/*******************************************************************************
*    exdem         (see manual, Command Ref., User Functions)    <fc>          *
*            This program calculates and writes a DEM export file.             *
*       Process starts by collecting profiles accross all dem grid points in   *
*       four directions (s to n,  w to e,  sw to ne,  se to nw).  After each   *
*       profile is calculated, elevations are calculated for the grid points   *
*       crossed.  When all profiles are completed, the elevations for each     *
*       grid points are averaged.  Next, these are smoothed by the number      *
*       of passes specified.  Each smoothing pas consists of giving each       *
*       grid elevation a weight of 2 and adding it to elevations on slope      *
*       between the points east and west, and north and south, then averaging  *
*       the elevations.  Note:  when smoothing, points outside the dem are     *
*       never used (for example, if a grid point is the western most point     *
*       in a grid row, then the elevation on slope between the points to the   *
*       east and west is not used.)  Finally, the elevations are converted     *
*       (if necessary) from the input elevation coordinates (say feet) to the  *
*       output elevation coordinates (usually meters).  Lastly, the dem file   *
*       is written.                                                            *
*                                                                              *
*       files/globals changed:  the output file & job_stats[51]  & diary file  *
*                                                                              *
*       values returned:  0 = file written,  1 = failure                       *
*                                                                              *
*       parameters:   char *fname  filename                                    *
*                     long c       grid coordinate system code                 *
*                     long w       width in grid coordinates                   *
*                     long h       height in griducoordinates                  *
*                     long g       clipping coordinate system code             *
*                     long e1      map attribute units code (1=ft, 2=meters)   *
*                     long e2      output dem units code (1=fl, 2=meters)      *
*                     long st      smooth type: 0=none, 1=all pts, 2+=variable *
*                     long f       number of filter/smoothing passes (0-3)     *
*                                                                              *
*            J.Dabritz   8/30/89                                               *
*******************************************************************************/

#include "../../incl/globals.h"
#include "../../incl/in.h"
#include <malloc.h>
#include <math.h>
#include <string.h>
#define PROFILE_PTS 10000


   /************************ globals for this file *************************/
typedef struct  {
   double dist;
   long elev;
   } PROF_STRUCT;

PROF_STRUCT *profile;

float **grid;                    /* grid array [setup to work like grid[][]*/
long **gweight;                 /* weight array for grid   */
long c1,cn,r1,rn;      /* limits of grid array */
long grid_w, grid_h;    /* grid cell width & height in output coords */
long output_coord_sys;
long grid_rows, grid_cols;      /* number of rows and columns in grid  */
long **column;                  /* for each column holds #pts, min, max,p1*/
long p_end;
long weight_index[2000];   /* index to weights for distances 0 - 2000  */


   /************************ forward references in this file ***************/
long raw_profile();
long get_profile();
int  compair_pr();
long prof2cellchk();
long on_grid();
long dem_free_all();
char *d24_15();
char *e12_6();
long accum_elev();
long propor_elev();


   /************************ primary reference in this file ****************/
long exdem( fname, c, w, h, g, e1, e2, st, f)
    char *fname;
    long c, w, h, g, e1, e2, st, f;
{
   OBJECTS object;
   ATTRIBUTES attribute;


    char string[2048];
   char actual[256];
   char file_name[256];
   char line[2048];
   char zeros[2048];
   char *gridptr0,*weightptr0,*smoothptr0,*columnptr0,*profileptr0;
   char part1[256],part2[256],part3[256],part4[256];
   char ddmmdddmm[32];
   char state[STRLIST];
   char ch[1024];
	char *ptr;

   float *gridptr;                  /* pointer to malloc space for grid */
   long *weightptr;                /* pointer to malloc space for weight ary */
   long *smoothptr;                /* pointer to malloc space for smooth ary */
   float **smooth;                  /* smooth array (holds 3 grid rows)       */
   long *columnptr;                /* pointer to malloc space for column     */
   long i,j,n,r,t;
	long t_end, limit;
   unsigned int size0, size1, size2, size3; /* bytes needed to make up 'size'*/
   long col,row;
   long col0, row0;
   long no_pts;
   long p1, p2;
   long ri[7];                     /* row-index array */
   long poly;                      /* polygon record number */
   long convert;                   /* flag */
   long dem_level, elev_pattern;   /* record A values */
   long planimetric_code, zone;    /* record A values */
   long ground_coords, sides, cc;  /* record A values */
   long accuracy_code;             /* record A values */
   long elev_coords;               /* record A values */
   long rows, cols;                /* record A values */
   long min_elev, max_elev;        /* record A & B values */
   long first_elev;                /* flag */
   long last_row;
   long strlength;
   long dd1,dd2,mm1,mm2;
   long col1;
   long blocks;
   long diff, max_change;
   long raster_size;
	long st1;

   int fd;

   float floatx, floaty;

	double gauge;
   double x,y;
   double dx, dy;
   double x1, y1;
   double matrix[2][3];
   double increment;
   double dist;
   double elev;
   double conversion;
   double angle, res_x, res_y, res_z;  /* record A values */
   double ln;
   double lat,lon;
	double dist1,dist2;
	double min_elv, max_elv;


		/***** init pointers ****/
	fd = -1;
   gridptr0 = (char *) 0;
	weightptr0 = (char *) 0;
	smoothptr0 = (char *) 0;
	columnptr0 = (char *) 0;
	profileptr0 = (char *) 0;

        /************ check for map loaded *****************************/
    if(!job_stats[0])
    {    message("no map is acquired, cannot write dem fale.");
        return(1);
    }

        /************ check for map assembled **************************/
    if(!job_stats[10])
    {    message("you must assemble the map befor exporting data.");
        return(1);
    }

        /************ check for map registration ***********************/
   if(!registered(0)    ||    !registered(c) ||  !registered(g) )
   {   message("This map needs to be registered (raster and/or output coords.) befor export.");
      return(1);
   }

        /************ check for attributes *****************************/
   if(job_stats[13] < 1)          
   {   message("you have no elevation (or any other) attributes.  DEM output not possible");
      return(1);
   }
   
      /************ check & set contour_interval *********************/
   get_cont_int();

   if(abort_cmd())
      return(1);

        /************ create & ouput file ******************************/
   if(setup_ofile(fname,actual))
   {   message("export dem, abort, location 1.0");
      return(1);
   }
   fd = openf(actual,WRITE);
   if(0 >= fd)
      return(1);


message("calculating limits & setting variables");
    /*================ calculate limits & set variables     ================*/
      /************ set file globals *********************************/
   grid_w = w;
   grid_h = h;
   output_coord_sys = c;

      /**************** find raster limits in output coords ********/
   corner[8][0][0] = 0.0;
   corner[8][0][1] = 0.0;
   corner[8][1][0] = 0.0;
   corner[8][1][1] = job_stats[2];
   corner[8][2][0] = job_stats[1];
   corner[8][2][1] = job_stats[2];
   corner[8][3][0] = job_stats[1];
   corner[8][3][1] = 0.0;
   for(i=0; i<4; i++)
      transform(&corner[8][i][0],&corner[8][i][1],0,c);


   if(abort_cmd())
   {   closef_ptr(&fd);
      return(1);
   }

      /**************** find limits of grid in output system ***/
   c1 = (corner[8][0][0] / w);          /* begin column of grid */
   c1 *= w;
   cn = (1 + (corner[8][0][0] / w));    /* end column of grid   */
   cn *= w;
   r1 = (corner[8][0][1] / h);          /* begin row of grid    */
   r1 *= h;
   rn = (1 + (corner[8][0][1] / h));    /* end row of grid      */
   rn *= h;
   for(i=0; i<4; i++)
   {   if(c1 > corner[8][i][0])
      {   c1 = (corner[8][i][0] / w);
         c1 *= w;
      }
      if(cn < corner[8][i][0])
      {   cn = (1 + (corner[8][i][0] / w));
         cn *= w;
      }
      if(r1 > corner[8][i][1])
      {   r1 =(corner[8][i][1] / h);
         r1 *= h;
      }
      if(rn < corner[8][i][1])
      {   rn = (1 + (corner[8][i][1] / h));
         rn *= h;
      }
   }


      /**************** find grid size *****************************/
   grid_cols = 1 + (cn - c1 + 0.5) / w;
   grid_rows = 1 + (rn - r1 + 0.5) / h;

   if(abort_cmd())
   {   closef_ptr(&fd);
      return(1);
   }

      /**************** malloc grid space **************************/
      /** size0 = size of grid array **/
   size0 = 4 * grid_rows * (1 + grid_cols);
      /** size1 = size of smooth array **/
/*   size1 = 12 * (grid_cols + 1);   */
   size1 = 28 * (grid_cols + 1);
      /** size2 = size of column array **/
   size2 = 20 * grid_cols;
      /** size3 = size of profile array **/
   size3 = 12 * PROFILE_PTS;

   if(update_temp())
   {   closef_ptr(&fd);
      return(1);
   }

      /**************** malloc space for grid array ****************/
   raster_size = (job_stats[1] + 1) * (job_stats[2] + 1) / 8;
   if(raster_size >= size0)
      gridptr0 = (char *) craster[0];
   else
   {   for(gridptr0 = malloc_ltp(size0);  gridptr0 == (char *) 0;  )
      {   if(freesomemore())
         {   closef_ptr(&fd);
            message("export dem, not enough RAM for grid variable.  aborting.");
            return(1);
         }
      }
   }

   gridptr = (float *) gridptr0;
   grid = (float **) gridptr;
   for(gridptr += grid_rows, i=0;  i<grid_rows; i++, gridptr += grid_cols)
      grid[i] = gridptr;
   for(i=0; i<grid_rows; i++)
   {   for(j=0; j<grid_cols; j++)
         grid[i][j] = 0.0;
   }

      /**************** malloc space for gweight array ****************/
   if(raster_size >= 2 * size0)
      weightptr0 = (char *) (craster[0] + size0);
   else
   {   for(weightptr0 = malloc_ltp(size0);  weightptr0 == (char *) 0;  )
      {   if(freesomemore())
         {   closef_ptr(&fd);
            dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
            message("export dem, not enough RAM for grid variable.  aborting.");
            return(1);
         }
      }
   }

   weightptr = (long *) weightptr0;
   gweight = (long **) weightptr;
   for(weightptr += grid_rows, i=0;  i<grid_rows; i++, weightptr += grid_cols)
      gweight[i] = weightptr;
   for(i=0; i<grid_rows; i++)
   {   for(j=0; j<grid_cols; j++)
         gweight[i][j] = 0;
   }

      /**************** malloc space for smoothing array ***********/
   for(smoothptr0 = malloc_ltp(size1);  smoothptr0 == (char *) 0;  )
   {   if(freesomemore())
      {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
         closef_ptr(&fd);
         message("export dem, not enough RAM for smooth variable.   aborting.");
         return(1);
      }
   }

   smoothptr = (long *) smoothptr0;
   smooth = (float **) smoothptr;
   for(smoothptr += 7,  i=0; i<7; i++, smoothptr += grid_cols)
      smooth[i] = (float *) smoothptr;

      /**************** malloc space for column array **************/
   for(columnptr0 = malloc_ltp(size2);  columnptr0 == (char *) 0;  )
   {   if(freesomemore())
      {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
         closef_ptr(&fd);
         message("export dem, not enough RAM for column variable.   aborting.");
         return(1);
      }
   }
   columnptr = (long *) columnptr0;
   column = (long **) columnptr;
   for(columnptr += grid_cols,  i=0; i<grid_cols; i++, columnptr += 4)
      column[i] = columnptr;

      /**************** malloc space for profile ********************/
   for(profileptr0 = malloc_ltp(size3);  profileptr0 == (char *) 0;  )
   {   if(freesomemore())
      {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
         closef_ptr(&fd);
         message("export dem, not enough RAM for profile variable.  aborting.");
         return(1);
      }
   }
   profile = (PROF_STRUCT *) profileptr0;


      /*** set weight_index array ****/
   weight_index[0] = 4000000;
   for(i=1; i<2000; i++)
   {   elev = 0.9999999 + (4000000.0  /  ((float) i * (float) i));
      weight_index[i] = elev;
   }


   if(abort_cmd())
   {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
      closef_ptr(&fd);
      return(1);
   }


      /**************** set first_row and # rows in column array ****/
   for(col = 0, x = c1; col < grid_cols; col++, x += w)
   {      /** let user know what is going on **/
      if(!(col %10))
      {   sprintf(string,"finding which grid points are on map area, column %d", col);
         message(string);

         if(abort_cmd())
         {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
            closef_ptr(&fd);
            return(1);
         }
      }
         /*  find first row on grid */
      for(row = 0, y = r1; row < grid_rows; row++, y += h)
      {      /* see if coords are on grid (see it point is on rt */
            /* of all 4 sides going clockwise) */
         x1 = x;
         y1 = y;
         transform(&x1,&y1,c,g);
         for(i=0; i<4; i++)
         {   dx = corner[g][(i+1) % 4][0] - corner[g][i][0];
            dy = corner[g][(i+1) % 4][1] - corner[g][i][1];
            matrix[0][0] = dx;
            matrix[0][1] = dy;
            matrix[0][2] = x1 - corner[g][i][0];
            matrix[1][0] = dy;
            matrix[1][1] = -dx;
            matrix[1][2] = y1 - corner[g][i][1];
            if(mat_solve(2,3,matrix[0]))
               break;
            if(matrix[1][2] < 0.0)
               break;
         }

         if(i == 4)   /*  inside map corners */
            break;
      }
      if(row < grid_rows)
      {   column[col][3] = row;
            /** find last row on column **/
         y = r1 + (grid_rows - 1) * h;
         for(row = grid_rows - 1; row > column[col][3]; row--, y -= h)
         {      /* see if coords are on grid (see it point is on rt */
               /* of all 4 sides going clockwise) */
            x1 = x;
            y1 = y;
            transform(&x1,&y1,c,g);
            for(i=0; i<4; i++)
            {   dx = corner[g][(i+1) % 4][0] - corner[g][i][0];
               dy = corner[g][(i+1) % 4][1] - corner[g][i][1];
               matrix[0][0] = dx;
               matrix[0][1] = dy;
               matrix[0][2] = x1 - corner[g][i][0];
               matrix[1][0] = dy;
               matrix[1][1] = -dx;
               matrix[1][2] = y1 - corner[g][i][1];
               if(mat_solve(2,3,matrix[0]))
                  break;
               if(matrix[1][2] < 0.0)
                  break;
            }
   
            if(i == 4)   /*  inside map corners */
            break;
         }
         column[col][0] = row - column[col][3] + 1;
      }
      else
         column[col][0] = 0;
   }


    /*================ accumulate grid elevations from 4 profile directions */
      /**************** accumulate from s to n profiles ************/
   for(col0=0; col0<grid_cols; col0++)
   {      /**************** 1. get profile *****************************/
      if(!(col0 % 5))  
      {   sprintf(string,"getting N profile for column %d",col0);
         message(string);

         if(abort_cmd())
         {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
            closef_ptr(&fd);
            return(1);
         }
      }
      if(0 > (no_pts = get_profile(col0,0,0,1)))      
      {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
         closef_ptr(&fd);
         message("export dem, abort location 2.");
         return(1);
      }

      if(no_pts >= PROFILE_PTS)
      {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
         closef_ptr(&fd);
         message("export dem, abort location 3.");
         return(1);
      }

      if(no_pts)
      {   /**************** 3. accumulate elevations on grid ***********/
         row = 0;
         dist = 0.0;
         increment = h;
         p1 = p2 = 0;
         for(;row < grid_rows ;  row++,  dist += increment)
			{
				if(accum_elev(no_pts,row,col0,&p1,&p2,dist,st))
					break;
			}
      }
   }


      /**************** accumulate from w to e profiles ************/
   for(row0=0; row0<grid_rows; row0++)
   {      /**************** 1. get profile *****************************/
      if(!(row0 % 5))  
      {   sprintf(string,"getting E profile for row %d",row0);
         message(string);

         if(abort_cmd())
         {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
            closef_ptr(&fd);
            return(1);
         }
      }
      if(0 > (no_pts = get_profile(0,row0,1,0)))      
      {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
         closef_ptr(&fd);
         message("export dem, abort location 4.");
         return(1);
      }
      if(no_pts >= PROFILE_PTS)
      {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
         closef_ptr(&fd);
         message("export dem, abort location 5.");
         return(1);
      }

      if(no_pts)
      {   /**************** 3. accumulate elevations on grid ***********/
         col = 0;
         dist = 0.0;
         increment = w;
         p1 = p2 = 0;
         for(;col < grid_cols ;  col++,  dist += increment)
			{	if(accum_elev(no_pts,row0,col,&p1,&p2,dist,st))
					break;
			}
      }
   }


      /**************** accumulate from sw to ne profiles **********/
   for(row0 = grid_rows - 1, col0=0; col0 < grid_cols; )
   {      /**************** 1. get profile *****************************/
      if(!((row0+col0) % 5))  
      {   sprintf(string,"getting NE profile from row %d, column %d",
          row0,col0);
         message(string);

         if(abort_cmd())
         {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
            closef_ptr(&fd);
            return(1);
         }
      }
      if(0 > (no_pts = get_profile(col0,row0,1,1)))      
      {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
         closef_ptr(&fd);
         message("export dem, abort location 6.");
         return(1);
      }
      if(no_pts >= PROFILE_PTS)
      {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
         closef_ptr(&fd);
         message("export dem, abort location 7.");
         return(1);
      }

      if(no_pts)
      {   /**************** 3. accumulate elevations on grid ***********/
         col = col0;
         row = row0;
         dist = 0.0;
         increment = sqrt((double) w*w + h*h);
         p1 = p2 = 0;
         for(;(col < grid_cols) && (row < grid_rows) ; row++, col++,
           dist += increment)
			{	if(accum_elev(no_pts,row,col,&p1,&p2,dist,st))
					break;
			}
      }
         /*** move down w side & e on s side ***/
      if(row0)
         row0--;
      else
         col0++;
   }



      /**************** accumulate from se to nw profiles **********/
   for(col0=0, row0=0; row0 < grid_rows; )
   {      /**************** 1. get profile *****************************/
      if(!((row0+col0) % 5))  
      {   sprintf(string,"getting NW profile from row %d, column %d",
          row0,col0);
         message(string);

         if(abort_cmd())
         {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
            closef_ptr(&fd);
            return(1);
         }
      }
      if(0 > (no_pts = get_profile(col0,row0,-1,1)))      
      {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
         closef_ptr(&fd);
         message("export dem, abort location 8.");
         return(1);
      }
      if(no_pts >= PROFILE_PTS)
      {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
         closef_ptr(&fd);
         message("export dem, abort location 8.");
         return(1);
      }

      if(no_pts)
      {   /**************** 3. accumulate elevations on grid ***********/
         col = col0;
         row = row0;
         dist = 0.0;
         increment = sqrt((double) w*w + h*h);
         p1 = p2 = 0;

         for(;(col>=0) && (row<grid_rows) ; row++, col--, dist += increment)
			{	if(accum_elev(no_pts,row,col,&p1,&p2,dist,st))
					break;
			}
      }
         /*** move e on s side & up on e side ***/
      if(col0 + 1 < grid_cols)
         col0++;
      else
         row0++;
   }

      /**************** average grid elevations ********************/
   for(row = 0; row < grid_rows; row++)
   {
      if(!(row % 10))
      {   sprintf(string,"averaging elevations from profiles onto grid, row %d",row);
         message(string);

         if(abort_cmd())
         {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
            closef_ptr(&fd);
            return(1);
         }
      }
      for(col = 0; col < grid_cols; col++)
      {   if(on_grid(row,col))
			{	if(0.0 < gweight[row][col])
            	elev = (double) grid[row][col] / (double) gweight[row][col];
				else
				{	elev = -999;
					gweight[row][col] = 99999999;
				}
            grid[row][col] =  elev;
         }
         else
            grid[row][col] = 0.0;
      }
   }


    /*================ smooth grid                          ================*/
   ri[0] = 0;
   ri[1] = 1;
   ri[2] = 2;
	ri[3] = 3;
	ri[4] = 4;
	ri[5] = 5;
	ri[6] = 6;
   x = contour_interval * .90  + 0.5;  /** limit change 90% of cont int. **/
   max_change = x;

      /**** case st==1,  smooth all grid points except fixed poly elevs **/
   if(st == 1) 
   {   for(i=0; i<f; i++)
      {      /*** copy first row onto smooth array **/
         for(col = 0; col < grid_cols; col++)
            smooth[ri[1]][col] = grid[0][col];
   
            /*** smooth all rows & columns which will be on final dem ******/
         for(row = 0; row < grid_rows ; row++)
         {   if(!(row % 10))
            {   sprintf(string,"smoothing pass %d,  row %d",i+1,row);
               message(string);
   
               if(abort_cmd())
               {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
                  closef_ptr(&fd);
                  return(1);
               }
            }
               /******* copy next row onto buffer ***/
            if((row + 1) < grid_rows)
            {   for(col = 0; col < grid_cols; col++)
                  smooth[ri[2]][col] = grid[row + 1][col];
            
            }
   
               /******* smooth grid row ******/
            for(col = 0;  col < grid_cols ; col++)
            {
                  /*** check to see if it is on final dem ***/
               if(on_grid(row,col))
               {
                     /*** give raw gridded elev a weight of 1 ****/
                  elev = smooth[ri[1]][col];
                  n = 1;
   
                     /*** calc smoothing e/w (weight of 2) ***/
                  if(on_grid(row,col-1)  &&  on_grid(row,col+1))
                  {   elev += smooth[ri[1]][col-1] + smooth[ri[1]][col+1];
                     n += 2;
                  }
   
                     /*** calc smoothing n/s (weight of 2) ***/
                  if(on_grid(row-1,col)  &&  on_grid(row+1,col))
                  {   elev += smooth[ri[2]][col] + smooth[ri[0]][col];
                     n += 2;
                  }
   
                     /*** calc smoothing sw/ne (weight of 1) ***/
                  if(on_grid(row-1,col-1)  &&  on_grid(row+1,col+1))
                  {   elev += (smooth[ri[2]][col+1] + smooth[ri[0]][col-1]) / 2;
                     n++;                  
                  }
   
                     /*** calc smoothing nw/se (weight of 1) ***/
                  if(on_grid(row+1,col-1)  &&  on_grid(row-1,col+1))
                  {   elev += (smooth[ri[2]][col-1] + smooth[ri[0]][col+1]) / 2;
                     n++;                  
                  }
   
                     /*** limit change to half of a contour interval ***/
                  elev = (elev / (double) n);
                  diff = elev - grid[row][col];
                  if(diff > max_change)
                     diff = max_change;
                  else if(diff < -max_change)
                     diff = -max_change;
   
                     /*** add smoothing to point ***/
                  grid[row][col] += diff;
               }
            }

               /******* increment buffer rows ****/
            j = ri[0];
            ri[0] = ri[1];
            ri[1] = ri[2];
            ri[2] = j;
         }
      }
   }
   else if((st > 1) || (st < -1))      /****** variable smoothing ************/
   {   if(st < 0)
         st = -st;
		dist1 = st + 0.5;
		i=dist1;
		if(i >= 2000)
			i = 1999;
		st1 = weight_index[i];

		dist2 = st * 1.41421 + 0.5;
		i=dist2;
		if(i >= 2000)
			i = 1999;
		st1 += weight_index[i];
		st1 += weight_index[i];

      for(i=0; i<f; i++)
      {      /*** copy first three rows onto smooth array **/
         for(col = 0; col < grid_cols; col++)
            smooth[ri[3]][col] = grid[0][col];

         if(1 < grid_rows)
			{	for(col = 0; col < grid_cols; col++)
					smooth[ri[4]][col] = grid[1][col];
			}

         if(2 < grid_rows)
			{	for(col = 0; col < grid_cols; col++)
					smooth[ri[5]][col] = grid[2][col];
			}

            /*** smooth all rows & columns which will be on final dem ******/
         for(row = 0; row < grid_rows ; row++)
         {   if(!(row % 10))
            {   sprintf(string,"smoothing pass %d,  row %d",i+1,row);
               message(string);
   
               if(abort_cmd())
               {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
                  closef_ptr(&fd);
                  return(1);
               }
            }
               /******* copy next row onto buffer ***/
            if((row + 3) < grid_rows)
            {   for(col = 0; col < grid_cols; col++)
                  smooth[ri[6]][col] = grid[row + 3][col];
            
            }
   
               /******* smooth grid row ******/
            for(col = 0;  col < grid_cols ; col++)
            {
                  /*** check to see if it is on final dem ***/
               if(on_grid(row,col))
               {
                     /*** variable smoothing controlled by st1 *****/
                  gauge = (double) gweight[row][col] / st1;
                  if(gauge < (double) 2.0)
                  {     /**** get initial weight *******/
							if(gauge >= (double) 1.0)
								n = 1.0 + (gauge - 1.0) * 10;
							else
								n = 1;
                     
                        /*** give raw gridded elev a weight of n ****/
							if(n)
                     	elev = n * smooth[ri[3]][col];
   
								/*** calc smoothing if gauge < .050 **/
							if(gauge > (double) 0.050)
								limit = 1;
							else if(gauge > (double) 0.010)
								limit = 2;
							else
								limit = 3;
							t_end = limit + 1;
							for(r = -limit; r < 1; r++)   /*** rows ***/
							{	if(!r)
									t_end = 0;
								for(t = -limit; t < t_end; t++)    /*** cols ***/
                     	{	if((on_grid(row+r,col+t)) && (on_grid(row-r,col-t)))
                     		{  elev += smooth[ri[3+r]][col+t] + smooth[ri[3-r]][col-t];
                       			n += 2;               
                     		}
								}
							}
     	 
								/*** limit change to 90% of one contour interval ***/
							max_elv = grid[row][col] + max_change;
							min_elv = grid[row][col] - max_change;

								/*** find weighted avg elev ***/
							if(n > 0)
                    		elev = (elev / (double) n);
							else
								elev = 0;

								/*** set smoothed grid elevation within limits ***/
							if(elev > max_elv)
								grid[row][col] = max_elv;
							else if(elev < min_elv)
								grid[row][col] = min_elv;
							else
								grid[row][col] = elev;
						}
               }
            }
   
               /******* increment buffer rows ****/
            j = ri[0];
            ri[0] = ri[1];
            ri[1] = ri[2];
            ri[2] = ri[3];
            ri[3] = ri[4];
            ri[4] = ri[5];
            ri[5] = ri[6];
            ri[6] = j;
         }
      }
   }


      /************ set lake/polygon elevations (overwrite previous) **/
   if(job_stats[7])          /** if there are polygon records **/
   {   for(row = 0; row < grid_rows; row++)
      {
         sprintf(string,"setting polygon elevations for row %d",row);
         message(string);

         if(abort_cmd())
         {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
            closef_ptr(&fd);
            return(1);
         }
         for(col = 0; col < grid_cols; col++)
         {
            if(on_grid(row,col))
            {      /*** convert grid center to raster coords ***/
               x = c1 + col * w;
               y = r1 + row * h;
               transform(&x,&y,c,0);
               floatx = x;
               floaty = y;
   
                  /*** find polygon that point is within    ***/
               poly = find_vpoly(floatx,floaty,0);
   
                  /*** if internal polygon, then get elevation **/
               if(poly > 0)
               {   if(get_objects(&object,poly))
                  {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
                     closef_ptr(&fd);
                     message("export dem, abort location 9.");
                     return(1);
                  }
                  if(object.i12[2])
                  {   if(get_attr(&attribute,object.i12[2]))
                     {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
                        closef_ptr(&fd);
                        message("export dem, abort location 10.");
                        return(1);
                     }
                     if(0 < sscanf(attribute.string,"%d",&i))
                        grid[row][col] = (float) i;
                  }
               }
            }
         }
      }
   }



      /******* convert to output elevation coordinate type ******/
   switch(e2 - e1)   {
   case 1:            /*  feet to meters  */
      conversion = 1200.0 / 3937.0;
      convert = 1;
      break;
   case 0:             /*  no change       */
      convert = 0;
      break;
   case -1:            /*  meters to feet  */
      conversion = 3937.0 / 1200.0;
      convert = 1;
      break;
   }
   if(convert != 0)
   {   sprintf(string,"converting elevations to output coords");
      message(string);

      if(abort_cmd())
      {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
         closef_ptr(&fd);
         return(1);
      }
      for(row = 0; row < grid_rows; row++)
      {  for(col = 0; col < grid_cols; col++)
         {  if(on_grid(row,col))
				{	if((grid[row][col] != -999) || (gweight[row][col] != 99999999))
               	grid[row][col] = (grid[row][col] * conversion);
				}
         }
      }
   }



message("****************************** writing output file,   A record.");

   if(abort_cmd())
   {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
      closef_ptr(&fd);
      return(1);
   }
    /*================ write output file                    ================*/
      /**************** read grid, count points, flag points not on map */
   
      /**************** write A record *****************************/
      /** 1.  file name & description **/
      /** calc GSC quad number (DDMMDDDMM, lat/lon) of SE corner **/
   lat = corner[1][3][1];
   lon = -corner[1][3][0];     /** lon = West longitude **/
   dd1 = lat;
   dd2 = lon;
   x = 60 * lat + .0000001;
   mm1 = x - dd1 * 60;
   x = 60 * lon + .0000001;
   mm2 = x - dd2 * 60;
   sprintf(ddmmdddmm,"%02d%02d%03d%02d",dd1,mm1,dd2,mm2);

      /** build part 1, name state code, quad number **/
   strcpy(state,header.state);
   state[2] = 0;
   trim(header.org);
   if(!strcmp(header.org,"usfs-gsc"))
      sprintf(part1,"%s, %2s %9s%64s",header.quad_unit,state,ddmmdddmm," ");
   else
      sprintf(part1,"(LTPlus dem NOT FROM GSC) %34 %2s  ",header.quad_unit,state,ddmmdddmm," ");
      
   part1[64] = 0;
   
      /** build part 2, "LINETRACE no_pts  CI = ci      **/
   sprintf(part2,"LINTRACE+ %-6d CI = %2d%9s",job_stats[11],
    contour_interval," ");
   part2[33] = 0;

      /** build part 3 "name, state code" **/
      /** check for standard/non-standard dem **/
   if((c!=2) || (w!=30) || (h!=30) || (g!=2) || (e2!=2))
      sprintf(part3,"%-33s","NON-STANDARD DEM");
   else
      sprintf(part3,"%s, %2s%34s",header.quad_unit,state," ");

   part3[33] = 0;

      /** build part 4, quad number FS **/
   sprintf(part4,"%s   FS",ddmmdddmm);
   part4[14] = 0;

      /** combine parts ***/
	trim(header.head1);
	trim(header.head2);
   if(!strcmp(header.org,"usfs-gsc") || 
      ((header.head1[0] == (char) 0)  && (header.head1[0] == (char) 0)))
   	sprintf(file_name,"%64s%33s%33s%14s",part1,part2,part3,part4);
	else
	{	for(i=0, ptr =file_name; i<144;  i++, ptr++)
		   *ptr = (char) 32;
		sprintf(file_name,"%s",header.head1);
		dollar_cnvt(file_name);
		sprintf(file_name + 72,"%s",header.head2);
		dollar_cnvt(file_name + 72);
		for(i=0, ptr =file_name; i<144;  i++, ptr++)
		{	if(!(*ptr))
				*ptr = (char) 32;
		}
		file_name[144] = (char) 0;
	}


      /** 2.  dem level code **/
   dem_level = 2;     

      /** 3.  code defining elevation pattern */
   elev_pattern = 1;    /** regular **/

      /** 4.  code defining ground planimetric reference **/
   planimetric_code = c - 1;      /** 1 = utm,  2 = state plane **/

      /** 5.  code defining zone in ground planimetric reference system **/
   if(c == 2)      /** utm **/
   {   trim(header.utmln);
      ln = 0.0;
       sscanf(header.utmln,"%lf",&ln);
      if(ln == 0.0)
      {   for(i=0, ln=0.0; i<4; i++)
            ln += corner[1][i][0];
         ln = ln / 4;
      }
      ln = (ln < 0.0) ?  -ln : ln;
      zone = 31 - (ln / 6.0);
   }   
   else            /** state plane **/
   {   zone = 0;
      sscanf(trim(header.stplzn),"%d",&zone);
   }

      /** 6.  map projection parameters **/
   zeros[0] = 0;
   for(i=0; i<15; i++)
      strcat(zeros,"    .000000000000000    ");

      /** 7.  code defining unit of measure for ground coords **/
   ground_coords = (c == 2) ?   2 :  1;

      /** 8.  code defining unit of measure for elevations **/
   elev_coords = e2;

      /** 9.  number of sides in the polygon which defines the coverage **/
   sides = 4;

      /** 10.  ground coordinates of the 4 corners **/
   cc = c;    /** corner[][][]   system reference **/

      /** 11.  min and max elevation on grid **/
      /** calculated with rows & columns in 15. below **/

      /** 12.  angle from primary axis to dem local reference **/
   angle = 0.0;

      /** 13.  accuracy code for elevations **/
   accuracy_code = 0;       /* unknown */

      /** 14.  dem spatial resolution x,y,z **/
   res_x = w;
   res_y = h;
   res_z = 1;


      /** 15.  number of rows and columns in grid **/
   rows = 1;
   cols = 0;
   min_elev = 0;
   max_elev = 0;
   first_elev = 1;
   for(col = 0; col < grid_cols; col++)
   {   if(column[col][0])
      {   cols++;
         last_row = column[col][3] + column[col][0] - 1;
         for(row = column[col][3]; row <= last_row; row++)
         {   elev = grid[row][col] + 0.5;
            n = elev;
            if(row == column[col][3])
            {   column[col][1] = n;
               column[col][2] = n;
            }
            else
            {   if(column[col][1] > n)
                  column[col][1] = n;
               else if(column[col][2] < n)
                  column[col][2] = n;
            }
         }
         if(first_elev)
         {   min_elev = column[col][1];
            max_elev = column[col][2];
            first_elev = 0;
         }
         else
         {   if(min_elev > column[col][1])
               min_elev = column[col][1];
            else if(max_elev < column[col][2])
               max_elev = column[col][2];
         }
      }
   }

   sprintf(line,"%-144s%6d%6d%6d%6d%360s%6d%6d%6d%s%s%s%s%s%s%s%s%s%s%s%6d%s%s%s%6d%6d",
   file_name, dem_level, elev_pattern, planimetric_code,zone,zeros,
   ground_coords, elev_coords, sides, 
   d24_15(corner[cc][0][0],ch),d24_15(corner[cc][0][1],ch+30),
   d24_15(corner[cc][1][0],ch+60),d24_15(corner[cc][1][1],ch+90),
   d24_15(corner[cc][2][0],ch+120),d24_15(corner[cc][2][1],ch+150),
   d24_15(corner[cc][3][0],ch+180),d24_15(corner[cc][3][1],ch+210),
   d24_15((double) min_elev,ch+240),d24_15((double) max_elev,ch+270),
   d24_15(angle,ch+300), accuracy_code, 
   e12_6(res_x,ch+330),e12_6(res_y,ch+360),e12_6(res_z,ch+390),rows,cols);


      /** write line **/
   n =strlen(line);
   if(0 > writeb(line,n,1,fd))
   {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
      closef_ptr(&fd);
      message("export dem, abort location 11.");
      return(1);
   }

      /** buffer line to increment of 1024 **/
   n = 1024 - n;
   for(i=0; i<n; i++)
      line[i] = 32;
   line[i] = 0;
      /** write line **/
   if(0 > writeb(line,n,1,fd))
   {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
      closef_ptr(&fd);
      message("export dem, abort location 12.");
      return(1);
   }


   
      /**************** write B records ****************************/
      /** find first column **/
   for(col1 = 0; column[col1][0] == 0; col1++)
   {   if((col1 + 1) == grid_cols)
         break;
   }
      /** write records **/
   blocks = 1;
   for(col = col1; col < grid_cols; col++)
   {
      if(((col-col1+1) % 10) == 0)
      {   sprintf(string,"writing B record for column %d",col-col1+1);
         message(string);

         if(abort_cmd())
         {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
            closef_ptr(&fd);
            return(1);
         }
      }
      if(column[col][0])
      {      /** row & column id **/
         row = 1;
   
            /** rows in record **/
         rows = column[col][0];
         cols = 1;
   
            /** planimetric coords of first elev in columm **/
         x = c1 + col * w;
         y = r1 + column[col][3] * h;
   
            /** local elevatin datum **/
         zeros[24] = 0;        /** datum is 0.0 **/
   
            /** min/max elevations for profile **/
         min_elev = column[col][1];
         max_elev = column[col][2];

            /** write header portion **/
         sprintf(line,"%6d%6d%6d%6d%s%s%24s%s%s",
            row,col-col1+1,rows,cols,d24_15(x,ch),d24_15(y,ch+30),zeros,
            d24_15((double)min_elev,ch+60),d24_15((double)max_elev,ch+90));

            /** add elevations **/
         strlength = strlen(line);
         
            /** write line header portion **/
         if(0 > writeb(line,strlength,1,fd))
         {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
            closef_ptr(&fd);
            message("export dem, abort location 13.");
            return(1);
         }

            /** write elevation records **/
         for(row = column[col][3], i=0; i<column[col][0]; i++, row++)
         {   sprintf(line,"%6.0f",grid[row][col]+0.5);
               /** write elevation onto record **/
            if(0 > writeb(line,6,1,fd))
            {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
               closef_ptr(&fd);
               message("export dem, abort location 14.");
               return(1);
            }

               /** check length **/
            strlength += 6;
            if(strlength > 1019)
            {      /** write 4 blanks onto record **/
               if(0 > writeb("    ",4,1,fd))
               {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
                  closef_ptr(&fd);
                  message("export dem, abort location 15.");
                  return(1);
               }
               strlength = 0;
               blocks++;
            }
         }

            /** pad file with blanks to increment of 1024 **/
         if(strlength)
         {   n = 1024 - strlength;
            for(i=0; i<n; i++)
               line[i] = 32;
            line[i] = 0;
               /** write blanks onto record **/
            if(0 > writeb(line,n,1,fd))
            {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
               closef_ptr(&fd);
               message("export dem, abort location 16.");
               return(1);
            }
            blocks++;
         }
      }
   }
   
      /**************** write C record (no rms data) ***************/
   message("writing C record.");
   sprintf(line,"%6d%24s%6d%24s%964s",
         0," ",0," "," ");
      /** write line to file **/
   if(0 > writeb(line,1024,1,fd))
   {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
      closef_ptr(&fd);
      message("export dem, abort location 17.");
      return(1);
   }
   blocks++;


      /***************** make file come out even on 4096 byte boundry **/
   sprintf(line,"%1024s"," ");
   for(; (blocks % 4) ; blocks++)
   {   if(0 > writeb(line,1024,1,fd))
      {   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);
         closef_ptr(&fd);
         message("export dem, abort location 18.");
         return(1);
      }
   }


   message("closing files & malloc space.");
    /*================ close out file & malloc space        ================*/
      /**************** free malloc space **************************/
   dem_free_all(gridptr0,weightptr0,smoothptr0,columnptr0,profileptr0);


        /************ close file ***************************************/
    if(closef_ptr(&fd))
    {   sprintf(string,"dem file complete, but cannot close file:  %s",actual);
        message(string);
        return(0);
    }

        /************ if a DOS file, copy to DOS partition/media *******/
    if(strchr(fname,':'))
    {   
#ifdef SV32
        sprintf(string,"Cannot copy to DOS under ATT Unix.  file is at: %s",
        actual);
        message(string);
#else
        sprintf(string,"Now copying file %s to DOS file %s.",actual,fname);
        message(string);
        sprintf(string,"doscp -m %s %s",actual,fname);
        if(system_ltp(string))
        {    sprintf(string,"copy failed, file is at %s",actual);
            message(string);
        }
        else
        {   purgef(actual);
            message("done.");
        }
#endif
    }

        /************ set file for dem exported ************************/
   job_stats[51] = job_stats[45] + 1;

   message("writing diary entry.");
        /************ write diary entry for dem export *****************/
   sprintf(string,"dem file output (#%d) to %s",job_stats[51],actual);
   put_diary(string);


        /************ return *******************************************/
    return(0);
}

/*******************************************************************************
*     get_profile                                                              *
*         This program generates a profile from map line and polygon data.     *
*     The profile generated has elevations in the output system (e_output)     *
*     and distances in the output coordinate system (c_output).  Disatances    *
*     are from the beginning point (parameter x0,y0 [in grid coords]) and      *
*     are in the direction of dx,dy (parameters, vector in grid coords).       *
*     Up to 10000 distance/elevation pairs are put into the profile.           *
*     Profiles are taken from line elevation attributes.                       *
*                                                                              *
*     parameters:  long x0, y0   row,col grid coordinates of start of profile  *
*                  long dx, dy   vector direction of profile, grid coordinates *
*                                                                              *
*     data changed:   (no files or system globals)                             *
*                     local/global profile  (array of prof_struct structures)  *
*                                           (holds dist/elev pairs where       *
*                           dist is in output coord system units and           *
*                           elev is in input attribute units  (meters or feet).*
*                                                                              *
*     Returned values:   long no_points    number of profile points (0 + )     *
*                                       returns -1 on error.                   *
*                                                                              *
*     John Dabritz    9/1/89                                                   *
*******************************************************************************/

long get_profile(x_0,y_0,dx0,dy0)
   long x_0,y_0;
   long dx0,dy0;
{
   double x,y;
   double d_x, d_y;
   double factor1, factor2;
   double base;
   double length;    
   double dx, dy;    
   double x1,y1,x2,y2;    
   double matrix[2][3];    
   double last_dist;    
   double dist1, delev1, slope1;    
   double dist2, delev2, slope2;    
   double dist, elev;    
   double de;                     /* difference in elevation */
   double max_elev;    
   double min_elev;    
   double runout_dist;    


   long hit;    
   long i, j, n;    
   long rounded;                  /* flag */
   long contour_above,contour_below;    


      /************ convert parameters from grid to output coords **/
   d_x = dx0 * grid_w;
   d_y = dy0 * grid_h;
   x = c1 + grid_w * x_0;
   y = r1 + grid_h * y_0;

      /************ normalize d_x, d_y vector **********************/
   length = sqrt(d_x * d_x  +  d_y * d_y);
	if(length > 0.0)
   {	d_x = d_x / length;
   	d_y = d_y / length;
	}

      /************ find profile endpoints/limit on raster area ****/
      /** (find limits in output coords) **/
      /** find intersections with all 4 sides of output system raster edges */
   for(hit=0, i=0; i<4; i++)
   {   matrix[0][0] = d_x;
      matrix[0][1] = corner[8][i][0] - corner[8][(i + 1) % 4][0];
      matrix[0][2] = corner[8][i][0] - x;
      matrix[1][0] = d_y;
      matrix[1][1] = corner[8][i][1] - corner[8][(i + 1) % 4][1];
      matrix[1][2] = corner[8][i][1] - y;
      if(!mat_solve(2,3,matrix[0]))
      {   if((matrix[1][2] >= 0.0)  &&  (matrix[1][2] <= 1.0))
         {   if(!hit)
            {   factor1 = factor2 = matrix[0][2];
               hit = 1;
            }
            else
            {   if(factor1 > matrix[0][2])
                  factor1 = matrix[0][2];
               if(factor2 < matrix[0][2])
                  factor2 = matrix[0][2];
            }
         }
      }
   }

   if(!hit)           /** if no hits, then off raster, return 0 (no pts) */
      return(0);

   x1 = x + d_x * factor1;     /** begin pt on raster area **/
   y1 = y + d_y * factor1;
   x2 = x + d_x * factor2;     /** end pt on raster area   **/
   y2 = y + d_y * factor2;

      /************ generate raw profile ***************************/
   base = factor1;
   p_end = 0;
   raw_profile(x1,y1,x2,y2,base);
   
   if(p_end == 0)
      return(0);

      /************ pack profile to end of array & *****************/
      /************ eliminate duplicate distance points ************/
   last_dist = -9999999.9;
   for(p_end--, i = PROFILE_PTS; p_end >= 0; p_end--)
   {   if(profile[p_end].dist != last_dist)
      {   i--;
         profile[i].dist = profile[p_end].dist;
         profile[i].elev = profile[p_end].elev;
      }
   }

      /************ round hilltops & depressions *******************/
   for(j = 2; (j<i-3) && (i < PROFILE_PTS); i++)
   {   rounded = 0;            /* flag **/
      if((j > 2) && (i < PROFILE_PTS - 2))
      {   if(profile[i].elev == profile[i+1].elev)
         {      /*** same elevation on both sides, check slopes each way */
            dist1 = profile[i].dist - profile[i-1].dist;
            delev1 = profile[i].elev - profile[i-1].elev;
            dist2 = profile[i+2].dist - profile[i+1].dist;
            delev2 = profile[i+2].elev - profile[i+1].elev;
            dist = profile[i+1].dist - profile[i].dist;
            if((dist1 != 0.0)  && (dist2 != 0.0))
            {  slope1 = delev1 / dist1 / 2.0;
               slope2 = delev2 / dist2 / 2.0;
               if((slope1 > 0.0)  &&  (slope2 < 0.0))    /* hilltop */
               {   rounded = 1;
                     /** find intersect of slopes **/
                  x = -slope2 * dist / (slope1 - slope2);
                  elev = x * slope1 + profile[i].elev;
               
                     /** find limiting elevation (90% to next contour) */
                  n = profile[i].elev / contour_interval;
                  contour_above = contour_interval * (n + 1);
                  de = contour_above - profile[i].elev;
                  max_elev = contour_above - 0.1 * de;
                  n = contour_above - 1;
                  if(max_elev > n)
                     max_elev = n;
                     
                  if(elev <= max_elev)     /** intersects within limit */
                  {      /*** write origional pt.  ***/
                     profile[j].dist = profile[i].dist;
                     profile[j].elev = profile[i].elev;
                     j++;
                        /*** add a single extra point **/
                     profile[j].dist = profile[i].dist + x;
                     profile[j].elev = elev + 10000000;   /* flag pt ***/
                     if((profile[j].dist != profile[i].dist) &&
                      (profile[j].dist != profile[i+1].dist))
                        j++;
                  }
						else
            		{  	/*** add 2 extra points on clipped slope */
                       	/*** write origional pt.  ***/
                    	profile[j].dist = profile[i].dist;
                    	profile[j].elev = profile[i].elev;
                    	j++;
                       	/*** write first clipped hilltop pt **/
                    	de = max_elev - profile[i].elev;
                    	x = de / slope1;
                    	if(x > 0.0)
                    	profile[j].dist = profile[i].dist + x;
                    	profile[j].elev = max_elev + 10000000; /* flag pt */
                    	if((profile[j].dist != profile[i].dist) &&
                     	(profile[j].dist != profile[i+1].dist))
                       	j++;
                       	/*** write second clipped hilltop pt **/
                    	de = profile[i].elev - max_elev ;
                    	x = de / slope2;
                    	if(x > 0.0)
                    	profile[j].dist = profile[i+1].dist - x;
                    	profile[j].elev = max_elev + 10000000; /* flag pt */
                    	if((profile[j].dist != profile[i].dist) &&
                     	(profile[j].dist != profile[i+1].dist))
                       	j++;
                  }
               }
               if((slope1 < 0.0)  &&  (slope2 > 0.0))    /* depression */
               {   rounded = 1;
                     /** find intersect of slopes **/
                  x = -slope2 * dist / (slope1 - slope2);
                  elev = x * slope1 + profile[i].elev;
               
                     /** find limiting elevation (90% to next contour) */
                  n = (profile[i].elev / contour_interval) - 0.000001;
                  contour_below = contour_interval * n;
                  de = profile[i].elev - contour_below ;
                  min_elev = contour_below + 0.1 * de;
                  n = contour_below + 1;
                  if(min_elev < n)
                     min_elev = n;
                     
                  if(elev >= min_elev)     /** intersects within limit */
                  {      /*** write origional pt.  ***/
                     profile[j].dist = profile[i].dist;
                     profile[j].elev = profile[i].elev;
                     j++;
                        /*** add a single extra point **/
                     profile[j].dist = profile[i].dist + x;
                     profile[j].elev = elev + 10000000; /* flag pt **/
                     if((profile[j].dist != profile[i].dist) &&
                      (profile[j].dist != profile[i+1].dist))
                        j++;
                  }
                  else
            		{  
                        /*** add 2 extra points on clipped slope */
                        /*** write origional pt.  ***/
                     profile[j].dist = profile[i].dist;
                     profile[j].elev = profile[i].elev;
                     j++;
                        /*** write first clipped depression pt **/
                     de = min_elev - profile[i].elev;
                     x = de / slope1;
                     profile[j].dist = profile[i].dist + x;
                     profile[j].elev = min_elev + 10000000; /* flag pt */
                     if((profile[j].dist != profile[i].dist) &&
                      (profile[j].dist != profile[i+1].dist))
                        j++;
                        /*** write second clipped depression pt **/
                     de = profile[i].elev - min_elev ;
                     x = de / slope2;
                     if(x > 0.0)
                     profile[j].dist = profile[i+1].dist - x;
                     profile[j].elev = min_elev + 10000000; /* flag pt */
                     if((profile[j].dist != profile[i].dist) &&
                      (profile[j].dist != profile[i+1].dist))
                        j++;
                  }
               }
            }
         }
      }

      if(!rounded)
      {   profile[j].dist = profile[i].dist;
         profile[j].elev = profile[i].elev;
         j++;
      }
   }

      /**** add runoput to beginning & end of profile.  go .90 toward
         next contour at current slope, then flat ****/
   dx = corner[8][3][0] - corner[8][0][0];
   dy = corner[8][3][1] - corner[8][0][1];
   runout_dist = sqrt(dx * dx  +  dy * dy);


      /** add runout to beginning of profile **/
   dist = profile[3].dist - profile[2].dist;
   de = profile[3].elev - profile[2].elev;
   if(j < 4)
   {   de = 0.0;
      dist = runout_dist;
   }
   if((dist > 0.0)  &&  (de != 0.0))
   {      /** write first runout point **/
      profile[1].dist = profile[2].dist - dist;
      profile[1].elev = profile[2].elev - de + 10000000; /* flag pt **/
   }
   else
   {      /** write first runout point **/
      profile[1].dist = profile[2].dist - runout_dist;
      profile[1].elev = profile[2].elev + 10000000; /* flag pt **/
   }

      /** write second runout point **/
   profile[0].dist = profile[1].dist - runout_dist;       
   profile[0].elev = profile[1].elev;  /** pt already flaged **/


      /** add runout to end of profile **/
   dist = profile[j-1].dist - profile[j-2].dist;
   de = profile[j-1].elev - profile[j-2].elev;
   if(j < 4)
   {   de = 0.0;
      dist = runout_dist;
   }
   if((dist > 0.0)  &&  (de != 0.0))
   {      /** write first runout point **/
      profile[j].dist = profile[j-1].dist + dist;
      profile[j].elev = profile[j-1].elev + de + 10000000;  /* flag pt **/
   }
   else
   {      /** write first runout point **/
      profile[j].dist = profile[j-1].dist + runout_dist;
      profile[j].elev = profile[j-1].elev + de + 10000000;  /* flag pt **/
   }

      /** write second runout point **/
   profile[j+1].dist = profile[j].dist + runout_dist;
   profile[j+1].elev = profile[j].elev; /*** point already flaged **/

   j += 2;

      /************ return number of profile points ****************/
   return(j);
}

/*******************************************************************************
*     raw_profile                                                              *
*         This program generates a profile from map line and polygon data.     *
*     The profile generated has elevations in the output system (e_output)     *
*     and distances in the output coordinate system (c_output).                *
*     Profiles are taken from line attributes.                                 *
*                                                                              *
*     parameters:  real ox1,oy1,ox2,oy2  output coord sys coords of endpoints  *
*                  real base         distance of x1 from profile 0 point       *
*                                                                              *
*     data changed:   (no files or system globals)                             *
*                     local/global profile  (array of prof_struct structures)  *
*                                           (holds dist/elev pairs where       *
*                           dist is in output coord system units and           *
*                           elev is in output elevation units (meters or feet).*
*                     long p_end      number of last profile point so far.     *
*                                                                              *
*     Returned values:   long 0 = ok,  1 = failure                             *
*                                                                              *
*     John Dabritz    9/5/89                                                   *
*******************************************************************************/

long raw_profile(ox1,oy1,ox2,oy2,base)
double ox1,oy1,ox2,oy2,base;
{
   long i;
   long ix1, iy1, ix2, iy2;
   long iystep, ixstep;
   long p_start;
   
   double x1,y1,x2,y2;
   double dx,dy;
   double slope;
   double xstep, ystep;
   double xenter, yenter;
   double xexit, yexit;
   double rx,ry,rx1,ry1,rx2,ry2;     /* fractional raster coordinates */
   double x,y;    double midpt_error;
   double length, lengtho;
   double cfactor;
   double oxm, oym;


      /************ find raster coordinates of endpts **************/
   rx1 = ox1;
   ry1 = oy1;
   transform(&rx1, &ry1, output_coord_sys, 0);
   rx2 = ox2;
   ry2 = oy2;
   transform(&rx2, &ry2, output_coord_sys, 0);

      /************ test midpoints for agreement within .5 pixel ***/
   x = oxm = (ox1 + ox2) / 2.0;
   y = oym = (oy1 + oy2) / 2.0;
   transform(&x,&y,output_coord_sys,0);
   rx = (rx1 + rx2) / 2.0;
   ry = (ry1 + ry2) / 2.0;

   dx = rx - x;
   dy = ry - y;
   midpt_error = sqrt(dx * dx  +  dy * dy);

      /************ if error greater than 0.5, break segment in half
                  & call recursively, otherwise do profile ********/
   if(midpt_error > 0.5)
   {      /*** call for first half ***/
      raw_profile(ox1,oy1,oxm,oym,base);
         /*** call for second half **/
      dx = oxm - ox1;
      dy = oym - oy1;
      length = sqrt(dx * dx  +  dy * dy);
      base += length;
      raw_profile(oxm,oym,ox2,oy2,base);
      return(0);
   }
   else
   {
         /************ find conversion factor between raster & output
                     coordinate systems for this profile ****/
      dx = ox2 - ox1;
      dy = oy2 - oy1;
      lengtho = sqrt(dx * dx  +  dy * dy);

      cfactor = lengtho;
   
         /************ generate raw profile ***************************/

         /*** check input for 0 length ********/
      if((rx1 == rx2)  &&  (ry1 == ry2))
         return(0);       /** can't cross a segment if ya don't move */

         /********* Generate list of all segments crossing profile section */
         /** first calculate index_array coordinate values for endpoints **/
      x1 = (rx1 + 0.5) / 32.0;            /* fractional cell x */
      y1 = (ry1 + 0.5) / 32.0;            /* fractional cell y */
      ix1 = x1;                         /* integer cell x    */
      iy1 = y1;                         /* integer cell y    */
      x2 = (rx2 + 0.5) / 32.0;            /* fractional cell x */
      y2 = (ry2 + 0.5) / 32.0;            /* fractional cell y */
      ix2 = x2;                         /* integer cell x    */
      iy2 = y2;                         /* integer cell y    */
   
         /******* next go through each cell accumulating a list of 
               segments which cross along with their distance from
               the beginning (rx1,ry1) of the profile section  */
   
      p_start = p_end;
         /** if profile section is entirely in 1 cell, check it ***/
      if((ix1 == ix2) && (iy1 == iy2))
      {   if(prof2cellchk(ix1,iy1,rx1,ry1,rx2,ry2))
            return(1);
      }
   
         /** otherwise check all cells through which the profile goes  **/
      else
      {   dx = x2 - x1;        /*  change in cell coordinates */
         dy = y2 - y1;
            /*** check for more vertical or horizontal vector */
         if(fabs(dy) > fabs(dx))       /* more vertical than hz */
         {      /* calc y slope */
            slope = dx / dy;
   
               /* set variables for checking by row */
            if(y2 > y1)
               ystep = 1.0;
            else
               ystep = -1.0;
            xstep = ystep * slope;
            if(xstep > 0.0)
               ixstep = 1;
            else
               ixstep = -1;
   
               /** set enter and exit variables for first col */
            xenter = x1;
            yenter = y1;
            if(y2 > y1)
            {   i = (yenter + ystep);
               yexit = i;
            }
            else
            {   i = (yenter);
               yexit = i;
            }
            xexit = xenter + (yexit - yenter) * slope;
   
               /*** check cells, one row at a time **/
            for( ; ; )
            {      /*** check cells for one row **/
               for(;; ix1 += ixstep)
               {      /** check 1 cell **/
                  if(prof2cellchk(ix1,iy1,rx1,ry1,rx2,ry2))
                     return(1);
   
                     /*** check for end of cell row used **/
                  if(ix1 == (long) xexit)
                     break;
   
                     /*** check if done **/
                  if((iy1 == iy2) && (ix1 == ix2))
                     break;
               }
                  /*** check if done **/
               if((iy1 == iy2) && (ix1 == ix2))
                  break;
   
                  /*** calc xenter, xexit for next row **/
               xenter = xexit;
               xexit = xenter + xstep;
   
               if(iy1 != iy2)
                  iy1 += ystep;
            }
         }
         else                          /* more horizontal that vt */
         {      /* calc x slope */
            slope = dy / dx;
   
               /* set variables for storing by col */
            if(x2 > x1)
               xstep = 1.0;
            else
               xstep = -1.0;
            ystep = xstep * slope;
            if(ystep > 0.0)
               iystep = 1;
            else
               iystep = -1;
   
               /** set enter and exit variables for first row */
            yenter = y1;
            xenter = x1;
            if(x2 > x1)
            {   i = (xenter + xstep);
               xexit = i;
            }
            else
            {   i = (xenter);
               xexit = i;
            }
            yexit = yenter + (xexit - xenter) * slope;
   
               /*** check path, one col at a time **/
            for( ; ; )
            {      /*** check 1 column of cells in a row **/
               for(;; iy1 += iystep)
               {      /** check 1 cell **/
                  if(prof2cellchk(ix1,iy1,rx1,ry1,rx2,ry2))
                     return(1);
   
                     /*** check for end of vector **/
                  if(iy1 == (long) yexit)
                     break;
   
                     /*** check if done ***/
                  if((iy1 == iy2) && (ix1 == ix2))
                     break;
               }
                  /*** check if done ***/
               if((iy1 == iy2) && (ix1 == ix2))
                  break;
   
                  /*** calc yenter, yexit for next col **/
               yenter = yexit;
               yexit = yenter + ystep;
                  /*** increment row **/
               if(ix1 != ix2)
                  ix1 += xstep;
            }
         }
      }
   
         /********* convert distances to output coord system *************/
      for(i=p_start; i<p_end; i++)
         profile[i].dist = base + profile[i].dist * cfactor;
   
         /********* return ***********************************************/
      return(0);
   }
}

/*******************************************************************************
*     compair_pr                                                              *
*          This is the compairison function for the qsort in atrbt_prifile.    *
*     It compairs the double (.dist) in two prof_struct structures.            *
*                                                                              *
*******************************************************************************/
int compair_pr(a,b)
   PROF_STRUCT *a, *b;
{
   if(a->dist > b->dist)
      return(1);
   if(a->dist < b->dist)
      return(-1);
   return(0);
}

/*******************************************************************************
*      prof2cellchk                                                            *
*             This program checks 1 index array cell against the input         *
*      contour profile section.  It stores segment numbers and their           *
*      intersection distances from the beginning of the section in the         *
*      global profile structure array.                                         *
*                                                                              *
*      J.Dabritz  9/1/89                                                       *
*******************************************************************************/



long prof2cellchk(cellx,celly,x1,y1,x2,y2)
long cellx,celly;           /** index_array cell coordinates **/
double x1,y1;               /** beginning point of attribute profile section */
double x2,y2;               /** ending point of attribute profile section */
{
   OBJECTS object;
   ATTRIBUTES attribute;
   INDEX_LIST index_list;
   COORD_LIST coord_list;

   long i;
   long hit;
   long pts;
   long v;
   long p_end1;
   long elevation;

   double px1,py1;            /** first profile point coords **/
   double pdx,pdy;            /** dx and dy for profile points **/
   double sx1,sy1,sx2,sy2;    /** adjacent segment vertex coords **/
   double matrix[2][3];       /** for solving for intersection of lines **/
   double side[5][2];
   double factor1, factor2;


      /********* check cell coordinates for on raster ****************/
   if((cellx < 0)  ||  (cellx > (job_stats[1] >> 5)))
      return(0);
   if((celly < 0)  ||  (celly > (job_stats[2] >> 5)))
      return(0);

      /** set marker at current profile points stored **/
   p_end1 = p_end;

   px1 = x1;
   py1 = y1;
   pdx = x2 - x1;   /** profile section deltas **/
   pdy = y2 - y1;

      /********* find limits of profile within cell ******************/
      /** load cell sides into array for easy processing **/
   side[0][0] = side[4][0] = side[1][0] = cellx * 32.0 - 0.5;
   side[2][0] = side[3][0] = side[0][0] + 32.0;
   side[0][1] = side[4][1] = side[3][1] = celly * 32.0 - 0.5;
   side[1][1] = side[2][1] = side[0][1] + 32.0;
      /** intersect sides with profile **/
   for(i=0, hit=0; i<4; i++)
   {   matrix[0][0] = x2 - x1;
      matrix[0][1] = side[i][0] - side[i+1][0];
      matrix[0][2] = side[i][0] - x1;
      matrix[1][0] = y2 - y1;
      matrix[1][1] = side[i][1] - side[i+1][1];
      matrix[1][2] = side[i][1] - y1;
      if(!mat_solve(2,3,matrix[0]))
      {   if((matrix[1][2] >= 0.0)  &&  (matrix[1][2] <= 1.0))
         {   if(!hit)
            {   hit = 1;
               factor1 = factor2 = matrix[0][2];
            }
            else
            {   if(matrix[0][2] < factor1)
                  factor1 = matrix[0][2];
               else if(matrix[0][2] > factor2)
                  factor2 = matrix[0][2];
            }
         }
      }
   }

   if(!hit)
      return(0);

   if(factor1 < 0.0)
      factor1 = 0.0;
   if(factor2 > 1.0)
      factor2 = 1.0;
   if(factor1 > factor2)
      return(0);
   

      /********* go through index_list linked-list for cell **********/
   index_list.next = index_array[celly][cellx];
   for(; index_list.next ;)
   {      /** get index_list record **/
      if(get_indexl(&index_list,index_list.next))
         return(1);

         /********* check object type for segment *******************/
      if(index_list.type == 1)      /** map segment **/
      {      /********* go through coord_list checking for intersections */
         if(get_coordlst(&coord_list,index_list.v1))
            return(1);
         sx1 = coord_list.x;
         sy1 = coord_list.y;
         for(v=index_list.v1+1;  v <= index_list.v2; v++)
         {      /********* get vertex ******************************/
            if(get_coordlst(&coord_list,v))
               return(1);
               /********* calc matrix variables *******************/
            sx2 = coord_list.x;
            sy2 = coord_list.y;

            matrix[0][0] = pdx;
            matrix[0][1] = sx2 - sx1;
            matrix[0][2] = sx2 - px1;
            matrix[1][0] = pdy;
            matrix[1][1] = sy2 - sy1;
            matrix[1][2] = sy2 - py1;

               /********* solve for intersection ******************/
            if(!mat_solve(2,3,matrix[0]))
            {      /***** if intersection between 0 & 1, a hit! ***/
               if((matrix[0][2] >= factor1)  &&  (matrix[0][2]  <= factor2)
                 && (matrix[1][2] >= 0.0)  &&  (matrix[1][2]  <= 1.0))
               {   /***** store elevation & dist ******************/
                  if(get_objects(&object,index_list.obj_rec))
                     return(-1);
                  if(object.i12[2])   /** segment has attribute **/
                  {   if(get_attr(&attribute,object.i12[2]))
                        return(-1);
                     if(sscanf(attribute.string,"%ld",&elevation))
                     {
                        profile[p_end].elev = elevation;
                        profile[p_end].dist = matrix[0][2];
                        p_end++;
                        if(p_end >= PROFILE_PTS)
                           return(1);
                     }
                  }
						else                /** segment has no attribute **/
						{		/******* check endpoints for elevations ********/
								/******* & proportion elevation between ********/
							if(propor_elev(&object,v-1,v,x1,y1,x2,y2,&elevation))
                     {
                        profile[p_end].elev = elevation;
                        profile[p_end].dist = matrix[0][2];
                        p_end++;
                        if(p_end >= PROFILE_PTS)
                           return(1);
                     }
						}
               }
            }
               /**** increment vector verticies **/
            sx1 = sx2;
            sy1 = sy2;
         }
      }
   }

      /********* sort profile pts added by distance ******************/
   pts = (p_end - p_end1);
   if(pts > 1)
      qsort(&profile[p_end1],pts,12,compair_pr);


      /********* return **********************************************/
   return(0);
}

/*******************************************************************************
*    on_grid                                                                   *
*       This program returns a true if the row,column specified is on the      *
*    output dem.  Otherwise it returns false (0).                              *
*                                                                              *
*    Parameters long row, col                                                  *
*                                                                              *
*    Returns    long 1   (if point on output dem)                              *
*                    0   (if point not on output dem)                          *
*    J.Dabritz   9/7/89                                                        *
*******************************************************************************/

long on_grid(row,col)
long row, col;
{
		/*** check for legal row & col numbers ***/
   if((row < 0)  ||  (row > grid_rows)  ||  (col < 0)  ||  (col > grid_cols))
      return(0);

		/*** check for row & col on active dem area ***/
   if((row < column[col][3])  ||  (row >= (column[col][3] + column[col][0])))
      return(0);

		/*** check for presence of profile/elevation data ***/
	if((grid[row][col] == -999) && (gweight[row][col] == 99999999))
		return(0);

   return(1);
}


/*******************************************************************************
*    d24_15                                                                    *
*      this program mimics the D24.15 format of the univac for formatting      *
*    output.                                                                   *
*                                                                              *
*    Parameter    double x        input value to be ascii formatted            *
*                 char   *ptr     string variable to put ascii data into       *
*    Returned value    string *string                                          *
*    J.Dabritz     9/8/89                                                      *
*******************************************************************************/

char *d24_15(x,ptr)
double x;
char *ptr;
{
   long power;

      /***        (C character numbers)     012345678901234567890123    */
      /*** write C language 24.14E format  (    X.XXXXXXXXXXXXXXE+YY) **/
      /*** to Univac Fortran 24.15D format (    .XXXXXXXXXXXXXXXD+YY) **/

      /*** first write E type format **/
   sprintf(ptr,"%24.14E",x);

      /** move the units digit to the rt of the decimal **/
   ptr[5] = ptr[4];
   ptr[4] = '.';

      /** increment the power by 1  &  replace 'E' with 'D' **/
   sscanf(ptr+21,"%d",&power);
   power++;
   if(x == 0.0)
      sprintf(ptr+20,"    ");      /** no exponent for 0 **/
   else
      sprintf(ptr+20,"D%+03d",power);

      /** return string **/
   return(ptr);
}

/*******************************************************************************
*    e12_6                                                                     *
*      this program mimics the E12.6  format of the univac for formatting      *
*    output.                                                                   *
*                                                                              *
*    Parameter    double x        input value to be ascii formatted            *
*                 char   *ptr     string variable to put ascii data into       *
*    Returned value    string *string                                          *
*    J.Dabritz    9/8/89                                                       *
*******************************************************************************/

char *e12_6(x,ptr)
double x;
char *ptr;
{
   long power;

      /***        (C character numbers)     012345678901    */
      /*** write C language 12.5 E format  ( X.XXXXXE+YY) **/
      /*** to Univac Fortran 12.6 E format ( .XXXXXXE+YY) **/

      /*** first write E type format **/
   sprintf(ptr,"%12.5E",x);

      /** move the units digit to the rt of the decimal **/
   ptr[2] = ptr[1];
   ptr[1] = '.';

      /** increment the power by 1  **/
   sscanf(ptr+9,"%d",&power);
   power++;
   if(x == 0.0)
      sprintf(ptr+8,"    ");      /** no exponent for 0 **/
   else
      sprintf(ptr+9,"%+03d",power);

      /** return string **/
   return(ptr);
}

/*******************************************************************************
*   dem_free_all                                                               *
*        This program frees up to 5 sets of malloc'ed space.                   *
*        J.Dabritz  2/13/90                                                    *
*******************************************************************************/

long dem_free_all(ptr0,ptr1,ptr2,ptr3,ptr4)
   char *ptr0,*ptr1,*ptr2,*ptr3,*ptr4;
{
   long raster_size;


   raster_size = (job_stats[1] + 1) * (job_stats[2] + 1) / 8;

   if((ptr0 != (char *) 0) && ((ptr0 < (char *) craster[0]) || (ptr0 >= (char *) (craster[0] + raster_size))))
      free_ltp(ptr0);
	else
		ptr0 = (char *) 0;

   if((ptr1 != (char *) 0) && ((ptr1 < (char *) craster[0]) || (ptr1 >= (char *) (craster[0] + raster_size))))
      free_ltp(ptr1);
	else
		ptr1 = (char *) 0;

   if((ptr2 != (char *) 0) && ((ptr2 < (char *) craster[0]) || (ptr2 >= (char *) (craster[0] + raster_size))))
      free_ltp(ptr2);
	else
		ptr2 = (char *) 0;

   if((ptr3 != (char *) 0) && ((ptr3 < (char *) craster[0]) || (ptr3 >= (char *) (craster[0] + raster_size))))
      free_ltp(ptr3);
	else
		ptr3 = (char *) 0;

   if((ptr4 != (char *) 0) && ((ptr4 < (char *) craster[0]) || (ptr4 >= (char *) (craster[0] + raster_size))))
      free_ltp(ptr4);
	else
		ptr4 = (char *) 0;

   disprstr_get(0,1);

   return(0);
}


/*******************************************************************************
*     accum_elev                                                               *
*          accumulates one elevation on the elevation grid.  This includes     *
*    rounding & determining weight & setting weight grid.                      *
*    JPD    code added/modified  5/7/90                                        *
*                                                                              *
*******************************************************************************/

long accum_elev(no_pts,row,col,p1ptr,p2ptr,dist,st)
	long no_pts,row,col,*p1ptr,*p2ptr;
	double dist;
	long st;
{
   static long p3,p4;
   static long i;
   static long elv1,elv2;
	static long st2,st5;

   static double de,dd,slope,elev,x;
   static double dist1,dist2,weight;


		/** calc st2 & st10 **/
	if(!st2)
	{	st2 = st / 2;
		st5 = 5 * st;
	}

      /** straddle distance with *p1ptr & *p2ptr **/
   for(i=0;(dist >= profile[*p2ptr].dist); (*p2ptr)++, i++)
	{	if((*p2ptr + 1) >= no_pts)
         return(1);
   }
   for(*p1ptr = *p2ptr;(dist < profile[*p1ptr].dist); (*p1ptr)--)
   {   if(*p1ptr == 0)
         return(1);
   }
   if(!((dist >= profile[*p1ptr].dist) && (dist <= profile[*p2ptr].dist)
     && (profile[*p1ptr].dist != profile[*p2ptr].dist)))
      return(1);

      /**** trim off inserted point flags, if any ****/
   elv2 = profile[*p2ptr].elev;
   if(elv2 > 100000)
{
      elv2 -= 10000000;
}

   elv1 = profile[*p1ptr].elev;
   if(elv1 > 100000)
{
      elv1 -= 10000000;
}

      /**** calc elevation on slope ***/
   de = elv2 - elv1;
   dd = (profile[*p2ptr].dist) - (profile[*p1ptr].dist);
	if(dd == 0.0)
		return(1);
   slope = de / dd;
   elev = elv1 + (dist-profile[*p1ptr].dist) * slope;

      /*** find distance to real (unprojected) profile points) **/
   for(p3 = *p1ptr;(profile[p3].elev >1000000); p3--) /** back pt **/
   {  if(p3 == 0)         
         break;
   }
   for(p4 = *p2ptr;(profile[p4].elev > 1000000); p4++) /** ahead pt **/
   {  if((p4 + 1) >= no_pts)
         break;
   }

      /*** dist1 = distance to nearest real (unprojected) pt.**/
   dist1 = dist - profile[p3].dist;
   dist2 = profile[p4].dist - dist;
   if(dist1 > dist2)
      dist1 = dist2;

      /*** dist2 = distance between unprojected pts.          **/
   dist2 = profile[*p2ptr].dist - profile[*p1ptr].dist;

      /*** find weight for elevation from this profile **/
		/** 1. get base weighted distance **/
	x = dist1 + dist2 + 0.5;
	i = x;

		/** 2. modify for pts very close to existing lines **/
	if(dist1 < st2)
	{	i = dist1 + 0.5;
		i = i << 1;
	}
	else if(dist1 < st)
	{	if(dist2 > st5)
		{	i = dist1 + 0.5;
			i += st5;
		}
	}
	
      /** 3. modify for projected data & rounding **********/
   if((profile[*p2ptr].elev > 1000000) || (profile[*p1ptr].elev > 1000000))
		i = i << 1;

      /*** flat (conflicting rounding info) ****/
   else if(profile[*p1ptr].elev == profile[*p2ptr].elev)
		i = i << 3;

      /***  set weight & weighted elevation ***/
   if((((unsigned long) i) >= 2000) || (i < 0))
      weight = 1;
   else
      weight = weight_index[i];


      /*** add weighted elevations                      ***/
   elev = elev * weight;
   grid[row][col] += elev;
   gweight[row][col] += weight;


   return(0);
}


/*******************************************************************************
*      propor_elev                                                             *
*             This program proportions an intersect elevation between          *
*      the nodes on either end of a segment.  If the nodes have no elevation   *
*      attributes, then the other segments attached to the segment are         *
*      querried.                                                               *
*                                                                              *
*      Returns 1 if proportioned elevation set, otherwise 0.                   *
*                                                                              *
*      J.Dabritz  12/4/90                                                      *
*******************************************************************************/

long propor_elev(object,v1,v2,x1,y1,x2,y2,elevation)
	OBJECTS *object;			  /** ptr to object record **/
	long v1,v2;               /** bounding object verticies **/
	double x1,y1,x2,y2;       /** bounding profile verticies **/
	long *elevation;          /** ptr to elevation to return **/
{
	OBJECTS node, segment, poly;
	ATTRIBUTES atrbt;
	COORD_LIST coord_list;

	long i,j,p,s;
	long hit;
	long elev[2];
	long v;

	double dist, length;
	double vx1,vy1,vx2,vy2;
	double dx,dy;
	double mat[2][3];


		/*********** check object for correct type ***********************/
	i = object->i12[0];
	if((i < 8) || (i > 10))
		return(0);

		/*********** check for v1, v2 order ******************************/
	if(v1 > v2)
	{	i = v1;
		v1 = v2;
		v2 = i;
	}

		/*********** check beginning of segment for elevation ************/
		/*********** check end of segment for elevation ******************/
	for(i=0; i<2; i++)
	{	hit = 0;      /** flag for elevation found **/
			/**** get node on end ****/
      if(get_objects(&node,object->i12[9+i]))
         return(0);
      if(node.i12[2])   /** node has attribute **/
      {   if(get_attr(&atrbt,node.i12[2]))
            return(0);
         hit = sscanf(atrbt.string,"%ld",&elev[i]);
		}
			/*** if no hit, check attached segments for elev ***/
		for(j=6;(j<10) && (node.i12[j] |= 0) && (hit < 1); j++)
		{	s = node.i12[j];
			if(s < 0)
				s = -s;
      	if(get_objects(&segment,s))
        		return(0);
      	if(segment.i12[2])   /** segment has attribute **/
      	{   if(get_attr(&atrbt,segment.i12[2]))
          			return(0);
        		hit = sscanf(atrbt.string,"%ld",&elev[i]);
			}
			else if(job_stats[7])        /* poly map, check adjacent polygons */
			{	for(p=7; segment.i12[p] && (p < 9); p++)
      		{	if(get_objects(&poly,segment.i12[p]))
        				return(0);
      			if(poly.i12[2])   /** poly has attribute **/
      			{   if(get_attr(&atrbt,poly.i12[2]))
          					return(0);
        				hit = sscanf(atrbt.string,"%ld",&elev[i]);
					}
				}
			}
		}
			/**** if no elevation, return 0 ***/
		if(hit < 1)
			return(0);
	}

		/*********** check for same elev   *******************************/
	if(elev[0] == elev[1])
	{	*elevation = elev[0];
		return(1);
	}

		/*********** get dist to intersect *******************************/
		/** add distances up to vertex pair containing intersect **/
	dist = 0;
	if(get_coordlst(&coord_list,object->i12[5]))
      return(0);
	vx1 = coord_list.x;
	vy1 = coord_list.y;
	
	for(v = 1 + object->i12[5]; v < v2; v++)
	{	if(get_coordlst(&coord_list,v))
         return(0);
		vx2 = coord_list.x;
		vy2 = coord_list.y;

		dx = vx2 - vx1;
		dy = vy2 - vy1;
		dist += sqrt(dx*dx + dy*dy);

		if(v > v2)
			break;

		vx1 = vx2;
		vy1 = vy2;
	}
		/** add distance from last vertex to intersect **/
	if(get_coordlst(&coord_list,v1))
        return(0);
	vx1 = coord_list.x;
	vy1 = coord_list.y;
	if(get_coordlst(&coord_list,v2))
        return(0);
	vx2 = coord_list.x;
	vy2 = coord_list.y;

	dx = vx2 - vx1;
	dy = vy2 - vy1;

	length = sqrt(dx*dx + dy*dy);
	if(length > 0.0)
	{	mat[0][0] = dx;
		mat[1][0] = dy;

		mat[0][1] = x2 - x1;
		mat[1][1] = y2 - y1;

		mat[0][2] = x1 - vx1;
		mat[1][2] = y1 - vy1;

      if(mat_solve(2,3,mat[0]))
         return(0);

		if((mat[0][2] > 0.0) && (mat[0][2] <= 1.0))
			dist += mat[0][2] * length;
	}

		/*********** set elevation ***************************************/
	if(object->f4[2] <= 0.0)
		return(0);

	dx = 0.5 + elev[0] + (dist / object->f4[2]) * (elev[1] - elev[0]);
	*elevation = dx;

		/*********** return **********************************************/
	return(1);
}
