/*******************************************************************************
*       exscsgef       (see manual, Command Ref., User Functions)       <fc>   *
*                                                                              *
*           This program writes a SCS Geographic Exchange Format file set      *
*           according to the parameters specified.                             *
*                                                                              *
*           Parameters:  char *fname    name of the file to be written         *
*                        long f_flag    feature file flag:  create if true     *
*                        char *doco     state abbr and soil survey name        *
*                                                                              *
*           Files/Values changed:  the exported file and job_stats[55]         *
*                                                                              *
*           Returns zero upon normal execution, 1=failure.                     *
*           C. Jonson    1/17/90                                               *
*******************************************************************************/
#include "../../incl/globals.h"
#include "../../incl/in.h"
#include <string.h>
#include <time.h>
#include <math.h>

static int fd[4];

long exscsgef( fname, f_flag )
	char *fname;
	long f_flag;
{
	double a;
	double x, y;

	long ex_abort();
	long d[4], g, i, j, k, m[4], s[4];
	long numgefs;
	long beg_coord, end_coord;
	long num_coords;
	long cent_end;
	long num_objs;
	long time(), clock;
	long rpoly_sn, lpoly_sn;

	char actual[4][256];
	char curr_date[32];
	char doco[STRLIST];
	char record[STRLIST];
	char string[256];
	char gefs[4][256];
	char lname[32];
	char org_name[32];
	char rpoly_attrstr[32];
	char lpoly_attrstr[32];
	static char suffix[4][8] = {".hdr", ".lin", ".txt", ".fea" };
	static char quad_cnr[4][8] = {"SW", "NW", "NE", "SE"};
	char *ptr;

	struct tm *tm_date;
	OBJECTS obj, obj2;
	ATTRIBUTES attr;
	COORD_LIST clist;


	fd[0] = -1;
	fd[1] = -1;
	fd[2] = -1;
	fd[3] = -1;

/************ check for map loaded *****************************/
    if(!job_stats[0])
    {	message("!-> No map acquired, cannot write SCSGEF file.");
        return(1);
    }

/************ check for map assembled **************************/
    if(!job_stats[5])
    {    message("!-> Map must have assembled lines or polygons to export in \
SCSGEF format.");
        return(1);
    }

/************ check for map registration ***********************/
	if( !registered(1) )
	{	message("This map needs to be registered (geodetic coords.) \
 before export.");
		return(1);
	}

/************ create & ouput file ******************************/
	message("setting up output files and variables.");
	if( f_flag )			/* Does user want optional feature file? */
		numgefs = 4;
	else
		numgefs = 3;

	for(ptr = fname, i=0; fname[i]; i++)
	{	if(fname[i] == '/')
			ptr = fname + i+1;
	}
	ptr[10] = CNULL;

	for( i = 0; i < numgefs; i++)
		sprintf(gefs[i], "%s%s", fname, suffix[i] );

	for(i = 0; i < numgefs; i++)
	{	if(setup_ofile(gefs[i], actual[i]))
		{	ex_abort(1);
			return(1);
		}
	}

/*** Initialize fd's to -1 ***/
	for( i = 0; i < 4; fd[i++] = -1);

/*** Open files ***/
	for( i = 0; i < numgefs; i++ )
	{	if( (fd[i] = openf(actual[i], WRITE | O_TRUNC)) == -1)
		{	ex_abort(2);
			return(1);
		}
	}
/*** Get date and set up mm/dd/yy format for header ***/
	message("Writing header file -->  ");
	if( (clock = time((long *)NULL)) == (time_t)-1)
	{	message("EXSCSGEF: Cannot get time for header. Date set to zeros.");
		strcpy(curr_date, "00/00/00");
	}
	if( clock != -1 )
		tm_date = localtime( &clock );
	sprintf(curr_date, "%d/%d/%d", tm_date->tm_mon + 1, 
				tm_date->tm_mday, tm_date->tm_year);

/*** Record 1 of header **/
	sprintf(doco, "%s %s", header.state, header.layer_name);
	doco[27] = CNULL;			/* Ensure it is only 27 characters */
	sprintf(record, "%-30s%-07.0f %-7s%-27s    ", lname, map_scale,
			"       ", doco);
	record[LINESZ - 1] = '\n';
	record[LINESZ] = CNULL;
	writea(record, fd[0]);

/*** Record 2 of header **/
	strncpy(org_name, header.org, 30);
	sprintf(record, "%-30s%-15s%-18s%-9s    ", org_name, curr_date, 
			"                  ", "lat/long");
	record[LINESZ - 1] = '\n';
	record[LINESZ] =  CNULL;
	writea(record, fd[0]);

/*** Record 3, 4, 5 and 6 of header **/
	for( i = 0; i < 4; i++)
	{	for(j = 0; j < 2; j++)
		{	if( j == 0 )		/* Convert longitude to West */
				a = -corner[1][i][j];
			d[j] = a;
			m[j] = (a - d[j]) * 60.0;
			s[j] = (((a - d[j]) * 60.0) - m[j]) * 60.0;
			if(s[j] > 59.9999)
			{	s[j] = 0.0;
				m[j]++;
			}
			if(m[j] == 60)
			{	m[j] = 0;
				d[j]++;
			}
		}
		sprintf(record, "%2s %+3.7f  %+3.7f  LONGITUDE   %3d/%02d/%02d  \
LATITUDE  %3d/%02d/%02d  ", quad_cnr[i], corner[1][i][0], corner[1][i][1], 
				d[0], m[0], s[0], d[1], m[1], s[1] );
		record[HDRSZ - 1] = '\n';
		record[HDRSZ] = CNULL;
		writea(record, fd[0]);
	}

/*** Close header file ***/
	append_msg("done");
	closef_ptr(&fd[0]);
	fd[0] = -1;

/***   Start line, text (point) and feature (optional) files   ***/
	if( job_stats[7] )
		num_objs = job_stats[7];
	else
		num_objs = job_stats[10];

	message("Writing line file -->  ");
	for( i = 0; i < num_objs; i++)
	{	if( get_objects(&obj, i) )
		{	message("EXSCSGEF:  Cannot get record from objects file. ");
			ex_abort(3);
			return(-1);
		}
	/*** Write segments to line file, write spurs to feature file ***/
		if( obj.i12[0] == 10 )	
		{ 	if( get_objects(&obj2, obj.i12[7]) )
			{	message("EXSCSGEF:  Cannot get record from objects file. ");
				ex_abort(4);
				return(-1);
			}
			if( obj2.i12[2] )
				if( get_attr(&attr, obj2.i12[2]) )
				{	message("EXSCSGEF:  Cannot get attribute. ");
					ex_abort(5);
					return(-1);
				}
			strcpy(rpoly_attrstr, attr.string);
			rpoly_sn = obj2.i12[1];

			if( get_objects(&obj2, obj.i12[8]) )
			{	message("EXSCSGEF:  Cannot get record from objects file. ");
				ex_abort(6);
				return(-1);
			}
			if( obj2.i12[2] )
				if( get_attr(&attr, obj2.i12[2]) )
				{	message("EXSCSGEF:  Cannot get attribute. ");
					ex_abort(7);
					return(-1);
				}
			strcpy(lpoly_attrstr, attr.string);
			lpoly_sn = obj2.i12[1];

		/*** Get the coord pairs that make up the segments   ***/
			beg_coord = obj.i12[5];
			end_coord = obj.i12[6];

		/*** Calculate number of coord pairs for current segment ***/
			num_coords = end_coord - beg_coord + 1;
			lpoly_attrstr[9] = rpoly_attrstr[9] = CNULL;

			sprintf(record, "%07d%-9s%-9s%-05d%-05d%05d  ", BOUNDARY,
				lpoly_attrstr, rpoly_attrstr, lpoly_sn, rpoly_sn,  num_coords );
			record[LINESZ - 1] = '\n';
			record[LINESZ] = CNULL;
			writea(record, fd[1]);

		/*** Get lat and long (decimal degrees) ***/
			for( ; beg_coord <= end_coord; beg_coord++ )
			{	if( get_coordlst(&clist, beg_coord) )
				{	message("EXSCSGEF:  Cannot get coordlist. ");
					ex_abort(8);
					return(-1);
				}
				x = clist.x;
				y = clist.y;
				if( transform(&x, &y, 0L, 1L) )
				{	message("EXSCSGEF: Cannot transform raster to lat/lon.  ");
					ex_abort(9);
					return(-1);
				}
				/*** Format and write coordinate records ***/
				x = fabs(x);	/* Ensure West is positive */
				sprintf(record,"%16s%+3.7f%+3.7f    ", "                ", x,y);
				record[LINESZ - 1] = '\n';
				record[LINESZ] = CNULL;
				writea(record, fd[1]);
			}
		}
	/*** If user wants feature file, write segs with 1 or 2 free ends to it ***/
		else if( f_flag && (obj.i12[0] == 8 || obj.i12[0] == 9) )
		{	if( obj.i12[2] )
		 		if( get_attr(&attr, obj.i12[2]) )
				{	message("EXSCSGEF:  Cannot get attribute. ");
					ex_abort(10);
					return(-1);
				}

			num_coords = obj.i12[6] - obj.i12[5] + 1;
			sprintf(record, "%-07d%-9s%19s%05d  ", LINEFEAT, attr.string,
					"                   ", num_coords );
			record[LINESZ - 1] = '\n';
			record[LINESZ] = CNULL;
			writea(record, fd[3]);

			for( g = obj.i12[5]; g <= obj.i12[6]; g++ )
			{	if( get_coordlst(&clist, g) )
				{	message("EXSCSGEF:  Cannot get coordlist. ");
					ex_abort(11);
					return(-1);
				}
				x = clist.x;
				y = clist.y;
				if( transform(&x, &y, 0L, 1L) )
				{	message("EXSCSGEF: Cannot transform raster to lat/lon.  ");
					ex_abort(12);
					return(-1);
				}
				x = fabs(x);	/* Ensure West is positive */
				sprintf(record, "%16s%+3.7f%+3.7f   ", "                ", x,y);
				record[LINESZ - 1] = '\n';
				record[LINESZ] = CNULL;
				writea(record, fd[3]);
			}
		}
	}
	append_msg("done");

/*** Find all isolated points and write to the text (point) file ***/
	message("Writing point file -->  ");
	for( i = 0; i < job_stats[5]; i++)
	{ 	if( get_objects(&obj, i) )
		{	message("EXSCSGEF:  Cannot get record from objects file. ");
			ex_abort(13);
			return(-1);
		}
		if( obj.i12[0] == 3 )		/*** Isolated point. Write to txt ***/
		{	if( obj.i12[2])
				if( get_attr(&attr, obj.i12[2]) )
				{	message("EXSCSGEF:  Cannot get attribute. ");
					ex_abort(14);
					return(-1);
				}
			x = obj.f4[0];
			y = obj.f4[1];
			if( transform(&x, &y, 0L, 1L) )
			{	message("EXSCSGEF: Cannot transform raster to lat/lon.  ");
				ex_abort(15);
				return(-1);
			}
			x = fabs(x);	/* Ensure West is positive */
			attr.string[9] = CNULL;
			sprintf(record, "%-07d%-9s%+3.7f%+3.7f  ", 
						SOILFEAT, attr.string, x,y);
			record[LINESZ - 1] = '\n';
			record[LINESZ] = CNULL;
			writea(record, fd[2]);
		}
	}

/*** Get centroids of interior polys and write to text (point) file ***/
	for( i = job_stats[7]; i < job_stats[10]; i++ )
	{	if( get_objects(&obj, i) )
		{	message("EXSCSGEF:  Cannot get record from objects file. ");
			ex_abort(16);
			return(-1);
		}
		if( obj.i12[0] != 14 )
			continue;

		if( obj.i12[2] )
			if( get_attr(&attr, obj.i12[2]) )
			{	message("EXSCSGEF:  Cannot get attribute. ");
				ex_abort(17);
				return(-1);
			}

		x = obj.f4[0];
		y = obj.f4[1];
		if( transform(&x, &y, 0L, 1L) )
		{	message("EXSCSGEF: Cannot transform raster to lat/lon.  ");
			ex_abort(18);
			return(-1);
		}
		x = fabs(x);	/* Ensure West is positive */
		attr.string[9] = CNULL;
		sprintf(record, "%-07d%-9s%+3.7f%+3.7f   ", MAPSYMBL,attr.string, x, y);
		record[LINESZ - 1] = '\n';
		record[LINESZ] = CNULL;
		writea(record, fd[2]);
	
		cent_end = obj.i12[11] > obj.i12[3] ? obj.i12[11] : obj.i12[3];
		for( k = obj.i12[3]; k <= cent_end; k++ )
		{	if( get_coordlst(&clist, k) )
			{	message("EXSCSGEF:  Cannot get coordlist. ");
				ex_abort(19);
				return(-1);
			}
			x = clist.x;
			y = clist.y;
			if( transform(&x, &y, 0L, 1L) )
			{	message("EXSCSGEF: Cannot transform raster to lat/lon.  ");
				ex_abort(20);
				return(-1);
			}	
			x = fabs(x);	/* Ensure West is positive */
			sprintf(record,"%-07d%-9s%+3.7f%+3.7f   ",MAPSYMBL,attr.string,x,y);
			record[LINESZ - 1] = '\n';
			record[LINESZ] = CNULL;
			writea(record, fd[2]);
		}
	}
	append_msg("done");

        /************ if a DOS file, copy to DOS partition/media *******/
    if(strchr(fname,':') != (char *)NULL )
    {   
#ifdef SV32
        sprintf(string,"Cannot copy to DOS under ATT UNIX. Files are at: %s \
(%s %s %s)", actual[0], suffix[1], suffix[2], suffix[3]);
        message(string);
#else
		for( i = 0; i < numgefs; i++)
		{	sprintf(string, "Now copying file %s to DOS file %s.",
					actual[i], gefs[i]);
        	message(string);
        	sprintf(string, "doscp -m %s %s", actual[i], gefs[i]);
        	if(system_ltp(string))
        	{	sprintf(string, "Copy failed, file is at %s", actual[i]);
            	message(string);
        	}
        	else
        	{	purgef(actual[i]);
            	message("done.");
        	}
		}
#endif
    }
    else
    {	if( numgefs == 3 )
		{	sprintf(string,"SCSGEF files written.  Files are :  \
%s (%s %s)", actual[0], suffix[1], suffix[2]);
        	message(string);
    	}
		else
		{	sprintf(string,"SCSGEF files written.  Files are :  \
%s (%s %s %s)",actual[0], suffix[1], suffix[2], suffix[3]);
        	message(string);
    	}
	}

/*** Set file for scsgef exported ***/
	job_stats[55] = job_stats[45] + 1;

/*** Write diary entry for digit export ***/
	sprintf(string,"scsgef file(s) output (#%d) to %s",job_stats[55],actual[0]);
	put_diary(string);

/*** Close all files ***/
	for( i = 0; i < 4; i++)
		if(fd[i] != -1)
			closef_ptr(&fd[i]);
	
/*** Return                 ***/
    return(0);
}


/*******************************************************************************
*      ex_abort                                                                *
*          This program is a convenient way of closing the output files and    *
*      writing a message to the user when the program aborts.  No system       *
*      globals or files are changed.                                           *
*                                                                              *
*      Parameter:   long  location    unique abort location to be printed.     *
*******************************************************************************/

long ex_abort(location)
	long location;
{	
	long i;
	char string[256];

/*** Close file fd if open ***/
	for( i = 0; i < 4; i++)
		if(fd[i] != -1)
			closef_ptr(&fd[i]);

/*** Print abort message ***/
	sprintf(string," Aborting, location %d", location);
	append_msg(string);

/*** Return ***/
	return(0);
}
