/*******************************************************************************
*       exxy           (see manual, Command Ref., User Functions)       <fc>   *
*                                                                              *
*           This program writes a xy (generic user formatted) ascii vector file*
*           according to the parameters specified.                             *
*                                                                              *
*           Parameters:  char *fname    name of the file to be written         *
*                        long c         coord_type code:  same as register     *
*                        long s         set number to be exported              *
*                        long cd        coord display type code                *
*                        long dl        delimeter (ascii decimal) value        *
*                        long dp        number of decimal places (0-16)        *
*                        long ht        header type code:                      *
*                                         1 = no header                        *
*                                         2 = header with attribute            *
*                                         3 = header, attribute in xy line(s)  *
*                                         4 = header only, with centroid       *
*                        long i1        item number begin column               *
*                        long i2        item number end column                 *
*                        long a1        attribute begin column                 *
*                        long a2        attribute end column                   *
*                        long vc1       vertex count begin column              *
*                        long vc2       vertex count end column                *
*                        long x1        x coord begin column                   *
*                        long x2        x coord end column                     *
*                        long y1        y coord begin column                   *
*                        long y2        y coord end column                     *
*                                                                              *
*           Files/Values changed:  the exported file and job_stats[52]         *
*                                                                              *
*           Returns zero upon normal execution, 1=failure.                     *
*           J.Dabritz,   2/20/91                                               *
*******************************************************************************/

#include "../../incl/globals.h"
#include <string.h>
#include "../../incl/in.h"

int fd;

long exxyabort();
long xy_col_check();
long xy_add_vtx_line();
long xy_add_long();
long xy_add_xys();
long xy_add_att();
long xy_add_dms();
long xy_add_double();

long exxy( fname,c,s,cd,dl,dp,ht,i1,i2,a1,a2,vc1,vc2,x1,x2,y1,y2 )
	char *fname;
	long c,s,cd,dl,dp,ht,i1,i2,a1,a2,vc1,vc2,x1,x2,y1,y2;
{
	OBJECTS object, segment;
	ATTRIBUTES attr;
	COORD_LIST coordlist;

	char actual [256];
	char string [256];

	double x,y;

	long item_no;
	long i,j,v,v1,v2;
	long type;
	long no_verticies;
	long seg,aseg;
	long increment;
	long head_len, xy_len;  /** output string lengths **/


        /************ init global **************************************/
	fd = -1;

        /************ check for map loaded *****************************/
    if(!job_stats[0])
    {    message("no map is acquired, cannot write xy file.");
        return(1);
    }

        /************ check for map assembled **************************/
    if(!job_stats[10])
    {    message("you must assemble the map befor exporting data.");
        return(1);
    }

        /************ check for map registration ***********************/
	if(!registered(0)    ||    !registered(c))
	{	message("This map needs to be registered (raster and/or output coords.) befor export.");
		return(1);
	}

      /************ check column specs *******************************/
		/*** decimal places, check for lt justification **/
	if((dp < 0) || (dp > 16))
	{	message("export xy:  Illegal decial place parameter.  Aborting.");
		return(1);
	}
	if(xy_col_check("item number",&i1,&i2,0,256))
		return(1);
	if(xy_col_check("attribute",&a1,&a2,0,256))
		return(1);
	if(xy_col_check("vertex count",&vc1,&vc2,0,256))
		return(1);
	if(xy_col_check("x coordinate",&x1,&x2,0,256))
		return(1);
	if(xy_col_check("y coordinate",&y1,&y2,0,256))
		return(1);
		/*** get max line length ****/
	head_len = 0;
	xy_len = 0;
	switch(ht) {
	case 1:     /** no header ***/
		xy_len = (i2 > a2) ?  i2:a2;
		xy_len = (xy_len > vc2) ?  xy_len:vc2;
		xy_len = (xy_len > x2) ?  xy_len:x2;
		xy_len = (xy_len > y2) ?  xy_len:y2;
		break;
	case 2:     /** header with attribute **/
		head_len = (i2 > a2) ?  i2:a2;
		head_len = (head_len > vc2) ?  head_len:vc2;
		xy_len = (x2 > y2) ?  x2:y2;
		break;
	case 3:     /** header, attribute in xy line(s)  **/
		head_len = (i2 > vc2) ?  i2:vc2;
		xy_len = (x2 > y2) ?  x2:y2;
		xy_len = (xy_len > a2) ?  xy_len:a2;
		break;
	case 4:     /** header only, with centroid       **/
		head_len = (i2 > a2) ?  i2:a2;
		head_len = (head_len > vc2) ?  head_len:vc2;
		head_len = (head_len > x2) ?  head_len:x2;
		head_len = (head_len > y2) ?  head_len:y2;
		break;
	}

        /************ create & ouput file ******************************/
	message("setting up output files and variables.");
	if(setup_ofile(fname,actual))
	{	exxyabort(11);
		return(1);
	}

        /************ open file ****************************************/
	fd = openf(actual,WRITE);
	if(fd < 0)
	{	exxyabort(12);
		return(1);
	}

	message("writing xy file.");
        /************ check all objects & write file  ******************/
	item_no = 0;
	for(i=0; i<job_stats[10]; i++)
	{		/********* inform user *************************************/
		if(!(i % 100))
		{	sprintf(string,"export xy, checking object %d, %d objects written to file.",i,item_no);
			message(string);
		}

			/********* check to see if object is in file ***************/
		if(get_objects(&object,i))
		{	exxyabort(1);
			return(1);
		}
		if(object.i12[0] < 8)
			type = 0;                      /** point     */
		else if(object.i12[0] < 14)
			type = 1;                      /** segment   */
		else if(object.i12[0] != 15)
			type = 2;                      /** polygon   */
		else
			type = -1;
		
		if(object.i12[4] & (1 << s))
		{		/********* count verticies *****************************/
			if(vc1)
			{	switch(type) {
				case 0:                         /* node */
					no_verticies = 1;
					break;
				case 1:                         /* segment */
					no_verticies = object.i12[6] - object.i12[5] + 1;
					break;
				case 2:                         /* polygon */
					no_verticies = 1;
					for(j=object.i12[5]; j<=object.i12[6]; j++) /* count seg vert */
					{	if(get_seglist(&seg,j))	    /* get each seg number */
						{	exxyabort(3);
							return(1);
						}
						aseg = abs(seg);
						if(get_objects(&segment,aseg))  /* get each segment */
						{	exxyabort(4);
								return(1);
						}
						no_verticies += segment.i12[6] - segment.i12[5];
					}
					break;
				}
			}
	
				/********* get attribute string ************************/
			if(a1)
			{	attr.string[0] = 0;         /* set to null */
				if(object.i12[2])
				{	if(get_attr(&attr,object.i12[2]))   /*  get attribute */
					{	exxyabort(5);
						return(1);
					}
				}
			}

				/********* write header line ***************************/
				/** init string ***/
			if(ht > 1)
			{	for(j=0; j<head_len; j++)
					string[j] = ' ';
				string[j] = (char) 10;    /*** linefeed ***/
				string[j+1] = (char) 0;   /*** null terminated string ***/
			}
				/** write header **/
			item_no++;
			switch(ht) {
			case 4:  /** header only with centroid values **/
					/** write x,y values **/
				x = object.f4[0];
				y = object.f4[1];
				transform(&x,&y,0,c);
				if(xy_add_xys(string,x,y,cd,dl,dp,x1,x2,y1,y2))
					return(1);
			case 2:  /** header, attribute in header **/
					/** write attribute **/
				if(xy_add_att(string,attr.string,a1,a2))
					return(1);
			case 3:  /** header, attribute in xy line(s) **/
					/** write item number **/
				if(xy_add_long(string,item_no,i1,i2))
					return(1);
					/** write vertex count **/
				if(xy_add_long(string,no_verticies,vc1,vc2))
					return(1);

					/*** output line ***/
				writea(string,fd);
				break;
			}

				/********* write  verticies ****************************/
			if(ht != 4)
			{	switch(type) {
				case 0:                         /* node */

					x = object.f4[0];
					y = object.f4[1];
					transform(&x,&y,0,c);
					if(xy_add_vtx_line(ht,xy_len,string,x,y,cd,dl,dp,x1,
					 x2,y1,y2,attr.string,a1,a2,item_no,i1,i2,no_verticies,vc1,vc2))
						return(1);
		
						/*** output line ***/
					writea(string,fd);
					break;
	
				case 1:                         /* segment */
					for(v=object.i12[5]; v<=object.i12[6]; v++)
					{	if(get_coordlst(&coordlist,v))
						{	exxyabort(6);
							return(1);
						}
						x = coordlist.x;
						y = coordlist.y;
						transform(&x,&y,0,c);
						if(xy_add_vtx_line(ht,xy_len,string,x,y,
						 cd,dl,dp,x1,x2,y1,y2,attr.string,a1,a2,item_no,i1,
						 i2,no_verticies,vc1,vc2))
							return(1);

							/*** output line ***/
						writea(string,fd);
					}
					break;
	
				case 2:                         /* polygon */
					for(j=object.i12[5]; j<=object.i12[6]; j++) /* do each object */
					{	if(get_seglist(&seg,j))	    /* get each seg number */
						{	exxyabort(8);
							return(1);
						}
						aseg = abs(seg);
						if(get_objects(&segment,aseg))  /* get each segment */
						{	exxyabort(9);
							return(1);
						}

						if(seg > 0)
						{	v1 = segment.i12[5];
							v2 = segment.i12[6];
							increment = 1;
						}
						else
						{	v1 = segment.i12[6];
							v2 = segment.i12[5];
							increment = -1;
						}

						for(v=v1; v != v2 ; v += increment)
						{	if(get_coordlst(&coordlist,v))
							{	exxyabort(10);
								return(1);
							}
							x = coordlist.x;
							y = coordlist.y;
							transform(&x,&y,0,c);
							if(xy_add_vtx_line(ht,xy_len,string,x,y,
						 	cd,dl,dp,x1,x2,y1,y2,attr.string,a1,a2,item_no,i1,
						 	i2,no_verticies,vc1,vc2))
								return(1);

								/*** output line ***/
							writea(string,fd);
						}
					}
						/******* write ending vertex ***********/
					if(get_coordlst(&coordlist,v))
					{	exxyabort(16);
						return(1);
					}
					x = coordlist.x;
					y = coordlist.y;
					transform(&x,&y,0,c);
					if(xy_add_vtx_line(ht,xy_len,string,x,y,
					 cd,dl,dp,x1,x2,y1,y2,attr.string,a1,a2,item_no,i1,
					 i2,no_verticies,vc1,vc2))
						return(1);

						/*** output line ***/
					writea(string,fd);
					break;
				}
			}
		}
	}

	message("closing files and freeing RAM space.");
        /************ close file ***************************************/
    if(closef_ptr(&fd))
    {   sprintf(string,"xy file complete, but cannot close file:  %s",actual);
        message(string);
        return(0);
    }

        /************ if a DOS file, copy to DOS partition/media *******/
    if(strchr(fname,':'))
    {   
#ifdef SV32
        sprintf(string,"Cannot copy to DOS under ATT Unix.  file is at: %s",
        actual);
        message(string);
#else
        sprintf(string,"Now copying file %s to DOS file %s.",actual,fname);
        message(string);
        sprintf(string,"doscp -m %s %s",actual,fname);
        if(system_ltp(string))
        {    sprintf(string,"copy failed, file is at %s",actual);
            message(string);
        }
        else
        {    purgef(actual);
            message("done.");
        }
#endif
    }
    else
    {    sprintf(string,"xy export done.  File is at:  %s",actual);
        message(string);
    }

        /************ set file for xy exported ************************/
	job_stats[61] = job_stats[45] + 1;

        /************ write diary entry for xy export *****************/
	sprintf(string,"xy file output (#%d) to %s",job_stats[52],actual);
	put_diary(string);

        /************ return *******************************************/
    return(0);
}


/*******************************************************************************
*      exxyabort                                                             *
*          This program is a convenient way of closing the output file and     *
*      writing a message to the user when the program aborts.  no system       *
*      globals or files are changed.                                           *
*                                                                              *
*      Parameter:   long  location    unique abort location to be printed.     *
*******************************************************************************/

long exxyabort(location)
long location;
{	
	char string[256];


		/************ close file fd if open ****************************/
	if(fd > 0)
		closef_ptr(&fd);

		/************* print abort message ******************************/
	sprintf(string,"export xy, aborted, location %d",location);
	message(string);

		/************* return *******************************************/
	return(0);
}


/*******************************************************************************
*      xy_col_check                                                          *
*          This program check column number specifications.                    *
*******************************************************************************/

long xy_col_check(name,i1,i2,min,max)
char *name;
long *i1, *i2, min, max;
{
	char string[96];
	long i;

		/*** check for legal values (min - max) ***/
	if((*i1 < min) || (*i2 < min) || (*i1 > max) || (*i2 > max))
	{	sprintf(string,"Export xy: Illegal column spec. for %s.  Aborting.",name);
		return(1);
	}

		/*** check for begin item < end item, invert if necessary ***/
	if(*i1 > *i2)
	{	i = *i1;
		*i1 = *i2;
		*i2 = i;
	}

	return(0);
}

/*******************************************************************************
*      xy_add_vtx_line                                                       *
*          writes 1 vertex line                                                *
*******************************************************************************/

long xy_add_vtx_line(ht,xy_len,string,x,y,cd,dl,dp,x1,x2,y1,y2,
 attr_string,a1,a2,item_no,i1,i2,no_verticies,vc1,vc2)
	long ht,xy_len;
	char *string;
	double x,y;
	long cd,dl,dp,x1,x2,y1,y2;
	char *attr_string;
	long a1,a2,item_no,i1,i2,no_verticies,vc1,vc2;
{
	long c;

		/****** build blank string ****/
	for(c=0; c<xy_len; c++)
		string[c] = ' ';
	string[c] = (char) 10;    /*** linefeed ***/
	string[c+1] = (char) 0;   /*** null terminated string ***/

		/****** add items as necessary *****/
	switch(ht) {
	case 1:  /** no header **/
			/** write item number **/
		if(xy_add_long(string,item_no,i1,i2))
			return(1);
			/** write vertex count **/
		if(xy_add_long(string,no_verticies,vc1,vc2))
			return(1);
	case 3:  /** header, attribute in xy line(s) **/
			/** write attribute **/
		if(xy_add_att(string,attr_string,a1,a2))
			return(1);
	case 2:  /** header, attribute in header **/
			/** write x,y values **/
		if(xy_add_xys(string,x,y,cd,dl,dp,x1,x2,y1,y2))
			return(1);
		break;
	}

	return(0);
}


/*******************************************************************************
*      xy_add_xys                                                            *
*          adds the x & y values to 1 output line                              *
*******************************************************************************/

long xy_add_xys(string,x,y,cd,dl,dp,x1,x2,y1,y2)
char *string;
double x,y;
long cd,dl,dp,x1,x2,y1,y2;
{
		/*** add x value ***/
	xy_add_dms(string,x,cd,dl,dp,x1,x2);

		/*** add y value ***/
	xy_add_dms(string,y,cd,dl,dp,y1,y2);

	return(0);
}

/*******************************************************************************
*      xy_add_dms                                                            *
*          adds a single double x or y value in dec deg or dm or dms format    *
*******************************************************************************/

long xy_add_dms(string,z,cd,dl,dp,z1,z2)
char *string;
double z;
long cd,dl,dp,z1,z2;
{
	double m,s;

	long id, im;
	long a, b;
	long i;
	long neg;


		/*** diferentiate by coordinate display type ***/
	switch(cd) {
	case 1:     /** decimal value ***/
		xy_add_double(string,z,z1,z2,dp);
		break;
	case 2:    /** degree decimal minute **/
			/** check for negative value **/
		if(z < 0.0)
		{	neg = 1;
			z = -z;
		}
		else
			neg = 0;

			/** break up number **/
		dms2(z,&id,&im,&s);
		m = im + (s / 60.0);  /** combine seconds with minutes **/

			/** add minutes **/
		i = dp + 2;
		if(dp)
			i++;
		a = z2 - i + 1;
		if(a < z1)
			a = z1;
		xy_add_double(string,m,a,z2,dp);

			/** add delim **/
		a--;
		if(a <  z1)
			break;
		string[a-1] = (char) dl;

			/** add degree **/
		a--;
		if(a <  z1)
			break;
		if(neg)
			id = -id;
		xy_add_long(string,id,z1,a);
		break;

	case 3:     /** degree, min, decimal second format **/
			/** check for negative value **/
		if(z < 0.0)
		{	neg = 1;
			z = -z;
		}
		else
			neg = 0;

			/** break up number **/
		dms2(z,&id,&im,&s);

			/** add seconds **/
		i = dp + 2;
		if(dp)
			i++;
		a = z2 - i + 1;
		if(a < z1)
			a = z1;
		xy_add_double(string,s,a,z2,dp);

			/** add delim **/
		a--;
		if(a <  z1)
			break;
		string[a-1] = (char) dl;

			/** add minutes **/
		a--;
		if(a <  z1)
			break;
		b = a;
		a--;
		if(a < z1)
			a = z1;
		xy_add_long(string,im,a,b);

			/** add delim **/
		a--;
		if(a <  z1)
			break;
		string[a-1] = (char) dl;

			/** add degree **/
		a--;
		if(a <  z1)
			break;
		if(neg)
			id = -id;
		xy_add_long(string,id,z1,a);
		break;
	}

	return(0);
}



/*******************************************************************************
*      xy_add_double                                                         *
*          adds a double value to one vertex line                              *
*******************************************************************************/

long xy_add_double(string,x,i1,i2,dp)
char *string;
double x;
long i1,i2,dp;
{
	char blank[40];
	char *ptr1, *ptr2, *end1;

	long len;
	long neg;


		/**** decrement column values for 0 base ***/
	i1--;
	i2--;

		/**** check for negative value ***/
	if(x < 0.0)
	{	neg = 1;
		x = -x;
	}
	else
		neg = 0;

		/**** write number to blank string ****/
	sprintf(blank,"0%.16f",x);

		/**** get string length ****/
	len = strlen(blank);

		/**** check rounding ****/
	ptr2 = blank + len - 16 + dp;
	if(*ptr2 > '4')
	{	for(ptr2--; ptr2 >= blank; ptr2--)
		{	if(*ptr2 == '.')
				continue;    /** do nothing **/
			else if(*ptr2 < '9')
			{	*ptr2 = (char) (*ptr2 + 1);
				break;
			}
			else
				*ptr2 = '0';
		}
	}

		/**** find begin, end pointers ***/
	if(dp)
		ptr2 = blank + len - 17 + dp;
	else
		ptr2 = blank + len - 18;
	ptr1 = string + i2;
	end1 = string + i1;
	
		/**** transfer digits ***/
	for(; (ptr1 >= end1) && (ptr2 >= blank); ptr1--, ptr2--)
		*ptr1 = *ptr2;

		/**** fill with zeros ***/
	for(;ptr1 >= end1; ptr1--)
		*ptr1 = '0';

		/**** add sign ***/
	if(neg)
	{	ptr1++;
		*ptr1 = '-';
	}

	return(0);
}

/*******************************************************************************
*      xy_add_att                                                            *
*          adds the attribute to one output line                               *
*******************************************************************************/

long xy_add_att(string,attr_string,a1,a2)
char *string, *attr_string;
long a1,a2;
{
	char *ptr1, *ptr2;
	char *endptr1, *endptr2;

		/** decrement a1, a2 for 0 base **/
	a1--;
	a2--;

		/** find item string length   **/
	attr_string[30] = (char) 0; /** truncate for safety **/
	endptr2 = attr_string + strlen(attr_string);

		/** transfer item into string **/
	endptr1 = string + a2;
	ptr1 = string + a1;
	ptr2 = attr_string;
	for(;(ptr1<=endptr1) && (ptr2<endptr2); ptr1++, ptr2++)
		*ptr1 = *ptr2;

		/** return **/
	return(0);
}

/*******************************************************************************
*      xy_add_long                                                           *
*          adds a long value to one vertex line                                *
*******************************************************************************/

long xy_add_long(string,item,i1,i2)
char *string;
long item,i1,i2;
{
	char blank[40];

	long c,i,neg;
	

		/** check for item addition at all **/
	if(!i1)
		return(0);
	
		/** decrement i1, 12 for 0 base *********/
	i1--;
	i2--;

		/** check for negative value ************/
	if(item < 0)
	{	neg = 1;
		item = -item;
	}
	else
		neg = 0;

		/** write long value into blank string **/
	sprintf(blank,"%d",item);
		
		/** transfer into string **/
	for(i=strlen(blank) - 1, c=i2; (i >= 0) && (c >= i1); i--, c--)
		string[c] = blank[i];

		/** fill with 0s **/
	for(; c >= i1; c--)
		string[c] = '0';

		/** add '-' sign if necess **/
	if(neg)
		string[i1] = '-';

		/** return **/
	return(0);
}
