/*******************************************************************************
*   impgef_fea       (see manual, Command Ref., User Functions)  <fc>          *
*                                                                              *
*       This program reads a GRASS SCSGEF line  format file of a specified     *
*       SCSGEF data set and stores the vectors on the specified level (35-95)  *
*                                                                              *
*            Parameters:   char *fname       the name of the dlg3 file to      *
*                                            import.                           *
*                          long level        the level number to import to     *
*                                            (35  - 95  inclusive)             *
*                          char *lname       new name for the level imported   *
*                                                                              *
*            files/globals changed:  (files) job_stats, index_array,           *
*                 index_list, level_names, v_objects,                          *
*                 v_coordlist,   (varaibles)  [bunches]                        *
*                                                                              *
*            Values returned:  (long) 0 = successful import                    *
*                                     1 = failure                              *
*                                                                              *
*            C. Jonson   2/5/90                                                *
*******************************************************************************/
#include "../../incl/globals.h"
#include "../../incl/in.h"
#include <string.h>

static int fd;
static long last_js16, last_js17, last_js18;

void impfea_abort();

long impgef_fea( fname, level, lname)
	char *fname;
	char *lname;
	long level;
{
	double lon, lat;

	long line;
	long num_coords;

	V_OBJECTS v_obj;
	COORD_LIST c_list;
	ATTRIBUTES v_atr;

	char actual[256];
	char *cptr, *end, *beg;
	char fea_atr[32];
	char record[64];
	char string[256];


	fd = -1;

/************* check for map present ***************************/
	if(!job_stats[0])
	{	message("You must ACQUIRE or CREATE a map before IMPORTing a file.");
		return(1);
	}

/************* check for raster registration *******************/
	if(!registered(0))
	{	message("You must REGISTER the Raster before importing a vector file.");
		return(1);
	}

/************* check input parameters **************************/
	if((level < 35)  ||  (level > 95))
	{	message("Invalid level number.  Valid level numbers are 35 to 95 inclusive.");
		return(1);
	}

/******* set up variables *************************************/
	message("import scsgef point, setting up variables.");
	last_js16 = job_stats[16];
	last_js17 = job_stats[17];
	last_js18 = job_stats[18];
	fd = -1;

/************* find & asgn pointer to import file **************/
	if(get_impfile(fname,actual))
	{	message("Cannot get_impfile().  ");
		impfea_abort(1);
		return(1);
	}

	if( (fd = openf(actual,READ)) == -1)
	{	sprintf(string, "Cannot open %s. ", actual);
		message(string);
		impfea_abort(2);
		return(1);
	}
	
	if(!registered(1))
	{	message("There is no map registration for latitude/longitude. ");
		impfea_abort(3);
		return(1);
	}

/*----------------------------------------------------------------------------
|  Logically treat all records as a series of header and coordinate  sets.   |
|  The 'first' record of a set contains the number of coordinate records for |
|  the segment it represents.  The following records contain the coordinte   |
|  pairs of the segment verticies.  There may be several sets per line file. |
----------------------------------------------------------------------------*/
	v_obj.type = 10 * level + 1;		/* Object type */
	for( line=0; reada(record, LINESZ, fd) != (char *)NULL; line++)   
	{	if(!(line %25))
		{	sprintf(string,"Reading & storing scsgef feature %d.", line+1);
			message(string);
		}

		/*** Read one record from line file and extract num of coord pairs ***/
			num_coords = 0;
			sscanf(record,"%*7c%9c%*19c%5d", fea_atr, &num_coords);
			if(num_coords < 1)
			{	message("Negative or zero number of verticies. ");
				impfea_abort(4);
				return(1);
			}
			v_obj.v1 = job_stats[18];

		/*** Calloc space for coord records (they are LINESZ bytes long) ***/
			cptr = (char *)calloc_ltp( num_coords, LINESZ );
			if( cptr == (char *)NULL )
			{	message("Cannot calloc %d bytes. ");
				impfea_abort(5);
				return(1);
			}

		/*** Read all coord records for above seg boundary 'header' record ***/
			if( readb(cptr, LINESZ, num_coords, fd) != LINESZ * num_coords )
			{	message("Cannot read all coord records for a segment. ");
				impfea_abort(6);
				return(1);
			}

		/*** Read all the coordinates and transform them to raster coords ***/
			end = cptr + num_coords * LINESZ;
			for( beg = cptr;  beg < end;  beg += LINESZ )
			{	sscanf(beg,"%*16c%lf%lf", &lon, &lat);
				if( transform(&lon, &lat, 1, 0) )
				{	message("Cannot transform from geo to raster. ");
					impfea_abort(7);
					return(1);
				}
				c_list.x = lon;
				c_list.y = lat;
				if(put_v_coord(&c_list, job_stats[18]))
				{	message("Cannot put_v_coord().  ");
					impfea_abort(8);
					return(1);
				}
			}

		/*** Write poly attributes to v_attribute file ***/
			strcpy(v_atr.string, fea_atr);	
			v_atr.next = 0;
			v_obj.atr = job_stats[16];
			if( put_v_attr(&v_atr, job_stats[16]) )
			{	message("Cannot put_v_attr(). ");
				impfea_abort(9);
				return(1);
			}

		/*** Write vobjects to the v_objects file ***/
			v_obj.v2 = job_stats[18] - 1;

			if(put_vobjects(&v_obj, job_stats[17]))		/* segment v_object */
			{	message("Cannot put_vobjects(). ");
				impfea_abort(10);
				return(1);
			}

		/*** Free calloced space for next set of coords ***/
			free_ltp(cptr);
	}

/************* close file **************************************/
	closef_ptr(&fd);

/************* if a copied dos file, purge it ******************/
#ifdef SV32
	/* can't access dos anyway */
#else
		/************* remove temp file, if any ************************/
	imp_rm_tmp(fname,actual);
#endif

/************* index v_obj imported ************************/
	if(index_vobj(last_js17))
	{	job_stats[17] = last_js17;
		return(1);
	}

/************* write diary *************************************/
	sprintf(string,"import scsgef feature file %s to level %d.",fname,level);
	put_diary(string);

/************* set job_stats ***********************************/
	job_stats[45]++;
	job_stats[35] = job_stats[45];

/************* set level name **********************************/
	if(*lname)
	{	lname[STRLIST - 2] = 0;
		sprintf(level_names[level],"%s",lname);
	}
	else
		sprintf(level_names[level],"imported scsgef feature file: %s",fname);

		/*********** change display to add imported level **************/
	if(level < 64)
		disp_i1hglight |= (1 << (level % 32));
	else
		disp_i2hglight |= (1 << (level % 32));

		/*********** re-display map ************************************/
	dispgraphics((long) ((rwindowxmin + rwindowxmax) / 2),(long) ((rwindowymin + rwindowymax) / 2),resolution_code);

/************* return ******************************************/
	return(0);
}


/*******************************************************************************
*      impfea_abort                                                            *
*          This program is a convenient way of closing the input  file and     *
*      writing a message to the user when the program aborts.  no system       *
*      globals or files are changed.                                           *
*                                                                              *
*      Parameter:   long  location    unique abort location to be printed.     *
*******************************************************************************/

void impfea_abort(location)
	long location;
{	
	char string[256];

/************ close file fd if open ****************************/
	if(fd > 0)
		closef_ptr(&fd);

/************* reset job_stats variables ************************/
	job_stats[16] = last_js16;
	job_stats[17] = last_js17;
	job_stats[18] = last_js18;

/************* print abort message ******************************/
	sprintf(string,"impgef_fea() aborting--> %d", location);
	append_msg(string);

/************* return *******************************************/
	return;
}
