/*******************************************************************************
*   impscsgef_hdr   (see manual, Command Ref., User Functions)  <fc>           *
*                                                                              *
*       This program reads a SCSGEF formatted header file and stores the       *
*       extracted data  on the specified 'level' (35 - 95).  Coordinates       *
*       are in lat/long format, which is standard for SCSGEF.                  *
*                                                                              *
*            Parameters:   char *fname       the name of the SCSGEF file to    *
*                                            import.                           *
*                          long level        the level number to import to     *
*                                            (35  - 95  inclusive)             *
*                          char *lname       new name for the level imported   *
*                                                                              *
*            files/globals changed:  (files) job_stats, index_array,           *
*                 index_list, level_names, v_attributes, v_objects,            *
*                 v_coordlist,   (varaibles)  [bunches]                        *
*                                                                              *
*            Values returned:  (long) 0 = successful import                    *
*                                     1 = failure                              *
*                                                                              *
*            C. Jonson  1/31/90                                                *
*******************************************************************************/

#include "../../incl/globals.h"
#include "../../incl/in.h"
#include <string.h>

long imphdr_abort();
static int fd;

long impscsgef( fname, level, lname)
char *fname;
long level;
char *lname;
{
	double lon, lat;

	long i;
	long offset;

	char actual[256];
	char dsn[64];
	char map_pro[32];
	char state_abbr[4];
	char rec_buf[HDRSZ * 6];
	char string[STRLIST];


/***  Initialize variable ***/
	fd = -1;

/***  Check for map present ***/
	if(!job_stats[0])
	{	message("You must ACQUIRE or CREATE a map before IMPORTing a file.");
		return(1);
	}

/***  Check for raster registration ***/
	if(!registered(0))
	{	message("You must REGISTER the Raster before importing a vector file.");
		return(1);
	}

/***  Find & assign pointer to import file ***/
	if(get_impfile(fname,actual))
	{	imphdr_abort(1);
		append_msg(" get_impfile aborted");
		return(1);
	}

	if( (fd = openf(actual,READ)) != -1)
	{	imphdr_abort(2);
		sprintf(string, " Cannot open %s", actual);
		append_msg(string);
		return(1);
	}
	
/***** Decode header section of scsgef header file **/
	if(readb(string, HDRSZ, 6, fd) != (HDRSZ * 6))
	{	imphdr_abort(3);
		message(" Header file has less than 6 records!");
		return(1);
	}

	/* Get data set name, scale, map projection, state abbr and survey name */
	sscanf(string, "%30c%7c %7c%27c", dsn, header.scale, map_pro, state_abbr);

	/*** Ignore the second header record.  Its fluffy data ***/
	/*** Extract corner coords from records 3 - 6          ***/
	for( i = 0, offset = HDRSZ * 2; i < 4; i++, offset += HDRSZ )
	{	sscanf(string + offset, "%*3c%lf %lf", &lon, &lat );
		corner[1][i][0] = lon;
		corner[1][i][1] = lat;
		if( transform(&lon, &lat, 1L, 0L) == -1 )
		{	imphdr_abort(4);
			message(" Cannot transform from geo to raster!");
			return(1);
		}
		
		corner[0][i][0] = lon;
		corner[0][i][1] = lat;
	}

/************* close file **************************************/
	message("import scsgef header, closing file.");
	closef_ptr(&fd);

/************* if a copied dos file, purge it ******************/
#ifdef SV32
	/* can't access dos anyway */
#else
		/************* remove temp file, if any ************************/
	imp_rm_tmp(fname,actual);
#endif

/************* write diary *************************************/
	sprintf(string,"import scsgef header file %s to level %d.",fname,level);
	put_diary(string);

/************* set job_stats ***********************************/
	job_stats[45]++;
	job_stats[34] = job_stats[45];

/************* set level name **********************************/
	if(*lname)
	{	lname[STRLIST - 2] = 0;
		sprintf(level_names[level],"%s",lname);
	}
	else
		sprintf(level_names[level],"imported scsgef header file: %s",fname);

		/*********** change display to add imported level **************/
	if(level < 64)
		disp_i1hglight |= (1 << (level % 32));
	else
		disp_i2hglight |= (1 << (level % 32));

		/*********** re-display map ************************************/
	dispgraphics((long) ((rwindowxmin + rwindowxmax) / 2),(long) ((rwindowymin + rwindowymax) / 2),resolution_code);

/************* return ******************************************/
	return(0);
}


/*******************************************************************************
*      imphdr_abort                                                            *
*          This program is a convenient way of closing the input  file and     *
*      writing a message to the user when the program aborts.  no system       *
*      globals or files are changed.                                           *
*                                                                              *
*      Parameter:   long  location    unique abort location to be printed.     *
*******************************************************************************/

long imphdr_abort(location)
	long location;
{	
	char string[STRLIST];

/************ close file fd if open ****************************/
	if(fd > 0)
		closef_ptr(&fd);

/************* print abort message ******************************/
	sprintf(string,"import scsgef header aborted: location %d",location);
	message(string);

/************* return *******************************************/
	return(0);
}
