/*******************************************************************************
*   impscsgef_pnt    (see manual, Command Ref., User Functions)  <fc>          *
*                                                                              *
*       This program reads a GRASS SCSGEF point format file of a specified     *
*       SCSGEF data set and stores the  nodes  on the specified level (35-95)  *
*       This 'ass-u-me's  special soil features are identified with the code   *
*       of '9000330'.  Soil map symbols are ignored as they cannot fit our     *
*       type of topology building.                                             *
*                                                                              *
*            Parameters:   char *fname       the name of the dlg3 file to      *
*                                            import.                           *
*                          long level        the level number to import to     *
*                                            (35  - 95  inclusive)             *
*                          char *lname       new name for the level imported   *
*                                                                              *
*            files/globals changed:  (files) job_stats, index_array,           *
*                 index_list, level_names, v_objects,                          *
*                 v_coordlist,   (varaibles)  [bunches]                        *
*                                                                              *
*            Values returned:  (long) 0 = successful import                    *
*                                     1 = failure                              *
*                                                                              *
*            C. Jonson   2/9/90                                                *
*******************************************************************************/
#include "../../incl/globals.h"
#include "../../incl/in.h"
#include <string.h>

long strtol();
static int fd;
static long last_js16, last_js17, last_js18;

void imppoint_abort();

long impgef_pnt( fname, level, lname)
	char *fname;
	char *lname;
	long level;
{
	double lon, lat;

	long code;
	long file_size;
	long point;

	V_OBJECTS v_obj;
	COORD_LIST c_list;
	ATTRIBUTES v_atr;

	char actual[256];
	char *cptr, *end, *beg;
	char soil_code[32], soil_symb[32];
	char string[256];


	fd = -1;

/************* check for map present ***************************/
	if(!job_stats[0])
	{	message("You must ACQUIRE or CREATE a map before IMPORTing a file.");
		return(1);
	}

/************* check for raster registration *******************/
	if(!registered(0))
	{	message("You must REGISTER the Raster before importing a vector file.");
		return(1);
	}

/************* check input parameters **************************/
	if((level < 35)  ||  (level > 95))
	{	message("Invalid level number.  Valid level numbers are 35 to 95 inclusive.");
		return(1);
	}

/******* set up variables *************************************/
	message("import scsgef point, setting up variables.");
	last_js16 = job_stats[16];
	last_js17 = job_stats[17];
	last_js18 = job_stats[18];
	fd = -1;

/************* find & asgn pointer to import file **************/
	if(get_impfile(fname,actual))
	{	message("Cannot get_impfile().  ");
		imppoint_abort(1);
		return(1);
	}

	if( (fd = openf(actual,READ)) == -1)
	{	sprintf(string, "Cannot open %s. ", actual);
		message(string);
		imppoint_abort(2);
		return(1);
	}
	
	if(!registered(1))
	{	message("There is no map registration for latitude/longitude. ");
		imppoint_abort(3);
		return(1);
	}


/*** Calloc space for coord records (they are POINTSZ bytes long) ***/
	file_size = filesize(actual);		/* Number of bytes in file */
	cptr = (char *)calloc_ltp( file_size, 1 );
	if( cptr == (char *)NULL )
	{	message("Cannot calloc %d bytes. ");
		imppoint_abort(4);
		return(1);
	}

/*** Read all coord records for above seg boundary 'header' record ***/
	if( reada(cptr, file_size, fd) == (char *)NULL )
	{	message("Cannot read all coord records for a segment. ");
		imppoint_abort(5);
		return(1);
	}

/*** Read all the coordinates and transform them to raster coords ***/
	v_obj.type = 10 * level + 0;		/* Node type */
	end = cptr + file_size;

	for( point = 0, beg = cptr;  beg < end;  beg += POINTSZ, point++ )
	{	sscanf(beg, "%7c%9c%3.8f%3.8f", soil_code, soil_symb, &lon, &lat);
	/*** Transform soil_code to long value and check for soil feature ***/
		code = strtol(soil_code, (char **)NULL, 10);
		if( code != 9000130 )
			continue;

	/*** Tell user whar in tarnation he/she is! ***/
		if(!(point %25))
		{	sprintf(string,"Reading & storing scsgef point %d.", point+1);
			message(string);
		}

	/*** Process info and write 'v' type files ***/
		if( transform(&lon, &lat, 1, 0) )
		{	message("Cannot transform from geo to raster. ");
			imppoint_abort(6);
			return(1);
		}
		c_list.x = lon;
		c_list.y = lat;
		if( put_v_coord(&c_list, job_stats[18]) )
		{	message("Cannot put_v_coord().  ");
			imppoint_abort(7);
			return(1);
		}

	/*** Write node attribute to v_attribute file ***/
		if( strlen(soil_symb) )
		{	strcpy(v_atr.string, soil_symb);	
			v_obj.atr = job_stats[16];
			v_atr.next = 0;
		}
		else
		{	v_obj.atr =  v_atr.next = 0;
			v_atr.string[0] = CNULL;
		}

		if( put_v_attr(&v_atr, job_stats[16]) )
		{	message("Cannot put_v_attr(). ");
			imppoint_abort(8);
			return(1);
		}

	/*** Write vobjects to the v_objects file ***/
		v_obj.v1 = v_obj.v2 = job_stats[18] - 1;
	
		if(put_vobjects(&v_obj, job_stats[17]))		/* segment v_object */
		{	message("Cannot put_vobjects(). ");
			imppoint_abort(9);
			return(1);
		}
	}

/*** Free calloced space for next set of coords ***/
	free_ltp(cptr);

/************* close file **************************************/
	message("import scsgef point, closing file.");
	closef_ptr(&fd);

/************* if a copied dos file, purge it ******************/
#ifdef SV32
	/* can't access dos anyway */
#else
		/************* remove temp file, if any ************************/
	imp_rm_tmp(fname,actual);
#endif

/************* index v_obj imported ************************/
	if(index_vobj(last_js17))
	{	job_stats[17] = last_js17;
		return(1);
	}

/************* write diary *************************************/
	sprintf(string,"import scsgef point file %s to level %d.",fname,level);
	put_diary(string);

/************* set job_stats ***********************************/
	job_stats[45]++;
	job_stats[35] = job_stats[45];

/************* set level name **********************************/
	if(*lname)
	{	lname[STRLIST - 2] = 0;
		sprintf(level_names[level],"%s",lname);
	}
	else
		sprintf(level_names[level],"imported scsgef point file: %s",fname);

		/*********** change display to add imported level **************/
	if(level < 64)
		disp_i1hglight |= (1 << (level % 32));
	else
		disp_i2hglight |= (1 << (level % 32));

		/*********** re-display map ************************************/
	dispgraphics((long) ((rwindowxmin + rwindowxmax) / 2),(long) ((rwindowymin + rwindowymax) / 2),resolution_code);

/************* return ******************************************/
	return(0);
}


/*******************************************************************************
*      imppoint_abort                                                          *
*          This program is a convenient way of closing the input  file and     *
*      writing a message to the user when the program aborts.  no system       *
*      globals or files are changed.                                           *
*                                                                              *
*      Parameter:   long  location    unique abort location to be printed.     *
*******************************************************************************/

void imppoint_abort(location)
	long location;
{	
	char string[256];

/************ close file fd if open ****************************/
	if(fd > 0)
		closef_ptr(&fd);

/************* reset job_stats variables ************************/
	job_stats[16] = last_js16;
	job_stats[17] = last_js17;
	job_stats[18] = last_js18;

/************* print abort message ******************************/
	sprintf(string,"impgef_pnt() aborting--> %d", location);
	append_msg(string);

/************* return *******************************************/
	return;
}
