/*******************************************************************************
*       imprlc          (see manual, Command Ref., User Functions)      <fc>   *
*            This program imports a run length encoded raster file.            *
*        The old raster is replaced by the new raster, which may be a          *
*        different size.  All vector type files, (map and import levels)       *
*        are erased as the the new raster forms a new coordinate basis         *
*        for all files.                                                        *
*                                                                              *
*        The rlc  format:                                                      *
*        1.  All entries are binary integers, 2 bytes each, Hi order byte      *
*            first.                                                            *
*                                                                              *
*        2.  The header is 2 entries (4 bytes):                                *
*                 1)  length of file (number of scan lines)                    *
*                 2)  width of longest line (-1 e.g. 192 means 0-192);         *
*        3.  After header come groups of entries, 1 scan line per group        *
*                 1)  number of 'on' runs in scan line (followed by pairs      *
*                     of entries per 'on' run).                                *
*                 2)  pairs of entries indicating the pixel (x) location of    *
*                     the beginning of the run, and the pixel (x) location     *
*                     the end of the run.                                      *
*        4. If the first entry for a scan line group is 0, then the line       *
*           has no 'on' pixels (runs), and is immediately followed by the      *
*           group for the next run.                                            *
*                                                                              *
*        Notes:  the file is binary, not text editable, no control characters  *
*                are allowed in the file unless they represent scan data,      *
*                there may be any number of scan rows of any 1 length in the   *
*                                                                              *
*            Parameters:   char *fname       the name of the rlc  file to      *
*                                            import.                           *
*                          long r            reduction factor (1-6) (linear)   *
*                          long t            treshold 1-rsquared               *
*                                                                              *
*            files/globals changed:  (files) job_stats, index_array,           *
*                 index_list, level_names, v_attributes, v_objects,            *
*                 v_coordlist, raster, mini-raster, attributes, errors,        *
*                 register                                                     *
*                                                                              *
*            Values returned:  (long) 0 = successful import                    *
*                                     1 = failure                              *
*                                                                              *
*            J.Dabritz   6/5/90                                                *
*******************************************************************************/

#include "../../incl/globals.h"
#include "../../incl/in.h"
#include <string.h>


long imprlc( fname,r,t )
   char *fname;
   long r,t;
{
   char actual[256];
   unsigned char string[256];
   char *mallocptr; 

   int fd;

   long i,h,n,w;
   long width;
   long height;
   long size;
   long x,y,x0,y0,xmax;
   long wrd,bit;
   long raster_size;
   long bytes;
   long low, high;
   long scan_height;
   long scan_width;
   long *density;
   long *lptr;
   long len;
   long x1,x2;


	fd = -1;

      /************* check for map present ***************************/
   if(!job_stats[0])
   {   message("You must ACQUIRE or CREATE a map befor IMPORTing a file.");
      return(1);
   }

      /*** check parameters **/
   if((r < 1) || (r > 10))
   {   message("The reduction factor must be within the range 1-10.");
      return(1);
   }

   if((t < 1) || (t > r*r))
   {   message("The threshold must be within the range 1-r squared.");
      return(1);
   }

   message("import rlc, setting up variables.");
      /************* find & asgn pointer to import file **************/
   if(get_impfile(fname,actual))
   {   message("import rlc, abort, location 1.");
      return(1);
   }
   fd = openf(actual,READ);
   if(fd <= 0)
   {   message("Cannot open input file. aborting.");
      return(1);
   }


		/**** set display coordinate type to 0 (raster pixels) ******/
	disp_coord_type = 0;


      /*************** free current map malloc space *****************/
   message("Updating temporary map file.");
   update_temp();
   free_malloc();


      /************* read the header for raster size *********************/
   bytes = readb(string,4,1,fd);
   if(bytes < 4)
   {   closef_ptr(&fd);
      message("Not Enough data in file for header.  Aborting.");
      return(1);
   }

      /** find scan ht.     **/
   low = string[1];
   high = string[0];
   scan_height = (high << 8) | low;

      /** find scan width **/
   low = string[3];
   high = string[2];
   scan_width = (high << 8) | low;


      /************* calculate map height & width ************************/
   height = scan_height / r;
   if(scan_height % r)
      height++;

   width = scan_width / r;
   if(scan_width % r)
      width++;


      /****** make width an integral multiple of 32 **/
   if(width % 32)
   {   n = width / 32 + 1;
      width = n * 32;
   }

   sprintf(string,"scan h/w: %d %d,  reduct: %d,  rstr h/w:  %d %d",
         scan_height,scan_width,r,height, width);
   message(string);


   message("resetting map vairables & allocating space for new raster.");
      /************* reset job_stats *********************************/
   job_stats[1] = width - 1;
   job_stats[2] = height - 1;
   for(i=3; i<=7; i++)
      job_stats[i] = 0;
   for(i=10; i<=18; i++)
   {   if((i==13)  ||  (i==15)  ||  (i==16))
         job_stats[i] = (job_stats[i])  ?    1 : 0;
      else
         job_stats[i] = 0;
   }
   job_stats[45]++;
   job_stats[26] = job_stats[45];
   job_stats[49] = 0;
   job_stats[75] = 0;


   message("calculating new raster size.");
      /***********  calculate sizes **************/
   h = job_stats[2] + 1;             /* height of raster in rows  */
   w = (job_stats[1] + 1) / 8;       /* width of raster in bytes  */
   raster_size = h*w;                /* size of raster in bytes   */
   size = raster_size + 4*w + 8*h;   /* raster+4lines + space for lraster/r */

      /***********  allocate memory **************/
   rasterm.malloc = (char *) calloc_ltp(size,1);
   if(rasterm.malloc == (char *) 0)
   {   message("  <<  cannot get enough RAM for raster.  You must re-ACQUIRE map & start over.");
      job_stats[0] = 0;
      map_name[0] = 0;
      return(1);

   }
      /*  set the raster_wincrement and raster_bincrement globals  */
   raster_wincrement = (job_stats[1] + 1) >> 5;
   raster_bincrement = (job_stats[1] + 1) >> 3;


   message("Setting pointers for RAM raser array.");
      /********  calculate & set global pointers to raster ****************/
   rasterm.rec0 = (char *)(rasterm.malloc + 2*w + 8*h);
   rasterm.recn = rasterm.rec0 + raster_size;
   craster = (unsigned char **)(rasterm.malloc);
   lraster = (unsigned long **)(rasterm.malloc + 4*h);
   for(i=0; i<h; i++)
   {   craster[i] = (unsigned char *)rasterm.rec0 + w*i;
      lraster[i] = (unsigned long *) (craster[i]);
   }

      /**** get space for density array ****/
	if(r > 1)
   {	size = (job_stats[1] + 1) * 4;
   	mallocptr = (char *) calloc_ltp(1,size);
   	if(mallocptr == (char *) 0)
   	{   message("  <<  cannot get enough RAM for import.  You must re-ACQUIRE map & start over.");
      	job_stats[0] = 0;
      	map_name[0] = 0;
      	return(1);
	
   	}
   	density = (long *) mallocptr;


   	message("Setting loop vairables.");
      	/************* read scan file *********************************/
   	for(i=0; i<=job_stats[1]; i++)
      	density[i] = 0;
	}
   xmax = job_stats[1];

      /*** read all scan rows ***/
   y0=job_stats[2];
	if(r == 1)
   {	for(y=1;(y<=scan_height) && (readb(string,2,1,fd) == 2); y++)
   	{   
         	/*** decode len, number of runs in row ****/
      	len = string[1] | (string[0] << 8);
	
         	/*** read runs ***/
      	for(i=0; (i<len) && (readb(string,4,1,fd) == 4); i++)
      	{  x1 = string[1] | (string[0] << 8);
         	x2 = string[3] | (string[2] << 8);
         	if((x1 >= 0) && (x2 <= scan_width))
				{	x2 = job_stats[1] - x2;
					for(x=job_stats[1]-x1; x > x2; x--)
            	{	wrd = x >> 5;              /* find 32 bit word */
               	bit = (7 & ~x) | (x & 24); /* find bit within word*/
               	lraster[y0][wrd] |= (1 << bit);
            	}
         	}
			}
           	/*** decrement y0 ***/
         y0--;
         if(!(y0 % 100))
         {   sprintf(string,"importing rlc raster file.  At row %d.",y0);
           	message(string);
         }
           	/**** check for scan complete (there is other data following) **/
         if(y0 <= 0)
           	break;
   	}
	}
	else
   {	for(y=1;(y<=scan_height) && (readb(string,2,1,fd) == 2); y++)
   	{   
         	/*** decode len, number of runs in row ****/
      	len = string[1] | (string[0] << 8);
	
         	/*** read runs ***/
      	for(i=0; (i<len) && (readb(string,4,1,fd) == 4); i++)
      	{  x1 = string[1] | (string[0] << 8);
         	x2 = string[3] | (string[2] << 8);
         	if((x1 >= 0) && (x2 <= scan_width))
         	{   for(x = x1; x<x2; x++)
            	{   switch(r) {
               	case 2:
                  	x0 = x >> 1;
                  	break;
               	case 4:
                  	x0 = x >> 2;
                  	break;
               	case 8:
                  	x0 = x >> 3;
                  	break;
               	default:
                  	x0 = x / r;
                  	break;
               	}
               	density[x0]++;
            	}
         	}
      	}
	
         	/*** check for r complete ***/
      	if(!(y % r))
      	{
            	/*** set raster row & reset density array ***/
         	for(x=xmax, lptr=density; x>= 0 ; x--, lptr++)
         	{   if(*lptr >= t)
            	{
               	wrd = x >> 5;              /* find 32 bit word */
               	bit = (7 & ~x) | (x & 24); /* find bit within word*/
               	lraster[y0][wrd] |= (1 << bit);
            	}
            	*lptr = 0;
         	}
            	/*** decrement y0 ***/
         	y0--;
         	if(!(y0 % 100))
         	{   sprintf(string,"importing rlc raster file.  At row %d.",y0);
            	message(string);
         	}
            	/**** check for scan complete (there is other data following) **/
         	if(y0 <= 0)
            	break;
      	}
   	}
	}
   message("closing files and updating raster file.");

      /************* close file  *************************************/
   closef_ptr(&fd);

		/************* remove temp file, if any ************************/
	imp_rm_tmp(fname,actual);

   
      /*** free malloc space ***/
	if(r > 1)
   	free_ltp(mallocptr);

      /*******   write raster to temp file  *****/
      /*  open file   */
   sprintf(string,"%s/raster",temp_map_path);
   if((fd = openf(string,WRITE)) <= 0)
   {   message("Cannot open raster temp file.  Aborting.");
      return(1);
   }
      /*  write all records  */
   size = (job_stats[1]+1)  *  (job_stats[2]+1)  /  8;
   if(0 > writeb(rasterm.rec0, size , 1, fd))
   {   message("Cannot write all raster recs to temp file.  Aborting.");
      closef_ptr(&fd);
      return(1);
   }
      /*  close file  */
   closef_ptr(&fd);
/*  free malloc space */
free_ltp(rasterm.malloc);
rasterm.malloc = (char *) 0;


      /* store new sized index_array file     */
      /***********  calculate sizes **************/
   h = job_stats[2] / 32 + 1;        /* height of index_array (rows) */
   w = (job_stats[1] / 32 + 1) * 4;  /* width of index_array (bytes)*/
   size = (h*w);                     /* size to calloc            */
      /***********  allocate memory **************/
   index_arraym.malloc = (char *) calloc_ltp(1,size);
   if(index_arraym.malloc == (char *) 0)
   {   message("  <<  cannot get enough RAM for index_array.  Re-ACQUIRE map & start over.");
      job_stats[0] = 0;
      map_name[0] = 0;
      return(1);
   }

      /***********  write index_array to file *********************/
   sprintf(string,"%s/index_array",temp_map_path);
   fd = openf(string,WRITE);
   if(0 > fd)
   {   message("  <<  cannot open index_array map file.");
      return(0);
   }
   if(0 > writeb(index_arraym.malloc,size,1,fd))
   {   message("  <<  cannot write index_array map file.");
      closef_ptr(&fd);
      return(0);
   }
   closef_ptr(&fd);
/*  free malloc space */
free_ltp(index_arraym.malloc);
index_arraym.malloc = (char *) 0;

      /***********  write job_stats to file ***********************/
   sprintf(string,"%s/job_stats",temp_map_path);
   if(put_jobstats(string))
      {   message("Cannot write job_stats to temp file.  Aborting.");
         return(1);
      }

      /************* re-acquire temp file ****************************/
   if(reacquire())
   {   message("You must re-acquire the map & start over.");
      job_stats[0] = 0;
      return(1);
   }
   view_ltp(0,0,0);

      /************* un-register raster ******************************/
   for(i=0; i<4; i++)
   {   corner[0][i][0] = 0.0;
      corner[0][i][1] = 0.0;
   }

      /************* write diary record ******************************/
   sprintf(string,"imported rlc file %s  (%d x %d)",fname,job_stats[2]+1,job_stats[1]+1);
   put_diary(string);


      /*** reset undo array    **********/
   for(i=0; i<32; i++)
      undo_array[i].mode = 0;

		/************* create rasterb **********************************/
	brstr_create();

   return(0);
}
