/*******************************************************************************
*       imprle1         (see manual, Command Ref., User Functions)      <fc>   *
*            This program imports a run length encoded raster file.            *
*        The old raster is replaced by the new raster, which may be a          **        different size.  All vector type files, (map and import levels)       *
*        are erased as the the new raster forms a new coordinate basis         *
*        for all files.                                                        *
*                                                                              *
*        The rle1 format:                                                      *
*        1.  each continuous stretch (run) of 'on' or 'off' pixels on the      *
*            raster are encoded as 1 or more integers.  A single run does      *
*            not wrap around to another raster row.                            *
*                                                                              *
*        2.  Each byte represents a run of 1 or more 'on' or 'off' pixels:     *
*                 bit 7 = high order flag:  1 = multiply integer by 64         *
*                                           0 = use integer without mult.      *
*                 bit 6 = run type flag:  1 = 'on' run,  0 = 'off' run         *
*                 bits 5 - 0 = 6 bit unsigned integer indicating length of     *
*                              the run.  (multiply integer by 64 if bit 7      *
*                              is equal to 1)                                  *
*                                                                              *
*        3. Consecutive bytes may be of same run type ('on' or 'off')          *
*                                                                              *
*        4. The last run in each scan row is followed by 1 (or more)           *
*           0 (null) bytes.  (0 [null] is never used in a legitimate run.)     *
*                                                                              *
*        Notes:  the file is binary, not text editable, no control characters  *
*                are allowed in the file unless they represent scan data,      *
*                there may be any number of scan rows of any 1 length in the   *
*                file, there may any number of nulls befor, between, or        *
*                following an encoded scan row.                                *
*                                                                              *
*            Parameters:   char *fname       the name of the rle1 file to      *
*                                            import.                           *
*                                                                              *
*            files/globals changed:  (files) job_stats, index_array,           *
*                 index_list, level_names, v_attributes, v_objects,            *
*                 v_coordlist, raster, mini-raster, attributes, errors,        *
*                 register                                                     *
*                                                                              *
*            Values returned:  (long) 0 = successful import                    *
*                                     1 = failure                              *
*                                                                              *
*            J.Dabritz   9/26/89                                               *
*******************************************************************************/

#include "../../incl/globals.h"
#include "../../incl/in.h"
#include <string.h>


long imprle1( fname )
	char *fname;
{
	char actual[256];
	char new_file[256];
	char raster3[256];
	char file[256];
	char nulls[256];
	char path[256];
	char string[256];
	char tmp_buf[256];
	char *malloc_ptr;
	char *end_ptr;
	char *ptr;
	unsigned char byte;

	int fd, fd1, fd2, filefd;

	long i,j,h,n,w;
	long width;                  /* width of 1 raster row (pixels) */
	long size;                   /* bytes in 1 raster row */
	long *rle1_row;              /* pointer to malloc space */
	long rows;                   /* number of raster rows */
	long rle1_words;             /* number of 32 bit words in a raster row */
	long *rle1_row_end;          /* ptr just beyond rle1_row */
	long *lptr;
	long max;
	long done;
	long bit_count;
	long wrd,bit,x,dx;
	long offset;
	long iend;
	long sizel;
	long raster_size;


	fd = -1;
	fd1 = -1;
	fd2 = -1;
	filefd = -1;

		/************* check for map present ***************************/
	if(!job_stats[0])
	{	message("You must ACQUIRE or CREATE a map befor IMPORTing a file.");
		return(1);
	}

	message("import rle1, setting up variables.");
		/************* find & asgn pointer to import file **************/
	if(get_impfile(fname,actual))
	{	message("import rle1, abort, location 1.");
		return(1);
	}
	fd1 = openf(actual,READ);
	if(fd1 <= 0)
	{	message("Cannot open input file. aborting.");
		return(1);
	}


		/**** set display coordinate type to 0 (raster pixels) ******/
	disp_coord_type = 0;


		/************* read top raster line & get raster width *********/
	width = 0;
	for( ;; )
	{	if(1 > readb(&byte,1,1,fd1))               /** read 1 byte ***/
			break;
		if(width && !byte)                   /** check for ending null **/
			break;
		if(byte & 128)       /** multiply integer by 64 **/
			width += ((long) byte & 63) << 6;  /** multiply by 64 (shift 6) */
		else
			width += (byte & 63);              /** don't multiply        **/
	}

		/****** close input file. ***/
	closef_ptr(&fd1);

		/****** make width an integral multiple of 32 **/
	if(width % 32)
	{	n = width / 32 + 1;
		width = n * 32;
	}

		/************* create new raster file & asgn pointer ***********/
	sprintf(new_file,"%s/raster3",temp_map_path);
	purgef(new_file);
	sprintf(new_file,"%s/raster2",temp_map_path);
	purgef(new_file);
	if(-1 == creatf(new_file))
	{	message("Cannot create new raster file. aborting.");
		return(1);
	}
	fd2 = openf(new_file,WRITE);
	if(fd2 < 0)
	{	message("Cannot open new raster file. aborting.");
		return(1);
	}
	
		/************** re-open input (rle1) file **********************/
	fd1 = openf(actual,READ);
	if(fd1 <= 0)
	{	message("Cannot open input file. aborting.");
		closef_ptr(&fd2);
		return(1);
	}


		/************* malloc enough space for a raster row ************/
	size = width / 8;
	malloc_ptr = malloc_ltp(size);
	if(malloc_ptr == (char *) 0)
	{	message("Not enough RAM available to import rle1.  Aborting.");
		closef_ptr(&fd1);
		closef_ptr(&fd2);
		return(1);
	}
	rle1_row = (long *) malloc_ptr;
	rle1_words = size / 4 ;
	rle1_row_end = (long *) (rle1_row + rle1_words);
	

		/************* read rle1 file one raster line at a time ********/
	for(rows = 0, done = 0; !done ; rows++)
	{		/*********inform user **************/
		if(!(rows % 25))
		{	sprintf(string,"imprle1, reading in raster row %d",rows);
			message(string);
		}

			/************ zero out storage row *************************/
		for(lptr = rle1_row; lptr < rle1_row_end; lptr++)
			*lptr = 0;

			/************ read & decode 1 raster row ****************/
		x = 0;
		for( ;; )
		{		/******** read 1 byte ***************/
			if(1 > readb(&byte,1,1,fd1))       /** read 1 byte ***/
			{	done = 1;
				break;
			}
			  /********* check for null which ends the scan line *******/
			if(x && !byte)                     /** check for ending null **/
				break;

			if(byte & 128)       /** multiply integer by 64 **/
				dx = ((long) byte & 63) << 6;  /** multiply by 64 (shift 6) */
			else
				dx = (byte & 63);              /** don't multiply        **/

				/****** if pixel flag is 'on' then set pixels to 'on' ***/
			if(byte & 64)
			{	for( ; dx && (x < width) ; x++, dx--)
				{	wrd = x >> 5;                  /* find 32 bit word */
					bit = (7 & ~x) | (x & 24);     /* find bit within word*/
					rle1_row[wrd] |= (1 << bit);   /* turn on rle1 pixel */
				}
			}
			else
				x += dx;
		}

			/************ store raster row in temp file ****************/
		if(x > 0)
		{	if(writeb(malloc_ptr,size,1,fd2) <= 0)
			{	message("write failed when storing a row of raster data.  Aborting.");
				closef_ptr(&fd1);
				closef_ptr(&fd2);
				free_ltp(malloc_ptr);
				return(1);
			}
		}
	}

		/************* close files *************************************/
	closef_ptr(&fd1);
	closef_ptr(&fd2);


		/************* remove temp file, if any ************************/
	imp_rm_tmp(fname,actual);


		/************* create final raster file ************************/
	message("imprle1, re-reflecting new raster & storing in raster file.");
	sprintf(raster3,"%s/raster3",temp_map_path);
	if(-1 == creatf(raster3))
	{	message("Cannot create raster3 raster file. aborting.");
		free_ltp(malloc_ptr);
		return(1);
	}
	fd2 = openf(raster3,WRITE);
	if(fd2 < 0)
	{	message("Cannot open raster3 raster file. aborting.");
		free_ltp(malloc_ptr);
		return(1);
	}
	
		/************* open new_file (upside down raster) **************/
	fd1 = openf(new_file,READ);
	if(fd1 < 0)
	{	message("Cannot open new_file raster file. aborting.");
		free_ltp(malloc_ptr);
		closef_ptr(&fd2);
		return(1);
	}
	
		/************* read upside down raster & store right side up ***/
	for(offset = (rows - 1) * size; offset >= 0; offset -= size)
	{		/******************** position file read pointer ***************/
		if(0 > movefp(fd1,offset,0))
		{	free_ltp(malloc_ptr);
			closef_ptr(&fd1);
			closef_ptr(&fd2);
			message("Move file pointer failed while reflecting rle1 file.  Aborting.");
			return(1);
		}

			/******************** read  record from file 1 ******************/
		if(0 > readb(malloc_ptr, size, 1, fd1))
		{	free_ltp(malloc_ptr);
			closef_ptr(&fd1);
			closef_ptr(&fd2);
			message("Read failed while reflecting rle1 file.  Aborting.");
			return(1);
		}

			/******************** store record in file 2 ********************/
		if(0 > writeb(malloc_ptr, size, 1, fd2))
		{	free_ltp(malloc_ptr);
			closef_ptr(&fd1);
			closef_ptr(&fd2);
			message("Write failed while reflecting rle1 file.  Aborting.");
			return(1);
		}
	}

		/************* close files & free malloc space *****************/
	free_ltp(malloc_ptr);
	closef_ptr(&fd1);
	closef_ptr(&fd2);
	purgef(new_file);

		/************* move raster file ********************************/
	sprintf(new_file,"%s/raster",temp_map_path);
	if(move_file(raster3, new_file))
	{	message("Move failed when attempting to move new raster into old file, Aborting.");
		return(1);
	}

	message("resetting map vairables & loading new raster into display RAM.");
		/************* reset job_stats *********************************/
	job_stats[1] = size * 8 - 1;
	job_stats[2] = rows - 1;
	for(i=3; i<=7; i++)
		job_stats[i] = 0;
	for(i=10; i<=18; i++)
	{	if((i==13)  ||  (i==15)  ||  (i==16))
			job_stats[i] = (job_stats[i])  ?    1 : 0;
		else
			job_stats[i] = 0;
	}
	job_stats[45]++;                   /** rle1 import flag **/
	job_stats[21] = job_stats[45];
	job_stats[49] = 0;
	job_stats[75] = 0;


		/********** free old space *************************************/
	free_ltp(rasterm.malloc);
	rasterm.malloc = (char *) 0;
		/***********  calculate sizes **************/
	h = job_stats[2] + 1;             /* height of raster in rows  */
	w = (job_stats[1] + 1) / 8;       /* width of raster in bytes  */
	raster_size = h*w;                /* size of raster in bytes   */
	sizel = raster_size / 4;          /* size of raster in longs (4 bytes ea.)*/
	size = raster_size + 4*w + 8*h;   /* raster+4lines + space for lraster/r */
		/***********  allocate memory **************/
	rasterm.malloc = (char *) calloc_ltp(size,1);
	if(!rasterm.malloc)
	{	message("  <<  cannot get enough RAM for raster.  You must re-ACQUIRE map & start over.");
		job_stats[0] = 0;
		map_name[0] = 0;
		return(1);
	}
		/********  calculate & set global pointers to raster ****************/

	rasterm.rec0 = (char *)(rasterm.malloc + 2*w + 8*h);
	rasterm.recn = rasterm.rec0 + raster_size;
	craster = (unsigned char **)(rasterm.malloc);
	lraster = (unsigned long **) (rasterm.malloc + 4*h);
	for(i=0; i<h; i++)
	{	craster[i] = (unsigned char *)rasterm.rec0 + w*i;
		lraster[i] = (unsigned long *) (craster[i]);
	}

		/***********  read raster from file *********************/
	sprintf(path,"%s/raster",temp_map_path);
	filefd = openf(path,READ);
	if(0 > filefd)
	{	message("  <<  cannot open raster map file.  You must re-ACQUIRE map & start over.");
		job_stats[0] = 0;
		map_name[0] = 0;
		return(1);
	}
	if(0 >= readb(rasterm.rec0,raster_size,1,filefd))
	{	message("  <<  cannot read raster map file.  You must re-ACQUIRE map & start over.");
		closef_ptr(&filefd);
		job_stats[0] = 0;
		map_name[0] = 0;
		return(1);
	}


		/* create index_arary in ram            */
		/* free existing array space */
	free_ltp(index_arraym.malloc);
	index_arraym.malloc = (char *) 0;
		/***********  calculate sizes **************/
	h = job_stats[2] / 32 + 1;        /* height of index_array (rows) */
	w = (job_stats[1] / 32 + 1) * 4;  /* width of index_array (bytes)*/
	size = (h*w + 4*h);               /* size to malloc            */
		/***********  allocate memory **************/
	index_arraym.malloc = (char *) calloc_ltp(size,1);
	if(!index_arraym.malloc)
	{	message("  <<  cannot get enough RAM for index_array.  Re-ACQUIRE map & start over.");
		job_stats[0] = 0;
		map_name[0] = 0;
		return(1);
	}
		/********  calculate & set global pointers to index_array ***********/
	index_arraym.rec0 = index_arraym.malloc;
	index_arraym.recn = (char *)(index_arraym.rec0 + h*w);
	index_array = (long **)index_arraym.recn;
	for(i=0; i<h; i++)
		index_array[i] = (long *) (index_arraym.malloc + w*i);

		/************* un-register raster ******************************/
	for(i=0; i<4; i++)
	{	corner[0][i][0] = 0.0;
		corner[0][i][1] = 0.0;
	}

		/************* write diary record ******************************/
	sprintf(string,"imported rle1 file %s  (%d x %d)",fname,job_stats[2]+1,job_stats[1]+1);
	put_diary(string);

		/************* redisplay map ***********************************/
		/*  set the raster_wincrement and raster_bincrement globals  */
	raster_wincrement = (job_stats[1] + 1) >> 5;
	raster_bincrement = (job_stats[1] + 1) >> 3;
		/** recompute view[] global ******/
	for(i=0; i<4; i++)
	{	view[i].lower_leftx = 0;
		view[i].lower_lefty = 0;
	}
	view[0].zoom_factor = mind(2,  (double) 1.0, 
			(double) (400.0/(maxl(2,job_stats[1]+1,job_stats[2]+1))));
	view[1].zoom_factor = 0.25;
	view[2].zoom_factor = 1.0;
	view[3].zoom_factor = 8.0;


	message("building new map display.");
	redisp_map();

		/*** reset undo array    **********/
	for(i=0; i<32; i++)
		undo_array[i].mode = 0;


		/************* create rasterb **********************************/
	brstr_create();

		/************* return ******************************************/
	return(0);
}
