/*******************************************************************************
*      mask            (see manual, Command Ref., User Functions)       <fc>   *
*          Description:  Overlays objects within a specified set               *
*          with a solid mask of a specified radius.  Used with                 *
*          the design of corridors and buffers.  Assembly is required          *
*                                                                              *
*          Parameters:  long s;      set number of objects to be masked        *
*                       long u;      length units code for radius:             *
*                         0 = pixels                                           *
*                         1 = map inches                                       *
*                         2 = ground feet                                      *
*                         3 = ground miles                                     *
*                         4 = map centimeters                                  *
*                         5 = ground meters                                    *
*                         6 = ground kilometers                                *
*                       double r;    radius                                    *
*                                                                              *
*          Files/globals changed:  raster file, job_stats                      *
*                                                                              *
*          Returns zero upon normal end-of-job (EOJ), 1 on failure             *
*          J.Dabritz  2/16/90                                                  *
*******************************************************************************/

#include "../../incl/globals.h"
#include "../../incl/in.h"
#include <math.h>

long mask_object();
long line_circles();
long draw_disk();
long draw_circle();
long pix_row();


long *disk;   /** pointer to disk array; **/
long rd;      /** size of disk array; **/
long horizontal;      /*  flag, 0=vertical, non-zero = horizontal  */


long mask(s,u,r)
	long s,u;
	double r;
{
	OBJECTS object;

	char *ptr;

	long size;
	long i;

	double radius;
	double w;


		/****************** check for map ********************************/
	if(!job_stats[0])
	{	message("You must Create or Acquire a map befor masking a set of objects.");
		return(1);
	}

		/************* check for VIEW ONLY raster **********************/
	if(job_stats[78])
	{	message("You CANNOT do ANYTHING to the VIEW ONLY ORIGINAL SCAN.");
		return(1);
	}
		/****************** check for objects ****************************/
	if(!job_stats[10])
	{	message("You must Assemble the map befor masking a set of objects.");
		return(1);
	}

		/****************** check u & r  *********************************/
	if((u < 0) || (u > 6))
	{	message("Illegal units code.  Units code must be in range 0-6.");
		return(1);
	}
	get_units();
	if(unit_factor[u][0] == 0.0)
	{	message("You must register raster/geodetics before choosing non-pixel units.");

		return(1);
	}
	if((r<= 0.0))
	{	message("Illegal mask radius.  Radius must be greater than 0.");
		return(1);
	}


		/****************** calculate r in pixels ************************/
	radius = r / unit_factor[u][0];
	if(radius < 0.5)
	{	message("Computed radius is < 0.5 pixels.  Mask adds nothing to raster.");
		return(1);
	}


		/****************** calculate disk array *************************/
	rd = radius;         /* whole number */
	rd++;
	size = 4 * rd;
	ptr = malloc_ltp(size);
	if(ptr == (char *) 0)
	{	message("Cannot malloc enough RAM to generate disk array.  Aborting comand.");
		free_ltp(ptr);
		return(1);
	}
	disk = (long *) ptr;
	for(i=0; i<rd; i++)
	{	w = sqrt((radius * radius) - (i * i));
		disk[i] = w;
	}

		/****************** check all objects, drawing masks as needed ***/
	message("masking objects . . .  ");
	for(i=0; i<job_stats[10]; i++)
	{	if(get_objects(&object,i))
		{	message("mask abort, location 0.");
			free_ltp(ptr);
			return(1);
		}
		if(object.i12[4] & (1 << s))
		{
			if(mask_object(i))
			{	free_ltp(ptr);
				return(1);
			}
		}
		if(abort_cmd())
			break;
	}
	append_msg("done.  ");

		/****************** free malloc'd space **************************/
	free_ltp(ptr);

		/****************** update display *******************************/
	redisp_map();
	view_ltp(0,0,0);

		/****************** set edited flag ******************************/
	if((job_stats[40] < job_stats[45]) || !job_stats[45])
	{	job_stats[45]++;
		job_stats[40] = job_stats[45];
	}
	
		/****************** return ***************************************/
	return(0);
}


/*******************************************************************************
*   mask_object                                                                *
*       This routine draws a mask around a single map object.                  *
*                                                                              *
*       Parameters:  long obj;     object record number                        *
*                                                                              *
*       Values changed:  raster, job_stats                                     *
*       J.Dabritz   2/16/90                                                    *
*******************************************************************************/

long mask_object(obj)
long obj;
{
	OBJECTS object;
	COORD_LIST coord_list;

	long x,y;
	long rec, s;
	long v;
	long last_x, last_y;


	if(abort_cmd())
		return(0);

		/************ recall object ***************************************/
	if(get_objects(&object,obj))
	{	message("mask_object abort, location 0.");
		return(1);
	}


		/************ if node, draw solid disk ****************************/
	if(object.i12[0] < 8)       /** node **/
	{	x = object.f4[0] + 0.5;
		y = object.f4[1] + 0.5;
		draw_disk(x,y);
	}
		/************ if polygon, find segments & call recursively ********/
	else if(object.i12[0] >= 14)
	{		/** first, mask all segments **/
		for(rec=object.i12[5]; rec<=object.i12[6]; rec++)
		{	if(get_seglist(&s,rec))
			{	message("mask_object abort, location 1.");
				return(1);
			}
			s = abs(s);
			if(mask_object(s))
				return(1);
			if(abort_cmd())
				break;
		}
			/** finally, mask next island polygon **/
		if(object.i12[8])
		{	if(mask_object(object.i12[8]))
				return(1);
		}
	}
		/************ if segment, draw starting disk & then circles *******/
	else
	{		/*** get first vertex **********************************/
		if(get_coordlst(&coord_list,object.i12[5]))
		{	message("mask_object abort, location 2.");
			return(1);
		}
			/*** draw disk at first vertex *************************/
		x = 0.5 + coord_list.x;
		y = 0.5 + coord_list.y;
		draw_disk(x,y);
		last_x = x;
		last_y = y;

			/*** draw circles between subsequent verticies *********/
		for(v = object.i12[5] + 1; v<=object.i12[6]; v++)
		{	if(get_coordlst(&coord_list,v))
			{	message("mask_object abort, location 3.");
				return(1);
			}
			x = 0.5 + coord_list.x;
			y = 0.5 + coord_list.y;
			line_circles(x,y,last_x,last_y);
			last_x = x;
			last_y = y;
		}
	}

		/************ return **********************************************/
	return(0);
}


/*******************************************************************************
*       line_circles                                                           *
*                                                                              *
*    Purpose:    Draw circles about each point of a line on the raster & screen*
*       between specified points.  Uses a diagonal jump between rows.          *
*       See line_solid for square jumps between rows.                          *
*                                                                              *
*    Arguments:    long x1,y1,x2,y2  line segment endpoints, (raster coords)   *
*                                                                              *
*    Returned Values: (long) 0                                                 *
*                                                                              *
*    Global Variables Modified:  none                                          *
*                                                                              *
*                      j.dabritz  2/16/90                                      *
*******************************************************************************/


long line_circles(x1,y1,x2,y2)
	long x1,y1,x2,y2;
{	
	long dx,dy;
	long e,e1,e2;
	long i;

		/*********   clip line to raster ***************************/
	if(clipl(&x1,&y1,&x2,&y2,0,job_stats[1],0,job_stats[2]))
		return(0);

		/*********   calc dx and switch points to make it non-negative.   */
	dx = x2 - x1;
	if(dx < 0) 
	{	i = x1;
		x1 = x2;
		x2 = i;
		i = y1;
		y1 = y2;
		y2 = i;
		dx = x2 - x1;
	}
		/*********   calc cy  */
	dy = y2 - y1;

		/*********   divide situation up into cases ****************/
	if(dy >= 0)              /* line has positive slope */
	{	if(dx > dy)          /* slope is < 1  */
		{	e  = (dy * 2) - dx;
			e1 = (dy - dx) * 2;
			e2 = dy * 2;
			for(i = -1; i != dx; i++, x1++)
			{	draw_circle(x1,y1);
				if(e > 0)
				{	if((i+1) != dx)         /* add square jump to diag. line */
					{	if(e < dy)
							draw_circle(x1+1, y1);
						else
							draw_circle(x1, y1+1);
					}
					y1++;
					e += e1;
				}
				else
					e += e2;
			}
		}
		else                  /* slope is > 1    */
		{	e  = (dx * 2) - dy;
			e1 = (dx - dy) * 2;
			e2 = dx * 2;
			for(i = -1; i != dy; i++, y1++)
			{	draw_circle(x1,y1);
				if(e > 0)
				{	if((i+1) != dy)         /* add square jump to diag. line */
					{	if(e < dx)
							draw_circle(x1, y1+1);
						else
							draw_circle(x1+1, y1);
					}
					x1++;
					e += e1;
				}
				else
					e += e2;
			}
		}
	}
	else                     /* slope is negative  */
	{	if(dx > -dy)          /* slope is > -1  */
		{	e  = ((-dy) * 2) - dx;
			e1 = ((-dy) - dx) * 2;
			e2 = (-dy) * 2;
			for(i = -1; i != dx; i++, x1++)
			{	draw_circle(x1,y1);
				if(e > 0)
				{	if((i+1) != dx)         /* add square jump to diag. line */
					{	if(e < -dy)
							draw_circle(x1+1, y1);
						else
							draw_circle(x1, y1-1);
					}
					y1--;
					e += e1;
				}
				else
					e += e2;
			}
		}
		else                  /* slope is < -1   */
		{	e  = (dx * 2) + dy;
			e1 = (dx + dy) * 2;
			e2 = dx * 2;
			for(i = 1; i != dy; i--, y1--)
			{	draw_circle(x1,y1);
				if(e > 0)
				{	if((i-1) != dy)         /* add square jump to diag. line */
					{	if(e < -dy)
							draw_circle(x1, y1-1);
						else
							draw_circle(x1+1, y1);
					}
					x1++;
					e += e1;
				}
				else
					e += e2;
			}
		}
	}

	return(0);
}


/*******************************************************************************
*     draw_disk                                                                *
*        This program draws a disk   around a specified point.                 *
*                                                                              *
*        Parameters  long x1,y1;   center of disk                              *
*                                                                              *
*        Values returned: 0                                                    *
*        J.Dabritz  2/16/90                                                    *
*******************************************************************************/


long draw_disk(x1,y1)    /** ,disk,r,odd); **/
long x1,y1;
{
	long i;
	long dx;


	if(abort_cmd())
		return(0);

	for(i=rd-1;i >= 0; i--)
	{	dx = disk[i];
		pix_row(x1 - dx, y1 + i, x1 + dx, y1 + i);
		if(i)
			pix_row(x1 - dx, y1 - i, x1 + dx, y1 - i);
	}

	return(0);
}



/*******************************************************************************
*     draw_circle;                                                             *
*        This program draws a circle around a specified point.                 *
*                                                                              *
*        Parameters  long x1,y1;   center of circle                            *
*                                                                              *
*        Values returned: 0                                                    *
*        J.Dabritz  2/16/90                                                    *
*******************************************************************************/


long draw_circle(x1,y1)    /** ,disk,r,odd); **/
long x1,y1;
{
	long i;
	long dx,dx2;

	if(abort_cmd())
		return(0);


	dx2 = 0;
	for(i=rd-1;i >= (dx2); i--)
	{	dx = disk[i];
		if(i == rd - 1)   /** first edge **/
		{	pix_row(x1 - dx, y1 + i, x1 + dx, y1 + i);
			pix_row(x1 - dx, y1 - i, x1 + dx, y1 - i);
			pix_row(x1 + i, y1 - dx, x1 + i, y1 + dx);
			pix_row(x1 - i, y1 - dx, x1 - i, y1 + dx);
		}
		else
		{	pix_row(x1-dx2,y1+i, x1-dx,y1+i);
			pix_row(x1+dx2,y1+i, x1+dx,y1+i);
			pix_row(x1-dx2,y1-i, x1-dx,y1-i);
			pix_row(x1+dx2,y1-i, x1+dx,y1-i);
			pix_row(x1-i,y1-dx2, x1-i,y1-dx);
			pix_row(x1-i,y1+dx2, x1-i,y1+dx);
			pix_row(x1+i,y1-dx2, x1+i,y1-dx);
			pix_row(x1+i,y1+dx2, x1+i,y1+dx);
		}
		dx2 = dx + 1;
	}

	return(0);
}

/*******************************************************************************
*   pix_row                                                                    *
*         turns on a row or column of pixels on the raster, no screen display. *
*                                                                              *
*         parameters: long x1,y1;  x,y coords of one end of row/column.        *
*                     long x2,y2;  x,y coords of other end of row/column.      *
*                                                                              *
*         returns 0, 1=error.                                                  *
*         J.Dabritz 7/11/90                                                    *
*******************************************************************************/

long pix_row(x1,y1,x2,y2)
long x1,y1,x2,y2;
{
	long key;
	long i;
	long word, bit;

		/************ find out if row, column, or pixel ***************/
	key = 0;        /* 1=row, 2=col, 3=pixel */
	if(y1 == y2)
		key |= 1;
	if(x1 == x2)
		key |= 2;

		/************ draw row/column/pixel ***************************/
	switch(key) {
	case 1:        /* row    */
			/*** check if row on raser ***/
		if((y1 < 0) || (y1 > job_stats[2]))
			return(0);
			/**** draw all cols ***/
		if(x1 < x2)
			i = x1;
		else
		{	i = x2;
			x2 = x1;
		}
			/**** check beginning/ending col **/
		if((i > job_stats[1]) || (x2 < 0))
			return(0);
		if(i < 0)
			i = 0;
		if(x2 > job_stats[1])
			x2 = job_stats[1];
		for(;i<=x2;i++)
		{	word = i >> 5;
			bit = (7 & ~i) | (i & 24);
			lraster[y1][word] |= (1 << bit);
		}	
		break;
	case 2:        /* column */
			/*** check if col on raser ***/
		if((x1 < 0) || (x1 > job_stats[1]))
			return(0);
			/**** draw all rows ***/
		if(y1 < y2)
			i = y1;
		else
		{	i = y2;
			y2 = y1;
		}
			/**** check beginning/ending row **/
		if((i > job_stats[2]) || (y2 < 0))
			return(0);
		if(i < 0)
			i = 0;
		if(y2 > job_stats[2])
			y2 = job_stats[2];
		word = x1 >> 5;
		bit = (7 & ~x1) | (x1 & 24);
		for(;i<=y2;i++)
			lraster[i][word] |= (1 << bit);
		break;
	case 3:        /* pixel  */
			/*** check if pixel on raster ***/
		if((x1 < 0) || (x1 > job_stats[1]) || (y1 < 0) || (y1 > job_stats[2]))
			return(0);
		word = x1 >> 5;
		bit = (7 & ~x1) | (x1 & 24);
		lraster[y1][word] |= (1 << bit);
		break;
	}
	
		/************ return ******************************************/
	return(0);
}
