/*******************************************************************************
*    report_ltp    (see manual, Command Ref., User Functions)    <fc>          *
*                                                                              *
*         Generates a report to the designated printer based on the report     *
*         parameters received.                                                 *
*         Parameters:                                                          *
*           outdev   Display device type.  Range 0-3                           *
*           len      Length type .  Range 0-6                                  *
*           area     Area type.  Range 0-6                                     *
*           r1       Primary sort type.  Range 0-6                             *
*           r2       Secondary sort type.  Range 0-6                           *
*           setnum   Set number for selecting obj recs.  0-31                  *
*           fname    Name of file report writes to.                            *
*                                                                              *
*         Returns 0 upon normal execution, else returns -1 upon error.         *
*         C. Jonson                                                            *
*******************************************************************************/

#include "../../incl/globals.h"
#include "../../incl/in.h"
#include <string.h>

long report_obj();
int report_sort();

long sort_type;
long len_glob;
long area_glob;
double cum_len,cum_area;
static char wr_type[][8] = { 
		"node","segment","polygon","node","node","node","node","node",
		"segment","segment","segment","","","","polygon","polygon","polygon" };
static char wr_subtype[][14] = { 
		"","",""," 0-branches"," 1-branch"," 2-branches", " 3-branches", 
		" 4-branches", " 2 free ends"," 1 free end", " 0 free ends", 
		"","",""," interior"," exterior"," map border" }; 



long report_ltp(  outdev, len, area, r1, r2, setnum, fname )
	long outdev, len, area, r1, r2, setnum;
	char *fname;
{
	long  i,j;                               /* for loop index         */
	long  obj_cnt;                           /* cnt of obj recs read   */
	long  no_sort;
	long  *lptr;

	int   tmp_fd;                            /* for tmp report file    */

	char  *mptr;
	char  tmp_msg[STRLIST];                  /* for misc messages      */
	char  outrec[STRLIST];                   /* array for writing recs */
	char  tmp_name[L_tmpnam];                /* used for tmpnam()      */
	OBJECTS obj_rec;                         /* holds obj rec just read*/
	static char len_type[][16] = {
		"PIXELS", "MAP INCHES", "GROUND FT", "GROUND MILES",
		"MAP CM'S", "GROUND METERS", "GROUND KM'S" };
	static char area_type[][16] = {
		"SQ PIXELS", "SQ INCHES (MAP)", "ACRES", "SQ MILES",
		"SQ CM (MAP)", "SQ METERS", "SQ KILOMETERS" };
	void qsort();


		/************* 0. set globals ********************************/
	len_glob = len;
	area_glob = area;
	cum_len = 0.0;
	cum_area = 0.0;
	get_units();

		/************* 1. count objects in report ********************/
	obj_cnt = 0;
	for(i=0; i<job_stats[10]; i++)
	{	if( get_objects( &obj_rec, i ) )
		{	message("REPORT_LTP: Cannot read objects file.  Bye.");
			return(-1);
		}
		if(obj_rec.i12[4] & (1 << setnum))
			obj_cnt++;
	}

	if(!obj_cnt)
	{	message("NO OBJECTS in set.  REPORT NOT WRITTEN.");
		return(0);
	}

		/************* 2. see if space can be malloc'd ***************/
	if(((r1) && (r1 != 2)) || ((r2) && (r2 == 2))) /*** sorting called for ***/
	{	mptr = (char *) calloc_ltp(obj_cnt,4);  /** ask for space (4 bytes/rec) */
		if(mptr == (char *) 0)                  /** didn't get space ***/
			no_sort = 1;
		else                                    /** got space **/
		{	no_sort = 0;
			lptr = (long *) mptr;                /** set a long ptr to space **/
			obj_cnt = 0;
			for(i=0; i<job_stats[10]; i++)       /** write rec nums in space **/
			{	if( get_objects( &obj_rec, i ) )
				{	message("REPORT_LTP: Cannot read objects file.  Bye.");
					return(-1);
				}
				if(obj_rec.i12[4] & (1 << setnum))
				{	lptr[obj_cnt] = i;
					obj_cnt++;
				}
			}
		}
	}	
	else
		no_sort = 1;
	
		/************* 3. sort primary sort variable *****************/
	if((r1) && (r1 != 2)) /*** sorting called for ***/
	{	sort_type = r1;
		qsort(mptr, obj_cnt, 4, report_sort);
	}

		/**** (qsort is unstable, must sort 2ndry second) ************/
		/**** 4. sort by 2ndary sort variable in groups **************/
	if(r2 && (obj_cnt > 1))
	{	for(j=0,i=1; i<obj_cnt; i++)
		{		/**** check for different group ****/
			if(report_sort(&lptr[i],&lptr[j]))     /*** different ***/
			{	/*** see if range is > 1 ****/
				if((i-j) > 1)
				{	 /*** sorting called for ***/
					sort_type = r2;
					qsort(mptr + 4 * j, (long) (i - j), 4, report_sort);
					sort_type = r1;
				}
				j=i;
			}
		}
			/*** check last group ***/
		if((i-j) > 1)
		{	 /*** sorting called for ***/
			sort_type = r2;
			qsort(mptr + 4 * j, (long) (i - j), 4, report_sort);
		}
	}

		/************* 5. print report *******************************/
/***  Create temp name for file.  Use system default temp directory  ***/
	tmpnam( tmp_name );		/* Arg must be L_tmpnam in length */

/***  Create and then open the file for writing  ***/
	if( creatf( tmp_name ) == -1 )
	{	if( !batch_mode )
			message("REPORT_LTP: Cannot create temp file. Bye.");
		return(-1);
	}
	
	if( (tmp_fd = openf( tmp_name, WRITE)) == -1 )
	{	if( !batch_mode )
			message("REPORT_LTP: Cannot open temp file. Bye.");
		return(-1);
	}

		/** print header **/
/***  Write standard header then report headers ***/
	std_header(outrec);
	if( writea(outrec, tmp_fd) == -1 )
	{	if( !batch_mode )
			message("REPORT_LTP: Cannot write to temp file. Bye.");
		closef(tmp_fd);
		return(-1);
	}

/*** Also...write the set number and the level name below the header ***/
	sprintf(outrec, "Set ->%3d  Subjects ->%4d  Level Name --> %-45s\n", 
		setnum,count_subj(setnum), level_names[setnum]);
	if( writea(outrec, tmp_fd) == -1 )
	{	if( !batch_mode )
			message("REPORT_LTP: Cannot write to temp file. Bye.");
		closef(tmp_fd);
		return(-1);
	}

/******** DON'T DIDDLE WITH THE SPACING IN THE FOLLOWING SPRINTF'S!!!  *******/
	sprintf(outrec,"\nTYPE           RECORD   %-13s   %-13s  ATTRIBUTE 1  \n",
			 len_type[len], area_type[area] );
	if( writea(outrec, tmp_fd) == -1 )
	{	if( !batch_mode )
			message("REPORT_LTP: Cannot write to temp file. Bye.");
		closef(tmp_fd);
		return(-1);
	}

	sprintf(outrec, "SUBTYPE        SERIAL    CUMULATIVE   \
   CUMULATIVE      ATTRIBUTE 2   \n");
	if( writea(outrec, tmp_fd) == -1 )
	{	if( !batch_mode )
			message("REPORT_LTP: Cannot write to temp file. Bye.");
		closef(tmp_fd);
		return(-1);
	}

		/** print report lines **/
	if(no_sort)
	{	for(i=0; i<job_stats[10]; i++)
		{	if( get_objects( &obj_rec, i ) )
			{	message("REPORT_LTP: Cannot read objects file.  Bye.");
				return(-1);
			}
			if(obj_rec.i12[4] & (1 << setnum))
				report_obj(i,tmp_fd);
		}
	}
	else
	{	for(i=0; i<obj_cnt; i++)
			report_obj(lptr[i],tmp_fd);
	}

		/************* free malloc space *****************************/
	if(no_sort == 0  &&  mptr != (char *) 0)
		free_ltp(mptr);

		/************* close file ************************************/
	closef(tmp_fd);

		/************* move files if necess **************************/
	switch(outdev)
	{
	case 0:				/* Display on CRT */
		more(tmp_name);
		break;
	case 1:				/* Send to printer1 */
		print_file(tmp_name);
		break;
	case 2:				/* Send to printer2 */
		print_file(tmp_name);
		break;
	case 3:				/* Create ascii file */
		if( move_file( tmp_name, fname ) )
		{	if(!batch_mode)
			{	sprintf(tmp_msg, "Cannot create file %s.  \
Check file %s for output instead.", fname, tmp_name);
			}
		}
		break;
	default:			/* SHOULD NEVER HAPPEN */
		message("REPORT_LTP: Darned serious problem occurred. Good luck.");
		return(-1);
	}

		/************* write diary of report *************************/
	put_diary("report written.");

		/************* update report flag ****************************/
	job_stats[43] = job_stats[45] + 1;

		/************* return ****************************************/
return(0);
}

/*******************************************************************************
*    report_obj                                                                *
*         writes report line for 1 object.                                     *
*******************************************************************************/

long report_obj(i,tmp_fd)
long i;
int tmp_fd;
{
	OBJECTS object;
	ATTRIBUTES attr;

	char outrec[160];

	double length, area;

		/*** get object record ***/
	if( get_objects( &object, i ) )
	{	message("report:  warning, cannot get object record.");
		return(0);
	}

		/*** get 1st attribute record **/
	if(object.i12[2])
	{	if( get_attr(&attr, object.i12[2]) )
		{	message("report:  warning, cannot get attribute record.");
			return(-1);
		}
	}
	else
	{	attr.string[0] = (char) 0; 
		attr.next = 0;
	}

		/*** write 1st line of report for this object ***/
	length = object.f4[2] * unit_factor[len_glob][0];
	cum_len += length;
	if(object.i12[0] > 10)
		area = object.f4[3] * unit_factor[area_glob][1];
	else
		area = 0.0;
	cum_area += area;

	sprintf(outrec, "\n%-12s %8d %+12.2f %+12.2f %-30s\n",
	 wr_type[object.i12[0]], i, length, area, attr.string);

	if( writeb(outrec, strlen(outrec), 1, tmp_fd) == -1 )
	{	if( !batch_mode )
			message("REPORT_LTP: Cannot write to temp file. Bye.");
		closef(tmp_fd);
		return(-1);
	}

		/*** get 2nd attribute record **/
	if(attr.next)
	{	if( get_attr(&attr, attr.next) )
		{	message("report:  warning, cannot get attribute record.");
			return(-1);
		}
	}
	else
		attr.string[0] = (char) 0; 

		/*** write 2nd line of report for this object ***/
	sprintf(outrec, "%-12s %8d %+12.2f %+12.2f %-30s\n", 
		wr_subtype[object.i12[0]],object.i12[1],cum_len,cum_area,attr.string );
	if( writeb(outrec, strlen(outrec), 1, tmp_fd) == -1 )
	{	if( !batch_mode )
			message("REPORT_LTP: Cannot write to temp file. Bye.");
		closef(tmp_fd);
		return(-1);
	}

	return(0);
}

/*****************************************************************************
*    report_sort                                                             *
*                Sorts object numbers based on the global sort_type:         *
*                0=none,1=type,2=rec,3=serial,4=len,5=area,6=1st attr.       *
*****************************************************************************/

int report_sort(a,b)
	long *a,*b;
{
	OBJECTS obj_a,obj_b;
	ATTRIBUTES atr_a,atr_b;

		/**** get object 1, save value ****/
	if((sort_type > 0) && (sort_type != 2))
	{	if(get_objects(&obj_a,*a))
			return(0);
		if(get_objects(&obj_b,*b))
			return(0);
	}

	switch(sort_type) {
	case 0:            /** none **/
		return(0);
	case 1:            /** type: 0=node, 1=seg, 2=poly **/
		if(obj_a.i12[0] < obj_b.i12[0])
			return(-1);
		if(obj_a.i12[0] > obj_b.i12[0])
			return(1);
		return(0);
	case 2:            /** record num **/
		if(*a < *b)
			return(-1);
		if(*a > *b)
			return(1);
		return(0);
	case 3:            /** serial num **/
		if(obj_a.i12[1] < obj_b.i12[1])
			return(-1);
		if(obj_a.i12[1] > obj_b.i12[1])
			return(1);
		return(0);
	case 4:            /** length ***/
		if(obj_a.f4[2] < obj_b.f4[2])
			return(-1);
		if(obj_a.f4[2] > obj_b.f4[2])
			return(1);
		return(0);
	case 5:            /** area ***/
		if(obj_a.f4[3] < obj_b.f4[3])
			return(-1);
		if(obj_a.f4[3] > obj_b.f4[3])
			return(1);
		return(0);
	case 6:            /** 1st attribute **/
		if(!obj_a.i12[2]  ||  !obj_b.i12[2])  /** both don't have attributes **/
		{	if(obj_a.i12[2])
				return(1);
			if(obj_b.i12[2])
				return(-1);
			return(0);
		}
		else                                  /** both have attributes **/
		{	if(get_attr(&atr_a,obj_a.i12[2]))
				return(0);
			if(get_attr(&atr_b,obj_b.i12[2]))
				return(0);
			return(strncmp(atr_a.string,atr_b.string,30));
		}
	}
	return(0);
}
