/*******************************************************************************
*    vertex_thin   (see manual, Command Ref., User Functions)    <fc>          *
*            This program thins verticies of map objects in the specified set  *
*       & corridor radius & smoothes the result if desired.                    *
*                                                                              *
*       files/globals changed:  coord_list, objects, index_list, index_array   *
*                               job_stats                                      *
*                                                                              *
*       parameters:    long s      set number of objects to vertex_thin        *
*                      long u      units code, same choices as in report cmd.  *
*                      double r    corridor radius, in units of code above     *
*                      long m      smoothing code: 0=no smoothing              *
*                                                                              *
*       values returned:  long 0= ok,  1=failure                               *
*                                                                              *
*            J.Dabritz   7/31/90                                               *
*******************************************************************************/

#include "../../incl/globals.h"
#include <math.h>

long thin_span();

long vt_nextwrite;
double vt_radius;

long vertex_thin(s,u,r,m)
	long s,u;
	double r;
	long m;
{
	OBJECTS object,poly;
	COORD_LIST coordlist;

	char string[256];

	long i,in_set,j;

	double x1,y1,x2,y2;
	double a;


		/************* check for map loaded ************************/
	if(!job_stats[0])
	{	message("You must Create or Acquire a map befor executing the vertex_thin command.");
		return(1);
	}

		/************* check set number ****************************/
	if((s < 0) || (s > 31))
	{	message("Illegal set number.  Aborting.");
		return(1);
	}


		/************* check radius & units code *******************/
	if(r < (double) 0.0)
	{	message("Snap radius must be greater than or equal to zero.  Sorry Charlie.");
		return(1);
	}
	if((u < 0) || (u > 6))
	{	message("Illegal units code.  Aborting.");
		return(1);
	}
	get_units();
	{	if(unit_factor[u][0] == (double) 0.0)
		{	message("You must REGISTER the GEODETICS and RASTER befor choosing non-pixel units.");
			return(1);
		}
	}
	vt_radius = r / unit_factor[u][0];    /** find radius in pixels **/


		/************* for each object in the map ******************/
	if(r == (double) 0.0)
		i = job_stats[10];     /** no vertex reduction **/
	else
		i = 0;                 /** check objects for vertex reduction **/

	for( ; i<job_stats[10]; i++)
	{		/************* inform user *****************************/
		if(!(i % 25))
		{	sprintf(string,"Vertex_thin:  checking verticies on object %d.",i);
			message(string);
		}

			/************* check for user abort ********************/
		if(abort_cmd())
			break;

			/************* see if object is in set *****************/
		if(get_objects(&object,i))
		{	message("vertex_thin, abort, location 1.");
			return(1);
		}

		in_set = 0;                      /** default **/
		if((object.i12[0] > 7) && (object.i12[0] < 14))      /** segment **/
		{	if(object.i12[4] & (1 << s))
				in_set = 1;
			else
			{	for(j=7; j<9; j++)
				{	if(object.i12[j])
					{	if(get_objects(&poly,object.i12[j]))
						{	message("vertex_thin, abort, location 2.");
							return(1);
						}
						if(poly.i12[4] & (1 << s))
							in_set = 1;
					}
				}
			}
		}			/** polygons checked as part of segments **/

		if(in_set)
		{
			/************* SEGMENTS: check all verticies *******/
			vt_nextwrite = object.i12[5] + 1;

			if(get_coordlst(&coordlist,object.i12[5]))
			{	message("vertex_thin, abort, location 3.");
				return(1);
			}
			x1 = coordlist.x;
			y1 = coordlist.y; 

			if(get_coordlst(&coordlist,object.i12[6]))
			{	message("vertex_thin, abort, location 3.");
				return(1);
			}
			x2 = coordlist.x;
			y2 = coordlist.y; 

			if(thin_span(object.i12[5],object.i12[6],x1,y1,x2,y2))
			{	message("vertex_thin, abort, location 3.");
				return(1);
			}

				/******** re-save object ********************/
			object.i12[6] = vt_nextwrite - 1;
			if(put_objects(&object,i))
			{	message("vertex_thin, abort, location 7.");
				return(1);
			}
		}
	}
		
		/************* smooth if desired ***************************/
	if(m)
	{		/******* set default r if r== 0 *******/
		if(r == (double) 0.0)
			r = 1.0;

			/******* smooth set *******************/
		if(vertex_smooth(r,s))
		{	message("vertex_smoothing failed.");
			return(1);
		}
	}

		/************* recalculate lengths & areas *****************/
	recalc_area();

		/************* pack map object verticies *******************/
	if(vertex_pack())
	{	message("vertex_thin, abort, location 16.");
		return(1);
	}

		/************* re-index map objects ************************/
	if(index_map())
	{	message("vertex_thin, abort, location 17.");
		return(1);
	}

      /************* re-set max snap dist in job_stats file ******/
	a = 100.0 * job_stats[77];
	if(a < r)
	{	a = 100 * r + 0.5;
		job_stats[77] = a;
	}

		/*** redisplay the map ****/
	if((disp_hglight || disp_normal))
	{		/************* call dispgraphics to re-display data *********/
		i = (rwindowxmin + rwindowxmax) / 2;
		j = (rwindowymin + rwindowymax) / 2;
		if(dispgraphics(i,j,resolution_code))
			return(1);
	}

		/************* return **************************************/
	return(0);
}




/*******************************************************************************
    thin_span
       This program performs vertex weeding/thinning for a span of verticies
    from v1 to v2.  It is assumed that v1 has already been written, and is
    the lower numbered vertex.  The span is checked to see if it is within 
    radius 'radius' of a straight line from v1 to v2.  If it is, then v2 
    is written next.  Otherwise, the furthest point from the span (vf) is
    used as an intermediate point to recursively call this routine:
    thin_span(v1,vf)  and thin_span(vf,v2).

    parameters:   long v1;  lower numbered vertex, assumed already written.
                  long v2;  higher numbered vertex.
                  double x1,y1,x2,y2; vertex coordinates.

    globals used:  double r;  max dist (pixels) allowed for thinning
                   long vt_nextwrite; next vertex record number to be written.
*******************************************************************************/


long thin_span(v1,v2,x1,y1,x2,y2)
	long v1,v2;
	double x1,y1,x2,y2;
{
	static COORD_LIST coordlist;

	static long v;
	static long zero_dist;
	long vf;

	static double length;
	static double dx,dy;
	static double dx1,dy1;
	static double dx0,dy0;
	static double mat[2][3];
	static double dist_greatest;
	static double dist;
	double x,y;


		/******** get a vector from v1 to v2 ******************/
	dx = x2 - x1;
	dy = y2 - y1;

		/******** normalize the vector ************************/
	length = sqrt(dx * dx + dy * dy);
	if(length == 0.0)
		zero_dist = 1;
	else
	{	zero_dist = 0;
		dx0 = dx / length;
		dy0 = dy / length;
	}

		/******** test all verticies **************************/
	vf = -1;
	dist_greatest = 0.0;
	for(v = v1 + 1; v < v2; v++)
	{		/***** read vertex ****************************/
		if(get_coordlst(&coordlist,v))
		{	message("vertex_thin, abort, location 3.");
			return(1);
		}

			/***** find perpendicular dist ****************/
		if(zero_dist)
		{	dx1 = coordlist.x - x1;
			dy1 = coordlist.y - y1;
			dist = sqrt(dx1*dx1 + dy1*dy1);
		}
		else
		{	mat[0][0] = dx0;
			mat[0][1] = dy0;
			mat[0][2] = coordlist.x - x1;
			mat[1][0] = dy0;
			mat[1][1] = -dx0;
			mat[1][2] = coordlist.y - y1;
			
			if(!mat_solve(2,3,mat))
			{	if((mat[0][2] >= 0.0) && (mat[0][2] <= length))
					dist = fabs(mat[1][2]);
				else if(mat[0][2] < 0.0)
				{	dx1 = coordlist.x - x1;
					dy1 = coordlist.y - y1;
					dist = sqrt(dx1*dx1 + dy1*dy1);
				}
				else
				{	dx1 = coordlist.x - x2;
					dy1 = coordlist.y - y2;
					dist = sqrt(dx1*dx1 + dy1*dy1);
				}
			}
		}

		if(dist > dist_greatest)
		{	dist_greatest = dist;
			vf = v;
			x = coordlist.x; 
			y = coordlist.y; 
		}
	}
		
		/******** check for greatest dist *********************/
	if(dist_greatest <= vt_radius)  /** ok(within radius) write v2 coordinate **/
	{	coordlist.x = x2;
		coordlist.y = y2;
		if(put_coordlst(&coordlist,vt_nextwrite))
		{	message("vertex_thin, abort, location 6.");
			return(1);
		}
		vt_nextwrite++;
	}
	else     /** split span into two pecies & process both **/
	{	if(thin_span(v1,vf,x1,y1,x,y))
			return(1);
		if(thin_span(vf,v2,x,y,x2,y2))
			return(1);
	}
			
		/************* return **************************************/
	return(0);
}
