/*******************************************************************************
*	registerM	(see manual, Command Ref., Function Manager)	        <fm>   *
*                                                                              *
*			Function manager for the register function.                        *
*			Does extensive queries for user input depending on                 *
*			registration type:                                                 *
*                                                                              *
*			type  queries                                                      *
*			 0    	collects user input from graphics screen for 4             *
*				corners (raster pixels)                                        *
*			 1	allows user to input (on keyboard) all 4                       *
*				corners or specify 1 corner and a map series                   *
*				or ht & wdth. (N latitude E longitude)                         *
*			 2 	Prompts user for keyboard input of all 4                       *
*				corners. (UTM meters)                                          *
*			 3 	Prompts user for keyboard input of all 4                       *
*				corners. (State Plane Feet)                                    *
*			 4	Prompts user for digitizer input of all 4                      *
*				corners. (Dig1 absolute)                                       *
*			 5	Prompts user for digitizer input of all 4                      *
*				corners. (Dig2 absolute)                                       *
*			 6      Map inches.                                                *
*			 7-9    No input allowed, reserved for program use.                *
*			10-19	Prompts user for keyboard input of all 4                   *
*				corners.                                                       *
*			Note: for all prompting, the default values are the                *
*			current registration values.                                       *
*                                                                              *
*			Prompts user for options flags (see manual).  Prompts              *
*			user for user defined registration (string) name if                *
*			a user defined registration is being made.                         *
*			The coordinate type is not currently implemented.                  *
*                                                                              *
*			Calls function or writes command to macro being built              *
*			C.  Jonson                                                         *
*******************************************************************************/
 
#include <math.h>
#include "../../incl/globals.h"
#include "../../incl/in.h"
#include "../../incl/fc.h"
#include "../../incl/stplzn.h"
#include <string.h>
#define GEO_PROMPT 172		/*** Starting prompt for geodetics        ***/
#define UTM_PROMPT 180		/*** Starting prompt for UTM Coords       ***/
#define SPC_PROMPT 184		/*** Starting prompt for State Plane Feet ***/
#define MAPI_PROMPT 188		/*** Starting prompt for Map Inches ***/
#define UDU_PROMPT 192		/*** Starting prompt for User Defined Units ***/
#define DIG_PROMPT 196		/*** Starting prompt for digitizer absolute ***/
 
/***  Function declarations  ***/

void ltpquit();

long calc_ddmmss();      /*** Computes total degrees from deg ' "   ***/
long check_stplzn();     /*** Function Declaration                  ***/
long cnvt_ht_wd();       /*** Converts map ht & wd mins to degrees  ***/
double corner_rms();     /*** Function Declaration                  ***/
long dig_absolute();     /*** Registers digitizers before abs mode  ***/
long get_coord_type();   /*** Queries user to enter coordinate type ***/
long get_map_dim();      /*** Prompts user for map ht & wd          ***/
long get_raster_pix();   /*** Prompts user for 4 corner coord pairs ***/
long get_stplzn();       /*** Function Declaration                  ***/
long parse_corners();    /*** Parses command line parms into corners **/
long select_corner();    /*** Prompts for corner and its coords     ***/
long strtol();           /*** Convert strings into long values      ***/

/*** Global variables and definitions ***/
struct cnr_info		 /*** Holds the raw corner info ***/
{	long map_cnr;
	char x[64];
	char y[64];
};

long registerM(parms)
char *parms;
{
	double  ht, wd;             /***  Map ht + wd in degrees         ***/
	double  lat, lon;           /***  Latitude + longitude           ***/
	double  coord_pairs[4][2];  /***  Holds x,y pairs for register_ltp**/
	double  strtod();           /***  String to double decl          ***/

	long    cnr;                /***  map corner 1=SW 2=NW 3=NE 4=SE ***/
	long    coord_t;            /***  Coordinate type                ***/
	long    coord_indx = 0;     /***  Coordinate index               ***/
	long    i, j;               /***  For loop counter               ***/
	long    num_corners;        /***  Number of i1 (2,3,4) parms     ***/
	long    zone_val;           /***  Holds stpln zone value         ***/

	struct cnr_info	st_cnr[4];  /***  Holds info of the 4 corners    ***/

	char    parsed_buf[3][STRLIST];	/***  Used for parsing routine   ***/
	char    *ref_ptr;           /***  Used in strtol calls          ***/
	char    tmp_msg[96];        /***  General duty temp char array  ***/

    /***  Used for calculating remaining corners from one East coord pair. **/
	static  long geo_m[5][4][2] = { { {0,0}, {0,0}, {0,0}, {0,0} },
				       { {0,0}, {0,1}, {1,1}, {1,0} },
				       { {0,-1}, {0,0}, {1,0}, {1,-1} },
				       { {-1,-1}, {-1,0}, {0,0}, {0,-1} },
				       { {-1,0}, {-1,1}, {0,1}, {0,0} } };
	char    action = '\0';      /***  Flag for display and/or print ***/

/***  Ensure map is loaded before anything else.  ***/
	if( !job_stats[0] )
	{	if( !batch_mode )
			append_msg("  Acquire map before registering it.");
		return(0);
	}

/***  Ensure all parms are in lower case  ***/
	trim(lowerc(parms));

/***  Examine the parm for the display flag.  Load 'action' if there. ***/
	if( strstr(parms, "-d") )
	{	action = 'd';
		strcpy(raw_buf, "register -d");
		if( !collecting_macro )
			register_ltp(0L, 0L, coord_pairs, 0L, action);
		return(0);
	}

/***  Examine the parm for the print flag.  Load 'action' if there. ***/
	if( strstr(parms, "-p") )
	{	action = 'p';
		strcpy(raw_buf, "register -p");
		if( !collecting_macro )
			register_ltp(0L, 0L, coord_pairs, 0L, action);
		return(0);
	}

/***  Examine the parm string.  Read until a number is found.  This is ***/
/***  assumed to be the coordinate type.                               ***/
	if( strlen(parms) )
	{	parse( parms, delim_buf, parsed_buf, 3 );
		if( parsed_buf[0][0] == '-' )		/* Indicates opt parm */
		{	coord_indx = 1;		/* Coord type may be in [1].  */
		 	if( parsed_buf[1][0] == '-' )   /* Indicates opt parm */
			 	coord_indx = 2; /*  Coord type may be in [2]  */
		}
		if( !(coord_t = strtol(parsed_buf[coord_indx], &ref_ptr, 10)) )
			if( !strcmp( ref_ptr, parsed_buf[coord_indx]) )
				coord_t = -1;	/*** Set flag for prompting ***/
	}		
	else
	{	if( batch_mode )
			return(0);
		coord_t = -1;	/*** Set flag for user prompting  ***/
	}

/*** Ensure coord_t is within allowable range  ***/
	if( coord_t < 0 || coord_t > 29 )
	{	if( batch_mode )
			return(0);
		coord_t = get_coord_type();	/*** Get valid coord value ***/
	}

/*** Find the corner identifiers, if present ***/
	switch( coord_t )
	{	
/*---------------------- Raster Coordinates -----------------------------*/
    case 0:
	/***  Check for coords within parm string if = sign present. ***/
	if( strchr(parms, '='))
		if(parse_corners(st_cnr, parms) == 4)
		{   for( i = 0; i < 4; i++ )
		    {	coord_pairs[i][0] = strtod(st_cnr[i].x, (char **)NULL);
		    	coord_pairs[i][1] = strtod(st_cnr[i].y, (char **)NULL);
			}
			if( !collecting_macro )
			    register_ltp( coord_t, 0L, coord_pairs, 4L, action);
		    break;
    	}

	/***  Missing/too few coords in parm string.  Prompt user for them. ***/
	if( batch_mode )
		return(0);

	/*** Prompt user for all 4 corners. If true, user wants to quit. ***/
    if( get_raster_pix(coord_pairs, 176L) )
		return(0);
	register_ltp( coord_t, 0L, coord_pairs, 4L, action);
	break;
/*-------------------------- Geodetics ----------------------------------*/
	    case 1:
	/***  Check for coords within parm string if = sign present. ***/
	/*** check header for map ht and wd. 0 returned if found ***/
	/*** Will convert ht and wd into degrees.  (min/60)      ***/
	    if( strchr(parms, '=') == (char *)NULL )	/*** if no corner parms ***/
	    {	if( batch_mode )
			    return(0);
	     	if( !cnvt_ht_wd(&ht, &wd) )
			{   if(!select_corner(GEO_PROMPT, &lon, &lat, &cnr))
			    {   corner[1][0][0] = 0.0;    
					for(;lon < -180.0; lon += 360.0);
					for(;lon > 180.0; lon -= 360.0);
					lon = -lon;
			    	for(i=0; i < 4; i++)
					{	coord_pairs[i][0] = lon + geo_m[cnr][i][0] * wd;
						coord_pairs[i][1] = lat + geo_m[cnr][i][1] * ht;
					}
				    if( !collecting_macro )
				   		register_ltp(coord_t, 0L, coord_pairs, 4L, action);
			    }
			}
			/*** Check and see if map is rectangular. True, get corners.  ***/
			else if( answerisyes(226) )
			{	/* Check for default ht & wd, and convert from deg to min */
				if ( cnvt_ht_wd(&ht, &wd) )		/* True if no ht & wd */
				{	if( !get_map_dim(&ht, &wd) )	
			    	{	if( !select_corner(GEO_PROMPT,&lon,&lat,&cnr) )
						{	for(;lon < -180.0; lon += 360.0);
							for(;lon > 180.0; lon -= 360.0);
							lon = -lon;
							wd /= 60.0;		/* Convert the deg to minutes */
							ht /= 60.0;		/* Convert the deg to minutes */
			    			for(i=0; i < 4; i++)
							{   coord_pairs[i][0] = lon + geo_m[cnr][i][0] * wd;
				    			coord_pairs[i][1] = lat + geo_m[cnr][i][1] * ht;
			        		}
				    		if( !collecting_macro )
				   				register_ltp(coord_t,0L,coord_pairs,4L, action);
						}
			    	}
				}
			    else
			    {   while(!select_corner(GEO_PROMPT,&lon,&lat,&cnr))
					{	for(;lon < -180.0; lon += 360.0);
				    	for(;lon > 180.0; lon -= 360.0);
						lon = -lon;
				    	coord_pairs[cnr - 1][0] = lon + 
										geo_m[cnr][cnr - 1][0] * wd;
				    	coord_pairs[cnr - 1][1] = lat + 
										geo_m[cnr][cnr - 1][1] * ht;
				    	if( !collecting_macro )
							register_ltp(coord_t,cnr - 1,coord_pairs,1L,action);
					}
				}
			}			/***  Keep prompting for corners, as map ***/
			else		/***  is not rectangular.                ***/
				while(!select_corner(GEO_PROMPT, &lon, &lat, &cnr))
				{	for(;lon < -180.0; lon += 360.0);
					for(;lon > 180.0; lon -= 360.0);
					lon = -lon;
					coord_pairs[cnr - 1][0] = lon;
					coord_pairs[cnr - 1][1] = lat;
			        if(!collecting_macro)
			    	    register_ltp(coord_t,(cnr - 1),coord_pairs,1L,action);
				}
		}
	/*** Corner parms on command line  ***/
		else if( (num_corners = parse_corners(st_cnr, parms)) == 4)
		{	/** Register all four corners **/
			{	for( i = 0; i < 4; i++ )
				{	/*** The next two lines are critical!  The st_cnr.y ***/
				/*** value is assigned to lon and st_cnr.x is assigned  ***/
				/*** to lat.  This is correct, as geodetic registration ***/
				/*** is entered as lat, lon.  All others are lon, lat!  ***/
					calc_ddmmss(&lon, st_cnr[i].y);
					calc_ddmmss(&lat, st_cnr[i].x);
					for(;lon < -180.0; lon += 360.0);
					for(;lon > 180.0; lon -= 360.0);
					lon = -lon;
					coord_pairs[i][0] = lon;
					coord_pairs[i][1] = lat;
				}
				if( !collecting_macro )
					register_ltp(coord_t, 0L, coord_pairs, 4L,  action);
				break;
			}
			
		}
		else if( num_corners >= 1 ) 	/** Not all corners present. prompt **/
			/*** check header for map ht and wd. 0 returned if found ***/
			/*** Will convert ht and wd into degrees.  (min/60)      ***/
		{	if( !cnvt_ht_wd(&ht, &wd) )
			/*** Determine which corner was entered and register.    ***/
			{	for( j=0; j < 4  && !st_cnr[j].map_cnr ; j++ );
				cnr = st_cnr[j].map_cnr;
				calc_ddmmss(&lat, st_cnr[j].x);
				calc_ddmmss(&lon, st_cnr[j].y);
				for( i = 0; i < 4; i++ )
				{	for(;lon < -180.0; lon += 360.0);
					for(;lon > 180.0; lon -= 360.0);
					lon = -lon;
					coord_pairs[i][0] = lon + geo_m[cnr][i][0] * wd;
					coord_pairs[i][1] = lat + geo_m[cnr][i][1] * ht;
				}
				if( !collecting_macro )
					register_ltp(coord_t, 0L, coord_pairs, 4L,  action);
				break;
			}
	/*** Check and see if map is rectangular. True, get corners.  ***/
			else if( batch_mode )
				return(0);
			else if( answerisyes(226) )
			{	/*** Get map heigth and width as they are not in header ***/
				if( !get_map_dim(&ht, &wd) )	
			/*** Find the corner that was entered and display it to user.  ***/
				{	for( j=0; j < 4  && !st_cnr[j].map_cnr ; j++ );
					if( j < 4 )
					{	cnr = st_cnr[j].map_cnr;
						sprintf(tmp_msg, " Using i%d as base corner.", cnr);
						append_msg(tmp_msg);
					}
					else 
						select_corner(GEO_PROMPT,&lon, &lat, &cnr);

					for(wd /= 60.0, ht /= 60.0, i=0; i < 4; i++)
					{	for(;lon < -180.0; lon += 360.0);
						for(;lon > 180.0; lon -= 360.0);
						lon = -lon;
						coord_pairs[i][0] = lon + geo_m[cnr][i][0] * wd;
						coord_pairs[i][0] = lat + geo_m[cnr][i][1] * ht;
					}
				    if(!collecting_macro)
				   		register_ltp(coord_t, 0L, coord_pairs, 4L, action);
			    }
				else
				{	wd /= 60.0;	/*** cnv from mins to degs ***/
					ht /= 60.0;	/*** cnv from mins to degs ***/
					while(!select_corner(GEO_PROMPT,&lon,&lat,&cnr))
					{	for(;lon < -180.0; lon += 360.0);
						for(;lon > 180.0; lon -= 360.0);
						lon = -lon;
						coord_pairs[cnr - 1][0] = lon + 
										geo_m[cnr][cnr -1][0] * wd;
						coord_pairs[cnr - 1][1] = lat + 
										geo_m[cnr][cnr -1][1] * ht;
						if(!collecting_macro)
							register_ltp(coord_t,cnr - 1,coord_pairs,1L,action);
					}
				}
			}		/***  Keep prompting for corners, as map ***/
			else 	/***  is not rectangular.                ***/
				while(!select_corner(GEO_PROMPT,&lon,&lat,&cnr))
				{	for(;lon < -180.0; lon += 360.0);
					for(;lon > 180.0; lon -= 360.0);
					coord_pairs[cnr - 1][0] = -lon;
					coord_pairs[cnr - 1][1] = lat;
					if(!collecting_macro)
						register_ltp(coord_t, cnr - 1, coord_pairs, 1L, action);
				}
		}
		    break;
/*-------------------------- UTM Meters ----------------------------------*/
		case 2:
		if( !strchr(parms, '=') )	/** No coords present **/
		{	if( batch_mode )
				return(0);
			while( !select_corner(UTM_PROMPT,&lat,&lon,&cnr) )
			{	coord_pairs[cnr - 1][0] = lon;
				coord_pairs[cnr - 1][1] = lat;
				if(!collecting_macro)
					register_ltp(coord_t, (cnr-1), coord_pairs, 1L, action);
		    }
		}
	 	else
		/*** if 4 corner coord pairs not present, prompt for them ***/
		{	if(parse_corners(st_cnr, parms) == 4)
			{	for( i = 0; i < 4; i++ )
				{	coord_pairs[i][0] = strtod(st_cnr[i].x, (char **)NULL);
					coord_pairs[i][1] = strtod(st_cnr[i].y, (char **)NULL);
			    }
				if(!collecting_macro)
					register_ltp( coord_t, 0L, coord_pairs, 4L, action);
			    break;
			}
			else
			{	if( batch_mode )		/** Register whats there **/
				{	for( i = 0; i < 4; i++ )
					{	coord_pairs[i][0] = strtod(st_cnr[i].x, (char **)NULL);
						coord_pairs[i][1] = strtod(st_cnr[i].y, (char **)NULL);
					}
					if(!collecting_macro)
						register_ltp(coord_t, 0L, coord_pairs, 4L, action);
				}
				else
					while(!select_corner(UTM_PROMPT,&lat,&lon,&cnr))
					{	if(!collecting_macro)
						{	coord_pairs[cnr - 1][0] = lon;
							coord_pairs[cnr - 1][1] = lat;
							register_ltp(coord_t, (cnr - 1), coord_pairs, 
													1L, action);
						}
			        }
			}
		}
		break;
/*------------------------- State Plane Feet -----------------------------*/
		case 3:
		/***  Prompt user for all four corners  ***/
		if( !strlen(header.stplzn) )	/** no stpln zone **/
		{			/*** get state plane zone and validate it ***/
			if( batch_mode )
				return(0);
			message(prompt[134]);
			trim( get_string(tmp_msg) );
			if( !(zone_val = strtol( tmp_msg, (char **)NULL, 10 )))
			{	get_stplzn(&zone_val);
				if( zone_val == 0 )
				{	message("No state plane zone chosen. Register terminated.");
					return(0);
				}
			}
			else if( 101 > zone_val || zone_val > 5300 )
			{	if( zone_val != 9999 )
					get_stplzn(&zone_val);
			}
			
	/*** Check stplnzone for valid number.  If not ok, show user list.  ***/
			if( check_stplzn(zone_val) )
				get_stplzn(&zone_val);
			if( zone_val == 0 )
			{	message("No state plane zone chosen. Register terminated.");
				return(0);
			}
		}

		if( !strchr(parms, '=') )	/** No coords present **/
		{	if( batch_mode )
				return(0);
			while( !select_corner(SPC_PROMPT,&lat,&lon,&cnr) )
			{	if(!collecting_macro)
				{	coord_pairs[cnr - 1][0] = lon;
				 	coord_pairs[cnr - 1][1] = lat;
					register_ltp(coord_t,(cnr-1), coord_pairs, 1L, action);
				}
			}
		}
		else
		{	if(parse_corners(st_cnr, parms) == 4)
			{	for( i = 0; i < 4; i++ )
				{	coord_pairs[i][0] = strtod(st_cnr[i].x, (char **)NULL);
					coord_pairs[i][1] = strtod(st_cnr[i].y, (char **)NULL);
				}
				if(!collecting_macro)
					register_ltp( coord_t, 0L, coord_pairs, 4L, action);
			}
			else
			{	if( batch_mode )	/** Register whats there **/
				{	for( i = 0; i < 4; i++ )
					{	coord_pairs[i][0] = strtod(st_cnr[i].x, (char **)NULL);
						coord_pairs[i][1] = strtod(st_cnr[i].y, (char **)NULL);
			    	}
					if(!collecting_macro)
						register_ltp(coord_t, 0L, coord_pairs, 4L, action);
				}
				else
					while(!select_corner(SPC_PROMPT,&lat,&lon,&cnr))
					{	if(!collecting_macro)
						{	coord_pairs[cnr - 1][0] = lon;
							coord_pairs[cnr - 1][1] = lat;
							register_ltp(coord_t, (cnr - 1), coord_pairs, 
								1L, action );
						}
			    	}
			}
		}
	    break;
/*----------------------- Digitizer 1, Absolute coords -------------------*/
		case 4:
		if( batch_mode )
			return(0);
		/*** Register digitizer 1 ***/
		if( !dig_absolute(1) )
			return(0);
		 /** copy reg to corner 7 for use later **/
		for(i=0; i < 4; i++)
		{	corner[7][i][0] = corner[4][i][0];
			corner[7][i][1] = corner[4][i][1];
		}
		 /** generate a raster registration if there isn't one **/
		if(!registered(0))
			reg_raster(4);
	    break;
/*----------------------- Digitizer 2, Absolute coords -------------------*/
		case 5:
		if( batch_mode )
			return(0);
		/*** Register digitizer 2 ***/
		if( !dig_absolute(2) )
			return(0);
		 /** copy reg to corner 7 for use plotting later **/
		for(i=0; i < 4; i++)
		{	corner[7][i][0] = corner[5][i][0];
			corner[7][i][1] = corner[5][i][1];
		}
		/** generate a raster regsitration if there isn't one **/
		if(!registered(0))
			reg_raster(5);
		break;
/*----------------------------- Map Inches -------------------------------*/
		case 6:
		if( !strchr(parms, '=') )	/** No coords present **/
	    {	while( !select_corner(MAPI_PROMPT,&lat,&lon,&cnr) )
			{	if( !collecting_macro )
				{	coord_pairs[cnr - 1][0] = lon;
					coord_pairs[cnr - 1][1] = lat;
					register_ltp(coord_t, (cnr-1), coord_pairs, 1L, action);
				}
			}
		}
	    else
	    {	if(parse_corners(st_cnr, parms) == 4)
			{	for( i = 0; i < 4; i++ )
				{	coord_pairs[i][0] = strtod(st_cnr[i].x, (char **)NULL);
					coord_pairs[i][1] = strtod(st_cnr[i].y, (char **)NULL);
			    }
				if(!collecting_macro)
					register_ltp( coord_t, 0L, coord_pairs, 4L, action);
			}
			else
			{	if( batch_mode )	/** Register whats there **/
				{	for( i = 0; i < 4; i++ )
					{	coord_pairs[i][0] = strtod(st_cnr[i].x, (char **)NULL);
						coord_pairs[i][1] = strtod(st_cnr[i].y, (char **)NULL);
					}
					if(!collecting_macro)
						register_ltp(coord_t, 0L, coord_pairs, 4L, action);
				}
				else
					while(!select_corner(MAPI_PROMPT, &lat,&lon, &cnr))
						if( !collecting_macro )
						{	coord_pairs[cnr - 1][0] = lon;
							coord_pairs[cnr - 1][1] = lat;
							register_ltp(coord_t, (cnr - 1), coord_pairs,
										1L, action);
						}
			}
		}
		break;
/*------------------- Other is 10 - 29, otherwise invalid reg ------------*/
		default:
		if( coord_t >= 10 && coord_t <= 29 )
		{	if( !strchr(parms, '='))      /** No coords present **/
			{   if( batch_mode )
					return(0);
				while( !select_corner(UDU_PROMPT, &lat, &lon, &cnr) )
				{	if( !collecting_macro )
					{	coord_pairs[cnr - 1][0] = lon;
					 	coord_pairs[cnr - 1][1] = lat;
			    	    register_ltp(coord_t, (cnr-1), coord_pairs, 1L, action);
					}
				}
			}
			else
			{	if(parse_corners(st_cnr, parms))
				{	for( i = 0; i < 4; i++ )
					{	if( st_cnr[i].map_cnr )
						{	coord_pairs[i][0] = strtod(st_cnr[i].x,
											(char **)NULL);
							coord_pairs[i][1] = strtod(st_cnr[i].y,
											(char **)NULL);
			            }
					}
					if( !collecting_macro )
						register_ltp(coord_t, 0L, coord_pairs, 4L, action);
			    }
				else
				{	if( batch_mode )
						return(0);
					while(!select_corner(UDU_PROMPT,&lat,&lon,&cnr))
					{	if( !collecting_macro )
						{	coord_pairs[cnr - 1][0] = lon;
						 	coord_pairs[cnr - 1][1] = lat;
							register_ltp(coord_t, (cnr - 1), coord_pairs, 
											1L, action);
						}
					}
				}
			}
		}
		else
		{	if( !batch_mode )
				message("Invalid coordinate type.  Register aborted.");
			return(0);
		}
		break;
	}

/*** Generate the raw_buf string  ***/
	switch(coord_t)   {
	case 0:               /** raster ***/
		sprintf(raw_buf, "register %d i1=%.1f %.1f i2=%.1f %.1f\
 i3=%.1f %.1f i4=%.1f %.1f", coord_t, corner[coord_t][0][0], 
		corner[coord_t][0][1], corner[coord_t][1][0], 
		corner[coord_t][1][1], corner[coord_t][2][0], 
		corner[coord_t][2][1], corner[coord_t][3][0], 
		corner[coord_t][3][1] );
		break;
	case 1:
		sprintf(raw_buf, "register %d i1=%.3f %.3f i2=%.3f %.3f\
 i3=%.3f %.3f i4=%.3f %.3f", coord_t, corner[coord_t][0][1], 
		-corner[coord_t][0][0], corner[coord_t][1][1], 
		-corner[coord_t][1][0], corner[coord_t][2][1], 
		-corner[coord_t][2][0], corner[coord_t][3][1], 
		-corner[coord_t][3][0] );
		break;
	default:
		sprintf(raw_buf, "register %d i1=%.0f %.0f i2=%.0f %.0f\
 i3=%.0f %.0f i4=%.0f %.0f", coord_t, corner[coord_t][0][0], 
		corner[coord_t][0][1], corner[coord_t][1][0], 
		corner[coord_t][1][1], corner[coord_t][2][0], 
		corner[coord_t][2][1], corner[coord_t][3][0], 
		corner[coord_t][3][1] );
		break;
	}

	return(1);
}

/*******************************************************************************
*	get_coord_type                                                         *
*			Queries user to enter new coordinate type.  Will not   *
*			leave user alone until a valid number is entered.      *
*******************************************************************************/
long get_coord_type()
{
	long ct;
	char buf[512];

	for( ct = -1; ct< 0 || ct > 29; )
	{	message( prompt[50] );
		ct = strtol( get_string( buf ), (char **)NULL, 10 );
	}
	
	return(ct);
}
		
/*******************************************************************************
*	get_raster_pix                                                             *
*			Queries user to register the corners.                              *
*			Returns 0 when successful, or -1 if user wants to quit             *
*******************************************************************************/
long get_raster_pix(coords, promptn)
	double coords[][2];
	long promptn;
{
	long i;
	char parsed_str[3][STRLIST];

	for( i = 0; i < 4; i++, promptn++)
	{	message(prompt[promptn]);
		for(coords[i][0] = coords[i][1] = -1.0; 
					coords[i][0] == -1.0 || coords[i][1] == -1.0; )
		{	if( abort_cmd() )
				return(-1);	
			if( km() > 0 )		/* Input from keyboard avail */	
			{	trim(lowerc(raw_buf));
				if( !strncmp(raw_buf, "view", 4) )
				{	if( viewM( (raw_buf + 4) ) == 1)
					{	message(prompt[promptn]);
						continue;
					}
				}
				else if( parse(raw_buf, ", ", parsed_str, 3L) == 2)
				{	coords[i][0] = strtod( parsed_str[0], (char **)NULL );
			 		coords[i][0] = strtod( parsed_str[1], (char **)NULL );
				}
			}
			else if( gr_dev )
/**** MOD 11/6/89   by C. Jonson   Changed in following 'if':
                 was  ...gi() >= 0 ...     now ...gi() == 1...    *****/
			{	if( gi() == 1 && ginfo.button )
				{	trim( lowerc(raw_buf) );
					if( !strncmp(raw_buf, "view", 4) )
					 	if( viewM( (raw_buf + 4) ) == 1)
						{	message(prompt[promptn]);
							continue;
						}
					if( ginfo.tile_num == 0 )
					{	coords[i][0] = ginfo.tcoord[0];
						coords[i][1] = ginfo.tcoord[1];
					}
				}
			}
		}
	}
	return(0);
}
/*******************************************************************************
*	parse_corners                                                          *
*			Parses input parms into seperate corners.              *
*******************************************************************************/

long parse_corners(raw_cnr, raw)
	struct cnr_info *raw_cnr;	/*** Array of structs to hold
					     parsed strings.          ***/
	char *raw;			/*** String to parse up.  ***/
{
	long i, j, cnt = 0;
	char cstr[32];
	char *c_ptr, *eq_ptr;
	char parsed_val[2][STRLIST];

/*** Coord pairs are assumed to be in x, y format ***/
	for( i = 0, j = 1; i <= 4; i++, j++ )
	{	sprintf(cstr, "i%d", j);
		if( c_ptr = strstr(raw, cstr))	/*** Is corner present? ***/
		{	/*** Look past the = sign ***/
			if( eq_ptr = trim(strchr(c_ptr, '=') + 1) )
				/*** If 2 strings parsed, store them.  ***/
				if( parse(eq_ptr, " ,", parsed_val, 2L) == 2)
				{	(raw_cnr + i)->map_cnr = j;
					strcpy((raw_cnr + i)->x, parsed_val[0]);
					strcpy((raw_cnr + i)->y, parsed_val[1]);
					cnt++;
				}
		}
	}

	return(cnt);
}

/*******************************************************************************
*	select_corner                                                          *
*		Prompts user to select and input coords for one corner.        *
*               Returns 0 when user enters coord pair, else returns -1.        *
*******************************************************************************/

long select_corner(prompt_num, lon, lat, cnr)
	long prompt_num;
	double *lon, *lat;
	long *cnr;			
{
	char parsed_num[2][STRLIST];	/***  Stores users input coords ***/
	long prmpt_indx;		/***  index for prompt array ***/
	char resp[512];			/***  String for user response ***/

/***  Prompt user to select corner.  1=SW 2=NW 3=NE 4=SE  ***/
	for( *cnr = 0; !(*cnr); )
	{	message(prompt[167]);
		lowerc( trim( get_string(resp) ) );
		if( !strlen(resp) )	/*** Empty...user wants to quit. ***/
			return(1);
		if( strstr(resp, "sw") || strchr(resp, '1') )
			*cnr = 1;
		else if( strstr(resp, "nw") || strchr(resp, '2') )
			*cnr = 2;
		else if( strstr(resp, "ne") || strchr(resp, '3') )
			*cnr = 3;
		else if( strstr(resp, "se") || strchr(resp, '4') )
			*cnr = 4;
	}

/***  Prompt for latitude and longitude.  Degrees/min/sec are used. ***/
/***  Loop until two strings are parsed from the input.             ***/

	for( prmpt_indx = prompt_num + *cnr - 1; ; )	/** Infinite loop **/
	{	message(prompt[prmpt_indx]);
		trim(get_string(resp));
		if( parse(resp, ", ", parsed_num, 2L) == 2)
		{	if((parsed_num[0][0] >= '-') && (parsed_num[0][0] <= '9') &&
			   (parsed_num[1][0] >= '-') && (parsed_num[1][0] <= '9'))
				break;
		}
		else
			return(-1);
	}

	if( prompt_num == GEO_PROMPT )
	{	calc_ddmmss(lat, parsed_num[0]);
		calc_ddmmss(lon, parsed_num[1]);
	}
	else
	{	*lat = strtod( parsed_num[0], (char **)NULL );
	 	*lon = strtod( parsed_num[1], (char **)NULL );
	}

	return(0);
}

/*******************************************************************************
*	calc_ddmmss                                                            *
*			Computes the total degrees from ddd.mmm.sss input.     *
*******************************************************************************/

long calc_ddmmss( result, raw_str )
	double *result;
	char *raw_str;
{
	double degrs, mins, secs;	/*** for intermediate calcs ***/
	long num_parsed;
	double strtod();
	char parsed_num[3][STRLIST];
	
	for( *result = 0.0;  !(*result); )
	{	num_parsed = parse(raw_str, "/:;\"\'", parsed_num, 3L);
		if( !num_parsed )		/*** No value, loop again ***/
			continue;
		if( num_parsed == 3 )
		{	degrs = strtod(parsed_num[0], (char **)NULL);
			mins = strtod(parsed_num[1], (char **)NULL) / 60.0;
			secs = strtod(parsed_num[2], (char **)NULL) / 60.0/60.0;
			*result = degrs + mins + secs;
		}
		else if( num_parsed == 2 )
		{	degrs = strtod(parsed_num[0], (char **)NULL);
			mins = strtod(parsed_num[1], (char **)NULL) / 60.0;
			*result = degrs + mins;
		}
		else if( num_parsed == 1 )
		{	degrs = strtod(parsed_num[0], (char **)NULL);
			*result = degrs;
		}
	}
	return(0);
}

/*******************************************************************************
*	cnvt_ht_wd                                                             *
*			Checks for the map height and width in the header      *
*			file, and if present, converts them to degrees.        *
*			                                                       *
*			Returns 0 if OK, else returns 1 if no header info.     *
*******************************************************************************/

long cnvt_ht_wd( ht, wd )
	double *ht, *wd;
{
	double ht_minutes, wd_minutes, strtod();

/*** Check header for presence of map height and width ***/
	if( strlen(header.map_ht) && strlen(header.map_wd) )
	{	ht_minutes = strtod( header.map_ht, (char **)NULL );
	 	wd_minutes = strtod( header.map_wd, (char **)NULL );
		*ht = ht_minutes / 60.0;		/** convt to degrees **/
		*wd = wd_minutes / 60.0;		/** convt to degrees **/
		return(0);
	}
	else
		return(1);
}

/*******************************************************************************
*	get_map_dim                                                            *
*			Prompts user to enter the map height and width in      *
*			minutes (taken as any decimal number), then stores     *
*			them in the header file.                               *
*			                                                       *
*			Returns 0 if OK, else returns 1 if unable to           *
*			update the header file.                                *
*******************************************************************************/

long get_map_dim(ht, wd)
	double *ht, *wd;
{
	double strtod();
	char resp1[STRLIST];
	char resp2[STRLIST];
	char header_type[32];

/***  Get height and width of current map  ***/
	for( *wd = 0.0; !*wd; )
	{
		message("Enter MAP WIDTH expressed in MINUTES.");
		trim( get_string( resp1 ) );
		*wd = strtod( resp1, (char **)NULL );
	}

	for( *ht = 0.0; !*ht; )
	{
		message("Enter MAP HEIGHT expressed in MINUTES.");
		trim( get_string( resp2 ) );
		*ht = strtod( resp2, (char **)NULL );
	}

/***  Create command string to pass to header ***/
	strcpy(header_type, "map_wd");
	if( !header_ltp(0, header_type, resp1))
		return(1);

	strcpy(header_type, "map_ht");
	if( !header_ltp(0, header_type, resp2))
		  return(1);

	return(0);
}

/***************************************************************************
*   dig_absolute                                                           *
*           Reads either dig1 or dig2 based on parameter.  Will register   *
*           using relative mode but the cursor will NOT be displayed       *
*           during the registration process.                               *
***************************************************************************/

long dig_absolute( dig_number )
	long dig_number;
{
	double  rms_dbl;
	double  coord_pairs[4][2];
	double  dig_perim, utm_perim;
	double  last_digx, last_digy;
	double  digx, digy;
	double  last_utmx, last_utmy;
	double  utmx, utmy;
	double  dx, dy;
	double  dig_scale;

	long    i, j;
	long    amode, last_gr_dev;
	long    register_indx;		/* Indicates device to be registered */

	char    tpath[STRLIST];
	char    tmp_msg[STRLIST];
	char    action = '\0';

/***  Place digitizer in relative mode until map registered ***/
	sprintf(tmp_msg, "Switching to digitizer %d...", dig_number);
	message(tmp_msg);
/*	last_gr_dev = gr_dev;			/*  Store current grdev   */
	last_gr_dev = (gr_dev + 1) / 2;
	last_gr_dev = (gr_dev > 4) ?  3 : last_gr_dev;
	amode = gr_dev % 2;
	amode = (gr_dev > 4) ?  0 : amode;
	if( ginput( dig_number, 0, 0 ) )		/*  Change graphics device */
	{	ginput(last_gr_dev, amode, 0); 		/*  Can't change? Restore */
		message("Can't initialize digitizer, returning.");
		return(0);			/*  graphics device global.  */
	}
	append_msg("done");
	register_indx = (dig_number == 1) ? 4 : 5 ;
	for( i = 0; i < 4; i++ )
	{	message( prompt[DIG_PROMPT + i]);
		ginfo.button = 0;			/* NULL out ginfo.button */
		/*** Loop until digcursor button hit or key pressed ***/
		while( gi() != 1 && !ginfo.button )		/* while no gr input */
		{	if( abort_cmd() )				/* if esc or del pressed, abort */
			{	message("register terminating per user request. ");
				ginput(last_gr_dev, amode, 0);    /*  reset to orig device */
				return(0);
			}
		}
	/* NULL out raw_buf to keep button hits from being displayed on tile 4. */
		for( j = 0; j < STRLIST; *(raw_buf + j++) = 0 );
		coord_pairs[i][0] = (double)ginfo.rawcoord[0];
		coord_pairs[i][1] = (double)ginfo.rawcoord[1];
	}
	if(!collecting_macro)
	{
		register_ltp(register_indx, 0L, coord_pairs, 4L, action);
	}

/***  report registered scale ****/
		/*** compare digitized perimeter with UTM perimeter ***/
	dig_scale = 0.0;
	if(registered(2))
	{
		dig_perim = 0.0;
		utm_perim = 0.0;
		last_digx = corner[register_indx][3][0];
		last_digy = corner[register_indx][3][1];
		last_utmx = corner[2][3][0];
		last_utmy = corner[2][3][1];
		for(i=0;i<4;i++)
		{	digx = corner[register_indx][i][0];
			digy = corner[register_indx][i][1];
			dx = digx - last_digx;
			dy = digy - last_digy;
			dig_perim += sqrt(dx * dx + dy * dy);
			last_digx = digx;
			last_digy = digy;

			utmx = corner[2][i][0];
			utmy = corner[2][i][1];
			dx = utmx - last_utmx;
			dy = utmy - last_utmy;
			utm_perim += sqrt(dx * dx + dy * dy);
			last_utmx = utmx;
			last_utmy = utmy;
		}
			/*** convert digitizer units (in) to meters ***/
		if(diginfo.units_per_inch > 0.0)
		{	dig_perim = dig_perim / diginfo.units_per_inch / 12.0 / 3937.0 * 1200.0;
			if(dig_perim > 0.0);
				dig_scale = utm_perim / dig_perim;
		}
	}

/***  Check for UTM registration.  If registered, report RMS of corners ***/
	if( registered(2) )		/** '2' is for UTM  **/ 
	{
		sprintf(tpath, "%s/tolerance", ltp_path);
		get_toler();
		rms_dbl = corner_rms(register_indx, 2);	
		rms_dbl *= (3937.0/1200.0);   /** Conv't meters -> ft */
		if( rms_dbl < 0.0 )
			message("Cannot calculate Root Mean Square.");
		else
		{	sprintf(tmp_msg, "Tolerance is %f  RMS is %.2f ground feet,  Scale = 1:%.0f", tolerance, rms_dbl,dig_scale);
			message(tmp_msg);
		}
	}
	else
		message("UTM not registered.  Cannot calculate Error or Scale.");

	ginput(last_gr_dev, amode, 0);     /*  reset to original device */

	return(1);
}

/***************************************************************************
*   get_stplzn                                                             *
*           Prompts the user to select a state plane zone number from      *
*           the displayed list.  The zone is then added to the header      *
*           file by using put_header().                                    *
*           Returns 0 upon successful EOJ, else returns -1.                *
***************************************************************************/

long get_stplzn( zone_num )
	long *zone_num;
{
	long arr_size;
	long choice;
	char hdr_path[STRLIST];

	arr_size = sizeof(stplzn_arr) / STRLIST;
	if( (choice = list_choice( stplzn_arr, arr_size, 229)) == -1 )
	{	message("Unable to display state plane zone choices.");
		return(-1);
	}
	
	if( choice == 0 )
		*zone_num = 0;
	else
	{	sscanf(stplzn_arr[choice], "%s", header.stplzn);
		sprintf(hdr_path, "%s/header", temp_map_path);
		*zone_num = strtol(header.stplzn, (char **)NULL, 10);
		if( put_header(hdr_path,&header) )
		{	message("Unable to update header file.");
			return(-1);
		}
	}

	return(0);
}

/***************************************************************************
*   check_stplzn                                                           *
*           Compares the parameter to state plane zone values.  If a       *
*           match is found, 0 is returned.                                 *
*           Returns 0 upon finding a match, else returns -1.               *
***************************************************************************/

long check_stplzn( zone_num )
	long zone_num;
{
	long arr_size;
	long found = FALSE;
	long i;
	long arr_zone;

/***  Calculate the size of the stplzn_arr ***/
	arr_size = sizeof(stplzn_arr) / STRLIST;

/***  Search array for a hit.  i set to 1 to avoid " -none- " (which is 0)  ***/
	for( i = 1; i <= arr_size; i++ )
	{
		sscanf( stplzn_arr[i], "%ld", &arr_zone );
		if( arr_zone == zone_num )
		{	found = TRUE;
			break;
		}
	}

	if( found )
		return(0);
	else
		return(-1);
}
