/*******************************************************************************
*       atrbt_profil    (see manual, Internal Functions)                <in>   *
*    assigns attributes to a series of segments crossed by the input           *
*    (atrbt_contur) profile section (using the increment or decrement          *
*    commands.  The elevations assigned are given by the globals               *
*    elev_above and elev_below and are changed by contour_interval as          *
*    incremented or decremented (parameter).   This is the case except for     *
*    the first segment crossed when the context_elev global flag is 1.         *
*    When the flag is == 1, the first segment crossed is found, and its        *
*    elevation is used to set the remaining contours.                          *
*                                                                              *
*    Arguments:    long rx1,ry1;  raster coords of profile section beginning   *
*                  long rx2,ry2;  raster coords of porfile section end.        *
*                  long dir;      direction: 1=up, -1=down, 0=remove atrbts    *
*                                                                              *
*    Returned Values: (long) 0 = ok,    1=failure                              *
*                                                                              *
*    Global Variables Modified:  elev_above, elev_below, context_elev.         *
*                                                                              *
*    Files modified:    objects, attributes.                                   *
*                                                                              *
*                      j.dabritz  8/15/89                                      *
*******************************************************************************/

#include "../../incl/globals.h"
#include <math.h>

int compair_ps();
long prof_cellchk();

		/*** global structure for collecting up to 250 segment numbers and
			their crossing distances from the profile beginning (rx1,ry1) **/
struct prof_struct { double dist;
		long obj;
		}   segs[250];
long segments;                  /** the number of segments collected **/


long atrbt_profil(rx1,ry1,rx2,ry2,dir)
long rx1,ry1,rx2,ry2;
long dir;
{
	OBJECTS object;
	ATTRIBUTES atrbt;

	char string[256];

	long i;
	long ix1, iy1, ix2, iy2;
	long iystep, ixstep;
	long last_seg;
	long last_i_atrbted;
	long elevation;

	double x1,y1,x2,y2;
	double dx,dy;
	double slope;
	double xstep, ystep;
	double xenter, yenter;
	double xexit, yexit;


		/********* check for map ****************************************/
	if(!job_stats[0])
	{	message("abort, atrbt_profil, location 1");
		return(1);
	}

		/*** check input for 0 length ********/
	if((rx1 == rx2)  &&  (ry1 == ry2))
		return(0);       /** can't cross a segment if ya don't move */

		/********* Generate list of all segments crossing profile section */
		/** first calculate index_array coordinate values for endpoints **/
	x1 = (rx1 + 0.5) / 32.0;         	/* fractional cell x */
	y1 = (ry1 + 0.5) / 32.0;            /* fractional cell y */
	ix1 = x1;                         /* integer cell x    */
	iy1 = y1;                         /* integer cell y    */
	x2 = (rx2 + 0.5) / 32.0;         	/* fractional cell x */
	y2 = (ry2 + 0.5) / 32.0;         	/* fractional cell y */
	ix2 = x2;                         /* integer cell x    */
	iy2 = y2;                         /* integer cell y    */


		/** initialize array of segments crossed **/
		/** (must initialize due to possible recursive calls) **/
	for(i=0; i<250; i++)
	{	segs[i].dist = 2.0;
		segs[i].obj = 0;
	}
	segments = 0;

		/******* next go through each cell accumulating a list of 
				segments which cross along with their distance from
				the beginning (rx1,ry1) of the profile section  */

		/** if profile section is entirely in 1 cell, check it ***/
	if((ix1 == ix2) && (iy1 == iy2))
	{	if(prof_cellchk(ix1,iy1,rx1,ry1,rx2,ry2))
			return(1);
	}

		/** otherwise check all cells through which the profile goes  **/
	else
	{	dx = x2 - x1;        /*  change in cell coordinates */
		dy = y2 - y1;
			/*** check for more vertical or horizontal vector */
		if(fabs(dy) > fabs(dx))       /* more vertical than hz */
		{		/* calc y slope */
			slope = dx / dy;

				/* set variables for checking by row */
			if(y2 > y1)
				ystep = 1.0;
			else
				ystep = -1.0;
			xstep = ystep * slope;
			if(xstep > 0.0)
				ixstep = 1;
			else
				ixstep = -1;

				/** set enter and exit variables for first col */
			xenter = x1;
			yenter = y1;
			if(y2 > y1)
			{	i = (yenter + ystep);
				yexit = i;
			}
			else
			{	i = (yenter);
				yexit = i;
			}
			xexit = xenter + (yexit - yenter) * slope;

				/*** check cells, one row at a time **/
			for( ; ; )
			{		/*** check cells for one row **/
				for(;; ix1 += ixstep)
				{		/** check 1 cell **/
					if(prof_cellchk(ix1,iy1,rx1,ry1,rx2,ry2))
						return(1);

						/*** check for end of cell row used **/
					if(ix1 == (long) xexit)
						break;

						/*** check if done **/
					if((iy1 == iy2) && (ix1 == ix2))
						break;
				}
					/*** check if done **/
				if((iy1 == iy2) && (ix1 == ix2))
					break;

					/*** calc xenter, xexit for next row **/
				xenter = xexit;
				xexit = xenter + xstep;

				if(iy1 != iy2)
					iy1 += ystep;
			}
		}
		else                          /* more horizontal that vt */
		{		/* calc x slope */
			slope = dy / dx;

				/* set variables for storing by col */
			if(x2 > x1)
				xstep = 1.0;
			else
				xstep = -1.0;
			ystep = xstep * slope;
			if(ystep > 0.0)
				iystep = 1;
			else
				iystep = -1;

				/** set enter and exit variables for first row */
			yenter = y1;
			xenter = x1;
			if(x2 > x1)
			{	i = (xenter + xstep);
				xexit = i;
			}
			else
			{	i = (xenter);
				xexit = i;
			}
			yexit = yenter + (xexit - xenter) * slope;

				/*** check path, one col at a time **/
			for( ; ; )
			{		/*** check 1 column of cells in a row **/
				for(;; iy1 += iystep)
				{		/** check 1 cell **/
					if(prof_cellchk(ix1,iy1,rx1,ry1,rx2,ry2))
						return(1);

						/*** check for end of vector **/
					if(iy1 == (long) yexit)
						break;

						/*** check if done ***/
					if((iy1 == iy2) && (ix1 == ix2))
						break;
				}
					/*** check if done ***/
				if((iy1 == iy2) && (ix1 == ix2))
					break;

					/*** calc yenter, yexit for next col **/
				yenter = yexit;
				yexit = yenter + ystep;
					/*** increment row **/
				if(ix1 != ix2)
					ix1 += xstep;
			}
		}
	}
		/********* check for more than 249 segments on list *************/
	if(segments >= 250)
	{		/*** inform user & abort **/
		message("cannot attribute more than 250 segment crossings on 1 profile section.");
		context_elev = 1;
		return(1);
	}
	else if(segments > 0)
	{		/********* sort list into asending order ********************/

		qsort(&segs[0],segments,sizeof(struct prof_struct),compair_ps);

			/********* attribute segments *******************************/
		last_seg = 0;
		last_i_atrbted = -1;
		for(i=0; i<segments; i++)
		{		/*** check for done **/
			if(!segs[i].obj)
				break;
				/********* see if seg is different that previous one ****/
			if(segs[i].obj  &&  (segs[i].obj != last_seg))
			{		/********* if object not displayed, don't attribute **/
/*				if(!disp_object(segs[i].obj))                                 */
/*					sleep(0);		  ** it's not displayed, do nothing **      */
					
/*				else if(context_elev)    ** need an elevation **/
				if(context_elev)    /** need an elevation **/
				{	if(get_objects(&object,segs[i].obj))
						return(1);
					if(object.i12[2])	/** has attribute **/
					{	if(get_attr(&atrbt,object.i12[2]))
							return(1);
						if(atrbt_action == 2)  /*  need 2nd attribute  */
						{	if(atrbt.next)
							{	if(get_attr(&atrbt,atrbt.next))
									return(1);
							}
							else
								atrbt.string[0] = 0;	
						}
						 /*** now decode string into an elevation **/
						if((*atrbt.string >= '0')  &&  (*atrbt.string <= '9'))
						{	elevation = strtol(atrbt.string, (char **)NULL, 10);
							context_elev = 0;
							  /*** last, reset elev_above & elev_below **/
							if(dir > 0)
							{	elev_below = elevation;
								elev_above = elevation + contour_interval;
							}
							else if(dir < 0)
							{	elev_below = elevation - contour_interval;
								elev_above = elevation;
							}
						}
					}
				}
				else
				{		/*** set elev, & reset elev_above & below *******/
					if(dir > 0)
					{	elevation = elev_above;
						elev_below = elevation;
						elev_above = elevation + contour_interval;
					}
					else if(dir < 0)
					{	elevation = elev_below;
						elev_below = elevation - contour_interval;
						elev_above = elevation;
					}
						/*** set attribute string, store & display **/
					sprintf(string,"%d",elevation);
					if(dir)
					{	if(asgn_atrbt(segs[i].obj,atrbt_action,string))
							return(1);
					}
					else
					{	if(asgn_atrbt(segs[i].obj,7,string))
							return(1);
					}
					last_i_atrbted = i;
				}
					/*** set marker to last_seg  ***/
				last_seg = segs[i].obj;
			}
			else if(segs[i].obj  &&  (segs[i].obj == last_seg))
			{	if(last_i_atrbted >= 0)
				{	if(fabs((double) (segs[i].dist - segs[last_i_atrbted].dist)) > 0.01)
					{	i = last_i_atrbted;
						asgn_atrbt(segs[i].obj,7,string);
						if(dir)
							pause_ltp(3,"Attempting to give same contour 2 different elevations.  Command aborted.");
						break;
					}
				}
			}
		}
	}

		/********* return ***********************************************/
	return(0);
}
/*******************************************************************************
*     compair_ps                                                               *
*          This is the compairison function for the qsort in atrbt_prifile.    *
*     It compairs the double (.dist) in two prof_struct structures.            *
*                                                                              *
*******************************************************************************/
int compair_ps(a,b)
struct prof_struct *a, *b;
{
	if(a->dist > b->dist)
		return(1);
	if(a->dist < b->dist)
		return(-1);
	return(0);
}

/*******************************************************************************
*      prof_cellchk                                                            *
*             This program checks 1 index array cell against the input         *
*      contour profile section.  It stores segment numbers and their           *
*      intersection distances from the beginning of the section in the         *
*      global segs structure array.  Intersection crossing duplications        *
*      are ignored at this point.  They will look after themselves after       *
*      sorting the structure array.                                            *
*                                                                              *
*      J.Dabritz  8/15/89                                                      *
*******************************************************************************/



long prof_cellchk(cellx,celly,x1,y1,x2,y2)
long cellx,celly;           /** index_array cell coordinates **/
long x1,y1;                 /** beginning point of attribute profile section */
long x2,y2;                 /** ending point of attribute profile section */
{
	INDEX_LIST index_list;
	COORD_LIST coord_list;

	long v;

	double px1,py1;            /** first profile point coords **/
	double pdx,pdy;            /** dx and dy for profile points **/
	double sx1,sy1,sx2,sy2;    /** adjacent segment vertex coords **/
	double matrix[2][3];       /** for solving for intersection of lines **/

		/********* check number of sections stored already *************/
	if(segments >= 250)
		return(0);

		/********* check cell coordinates for on raster ****************/
	if((cellx < 0)  ||  (cellx > (job_stats[1] >> 5)))
		return(0);
	if((celly < 0)  ||  (celly > (job_stats[2] >> 5)))
		return(0);

	px1 = x1;
	py1 = y1;
	pdx = x2 - x1;   /** profile section deltas **/
	pdy = y2 - y1;
		/********* go through index_list linked-list for cell **********/
	index_list.next = index_array[celly][cellx];
	for(; index_list.next ;)
	{		/** get index_list record **/
		if(get_indexl(&index_list,index_list.next))
			return(1);

			/********* check object type for segment *******************/
		if(index_list.type == 1)      /** map segment **/
		{		/********* go through coord_list checking for intersections */
			if(get_coordlst(&coord_list,index_list.v1))
				return(1);
			sx1 = coord_list.x;
			sy1 = coord_list.y;
			for(v=index_list.v1+1;  v <= index_list.v2; v++)
			{		/********* get vertex ******************************/
				if(get_coordlst(&coord_list,v))
					return(1);
					/********* calc matrix variables *******************/
				sx2 = coord_list.x;
				sy2 = coord_list.y;
				matrix[0][0] = pdx;
				matrix[0][1] = sx2 - sx1;
				matrix[0][2] = sx2 - px1;
				matrix[1][0] = pdy;
				matrix[1][1] = sy2 - sy1;
				matrix[1][2] = sy2 - py1;

					/********* solve for intersection ******************/
				if(!mat_solve(2,3,matrix[0]))
				{		/***** if intersection between 0 & 1, a hit! ***/
					if((matrix[0][2] >= 0.0)  &&  (matrix[0][2]  <= 1.0)  &&
					   (matrix[1][2] >= 0.0)  &&  (matrix[1][2]  <= 1.0))
					{	/***** store object no & dist ******************/
						segs[segments].dist = matrix[0][2];
						segs[segments].obj = index_list.obj_rec;
						segments++;
						if(segments >= 250)
							return(0);
					}
				}
					/**** increment vector verticies **/
				sx1 = sx2;
				sy1 = sy2;
			}
		}
	}
		/********* return **********************************************/
	return(0);
}
