/*******************************************************************************
* discio.c  This file contains all routines which do/control disc input/output *
*      operations for LTPlus.   These include:                                 *
*         openf()           opens files                                        *
*         closef()          closes files                                       *
*         closef_ptr()      closes files & resets file pointers to -1          *
*         openfiles_list()  fills a list with the names of all open files      *
*         movefp()          repositions a file read/write pointer within a file*
*         readb()           reads a file binary-wise (given number of bytes)   *
*         reada()           reads a file ascii-wise (lf, eof, or #bytes read)  *
*         reada_old()       reads a file ascii-wise (but slower)               *
*                                                                              *
*     The primary reason from grouping all of these functions into a single    *
*   block of code is to share global variables in order to track which files   *
*   are left open accidently & to speed reading of ascii (data sensitive)      *
*   files.     file grouped 1/2/91, JPD                                        *
*******************************************************************************/
/* Modified 6/92 by DBS@CWU for Ultrix support */

#include "../../incl/globals.h"
#include <string.h>

char open_fnames[60][STRLIST];     /** names of all currently open files **/
int  file_pointer[60];             /** pointers to all currently open files **/

char readblock[5][1028];           /** 1k blocks of data used for reada()   **/
int  fileptr[5] = {-1,-1,-1,-1,-1};/** file pointers for each readblock    **/
char *readptr[5][2];               /** next & end pointers for each readblock**/


int openf();
int closef();
int closef_ptr();
long openfiles_list();
int movefp();
int readb();
char *reada();
char *reada_old();


/*******************************************************************************
* openf                (see manual, Internal Functions)                 <in>   *
*                       Opens a file based on the type parameter.  See the     *
*                       ../../incl/define.h file for type choides.             *
*                                                                              *
*                       Returns  -1  upon abnormal end-of-job (EOJ), else      *
*                       returns  the opened file descriptor.                   *
*                  HLM&JPD030789                                               *
*******************************************************************************/

int openf( fname, type )	
	char *fname;
	int type;
{
/*	char string[256];     */
	int fd;
	long i;

	fd = open(fname, type);

/* Next line Added for ULTRIX port  */
	if (fd < 0) fd = -1;   /* calling routines expect -1, not less than 0 */

/*	sprintf(string," =======> OPEN(%3d) %s  %d<==========",fd,fname,type);     */
/*	message(string);     */
	for(i=0; file_pointer[i]>0; i++);
	file_pointer[i] = fd;
	strncpy(open_fnames[i],fname,STRLIST);
	open_fnames[i][STRLIST-1] = (char) 0;
	return( fd );
}


/*******************************************************************************
* closef               (see manual, Internal Functions)                 <in>   *
*                       Close an open file descriptor.                         *
*                                                                              *
*                       Returns zero upon normal end-of-job (EOJ), else        *
*                       returns -1 and sets errno.                             *
*                  HLM&JPD030789                                               *
*******************************************************************************/


int closef( fd )
	int fd;
{
/*	char string[256];   */
	long i;

	if(fd >= 0)
	{
/*		sprintf(string," =======> close(%3d) <==========",fd);   */
/*		message(string);   */
			/** check for & reset any file pointers to reada data  **/
		for(i=0; i<5; i++)
		{	if(fileptr[i] == fd)
				fileptr[i] = -1;
		}

		for(i=0;(file_pointer[i] != fd) && (i<60); i++);
		if(i<60)
			file_pointer[i] = -1;
		else
			message("  Attempting to close a file NOT opened using openf().");
		return( close(fd) );
	}
	else
		return(-1);
}


/*******************************************************************************
*    closef_ptr                                                                *
*           This program closes a file if given a legetimate pointer & the     *
*      value pointed to is also legitimate.  After closing, the value is set   *
*      to -1.                                                                  *
*                                                                              *
*      Parameter:   int *ptr     pointer to int (fd) file pointer.             *
*                                                                              *
*      Value returned:    0                                                    *
*                                                                              *
*      JPD 9/12/90                                                             *
*******************************************************************************/

int closef_ptr( ptr )
	int *ptr;
{
	if(ptr > (int *) 0)
	{	if(*ptr >= 0)
			closef(*ptr);
		*ptr = -1;
	}
	return(0);
}
			
/*******************************************************************************
*    openfiles_list                                                            *
*           This program fills a list with the names of all currently open     *
*      files and their file pointers.                                          *
*                                                                              *
*      Parameters:       char *list_ptr   pointer to list of up to 60 filenames*
*                        long *ptr_ptr    pointer to list of up to 60 pileptrs *
*                                                                              *
*      Value returned:   number of files in list (0-60) or -1 for error.       *
*                                                                              *
*      JPD 9/18/90                                                             *
*******************************************************************************/

long openfiles_list( list_ptr, ptr_ptr )
	char list_ptr[][STRLIST];
	int  ptr_ptr[];
{
	long i;
	long count;

	for(i=0, count=0; i<60; i++)
	{	if(file_pointer[i] > 0)
		{	strncpy(list_ptr[count],open_fnames[i],STRLIST);
			ptr_ptr[count] = file_pointer[i];
			count++;
		}
	}

	return(count);
}
			
/*******************************************************************************
*	movefp		(see manual, Internal Functions)            	<in>   *
*			Repositions file pointer within a file.                *
*			                                                       *
*			Returns file pointer upon EOJ, else                    *
*			returns -1 upon error.                                 *
*			HLM&JPD030789					       *
*******************************************************************************/

int movefp( fd, offset, origin )
	int fd;
	long offset;	
	int origin;
{
	long i;

		/** check for & reset any blocks of reada data present **/
	for(i=0; i<5; i++)
	{	if(fileptr[i] == fd)
			readptr[i][0] = readptr[i][1];
	}

	return( lseek( fd, offset, origin ) );
}

/*******************************************************************************
*	readb		(see manual, Internal Functions)            	<in>   *
*			Reads a file in binary fashion.                        *
*			                                                       *
*			Returns greater than zero for successful read, else    *
*			returns zero when at end of file, else                 *
*			returns -1 upon error.                                 *
*			HLM&JPD030789					       *
*******************************************************************************/

int readb( string, size, num, fd )
	char *string;	
	int size, num;
	int fd;
{
/*  Locals  */
	long bytes;
	long i;

	bytes = size * num;

		/*** check for reada block ponters to same file **/
	for(i=0; i<5; i++)
	{	if(fileptr[i] == fd)
			break;
	}
	if(i < 5)
	{	readptr[i][0] = readptr[i][0] + bytes;
		if(readptr[i][1] < readptr[i][0])
			readptr[i][0] = readptr[i][1];
	}

	return( read( fd, string, bytes ) );
}



/*******************************************************************************
*	reada		(see manual, Internal Functions)            	<in>   *
*			Reads an ascii file using low level I/O.  When a       *
*			newline is read or if "bytes" number of bytes have     *
*			been read, a null is appended and the string returned. *
*			                                                       *
*			Reads file in blocks of 1024 characters for faster response   *
*			Heavily modified 1/2-3/91, JPD                         *
*			HLM&JPD030789					       *
*******************************************************************************/

char *reada( string, bytes, fd  )
	char *string;	
	int bytes;
	int fd;
{
	long brd;
	long p;

	char *ptr, *endptr;


		/*** init string ***/
	*string = (char) 0;

		/**** check for legal file pointer **/
	if(fd < 0)
		return((char *) 0);

		/**** see if file pointer is in list fileptr[], & add if necessary ****/
	for(p=0; p<5; p++)
	{	if(fileptr[p] == fd)
			break;
	}
	if(p > 4)						/*** not in list ***/
	{		/***  find an unused ptr ***/
		for(p=0; p<5; p++)
		{	if(fileptr[p] < 0)
				break;
		}

		if(p < 5)         /** found an unused pointer **/
		{	fileptr[p] = fd;
			readptr[p][0] = readptr[p][1];
		}
		else              /** all pointers in use    **/
			return(reada_old(string,bytes,fd));     /** read it the old way. **/
	}

		/**** transfer data from readblock to parameter string *****/
	ptr = string;
	endptr = ptr + bytes;
	for(;;)
	{		/**** make sure that there is enough data ***/
		if(readptr[p][0] >= readptr[p][1])
		{		/*** read another block of data ***/
			brd = read(fd, readblock[p],1024);
			if(brd <= 0)
				return(0); 
				/*** set pointers ***/
			readptr[p][0] = readblock[p];
			readptr[p][1] = readptr[p][0] + brd;
		}

			/**** transfer one character ****/
		*ptr = *readptr[p][0];
		readptr[p][0]++;
/*		readptr[p][0] = readptr[p][0] + 1;   */
		
			/**** check for lf ****/
		if(*ptr == '\n')  /** line feed (lf) **/
		{	ptr++;             /** move past lf & return string **/
			*ptr = (char) 0;
			break;
		}

			/**** check for bytes read ****/
		ptr++;
		if(ptr >= endptr)
		{	*ptr = (char) 0;
			break;
		}
	}

		/**** return ***/
	return( string );
}

/*******************************************************************************
*	reada		(see manual, Internal Functions)            	<in>   *
*			Reads an ascii file using low level I/O.  When a       *
*			newline is read or if "bytes" number of bytes have     *
*			been read, a null is appended and the string returned. *
*			Reads file one character at a time (slow due to heavy bus use JPD) *
*			                                                       *
*			Returns string upon normal end-of-job (EOJ), else      *
*			returns (char *)0 upon error.                          *
*			HLM&JPD030789					       *
*******************************************************************************/

char *reada_old( string, bytes, fd  )
	char *string;	
	int bytes;
	int fd;
{
	long i, brd;

	for(bytes--, i=0; (brd = read(fd, (string + i), 1)) && i < bytes; i++ )
	{	if( brd == -1 )
			return(0);
		if( *(string + i) == '\n' )
		{
			*(string + i + 1) = '\0';	/* null term string  */
			break;
		}
		else if( i == bytes)
		{
			*(string + i) = '\0';
			break;
		}
	}

	if( brd == 0 )
		return((char *) 0);
	else
		return( string );
}
