/*******************************************************************************
*       find_gap        (see manual, Internal Functions)                <in>   *
*                                                                              *
*    Purpose:    searches a square area around raster location x,y up to a     *
*    'radius' r.  Returns 1 if a logical connection can be made and sets       *
*    x's & y's accordingly.  Otherwise, returns 0.                             *
*                                                                              *
*    'Logical connection' is evaluated as follows:                             *
*    A.  Find nearest spur endpoint (spur1, direction v1 [normalized vector])  *
*        (if none found, return 0)                                             *
*    B.  Continue search for spurs within search radius, assign priority as    *
*        follows (low number = high priority):                                 *
*        1.  check for spur1-spurx connection touch, if touch then ignore spur.*
*        2.  calc v2 (spurx, direction vector, normalized)                     *
*        3.  calc v3 (spur1 to spurx vector, normalized)                       *
*        4.  assign priority = length(v1 - v3) + length(v2 + v3)               *
*        5.  Keep track of spur with best priority < 2 * sqrt(2)               *
*    C.  if best priority is < sqrt(2), then return(1) with connections to it. *
*    D.  else check extension of spur1 up to length search_radius/5 (min 3 pix)*
*        1.  if it connects, then take it as connection & return(1).           *
*    E.  else if best priority is < 2 * sqrt(2), return(1) with connections    *
*        to it.                                                                *
*    F.  else return(0).                                                       *
*                                                                              *
*                                                                              *
*    Arguments:    long center_x,center_y: coords of center of search square   *
*                  long spur[2][2]: array for coords of both spur endpoints    *
*                  long radius:  search limit (half of a side of max square)   *
*                  long show_msg:  flag for displaying messages.               *
*                                                                              *
*    Returned Values: (long) 1 = endpoints found, x1,y1,x2,y2 found location.  *
*                            0 = no logical endpoints found.                   *
*                                                                              *
*    Global Variables Modified:  none                                          *
*                                                                              *
*                      j.dabritz  2/27/90                                      *
*******************************************************************************/

#include <math.h>
#include "../../incl/globals.h"

long check_spur();
long get_location();
long is_pix_on();
long get_spur_dir();
long connect_touch();


long find_gap(center_x,center_y,spur,radius,show_msg)
	long center_x,center_y;
	long spur[][2];
	long radius;
	long show_msg;
{	
	long i;
	long x,y;
	long r1, r2;
	long hit;
	long state;
	long spur1_x, spur1_y;
	long spur2_x, spur2_y;

	
	double dx, dy;
	double dx1,dy1;
	double dx2,dy2;
	double dx3,dy3;
	double priority, best_priority;
	double length;
	double divisor;
	double max;
	double dummy_x, dummy_y;



		/***************** check for map present *********************/
	if(!job_stats[0])
		return(0);

		/***************** check input paramters *********************/
	if(radius < 0)
		return(0);

/*  A.  Find nearest spur endpoint (spur1, direction v1 [normalized vector])  */
/*      (if none found, return 0)                                             */
	for(r1=0, hit=0; r1<=radius ; r1++)
	{	for(r2=0; r2<=r1 ; r2++)
		{	for(state=0; state<8 && !hit; )
			{		/****** check location *****/
				get_location(center_x,center_y,r1,r2,state,&x,&y);
				if(check_spur(x,y))
				{
					spur1_x = x;
					spur1_y = y;
					get_spur_dir(spur1_x,spur1_y,&dx1,&dy1);
					hit = 1;
				}
					/****** increment state ****/
				if(!r1)
					state += 8;
				else if(state < 3)
					state++;
				else if((r2 == r1)  ||  !r2)
					state += 5;
				else
					state++;
			}
			if(hit)
				break;
		}
		if(hit)
			break;
	}

	if(!hit)
		return(0);

	spur[0][0] = spur1_x;
	spur[0][1] = spur1_y;

/*  B.  Continue search for spurs within search radius, assign priority as    */
/*     follows (low number = high priority):                                  */
/*     1.  check for spur1-spurx connection touch, if touch then ignore spur. */
/*      2.  calc v2 (spurx, direction vector, normalized)                     */
/*      3.  calc v3 (spur1 to spurx vector, normalized)                       */
/*      4.  assign priority = length(v1 - v3) + length(v2 + v3)               */
/*      5.  Keep track of spur with best priority < 2 * sqrt(2)               */

	best_priority = (double) 3.1;
	for(hit=0; r1<=radius ; r1++)
	{	for(; r2<=r1 ; r2++)
		{	for( ; state<8 ; )
			{		/****** check location *****/
				get_location(center_x,center_y,r1,r2,state,&x,&y);
				if(check_spur(x,y))
				{		/*** check for touch ***/
					if(!connect_touch(spur1_x,spur1_y,x,y,0))
					{		/**** get spur direction ****/
						get_spur_dir(x,y,&dx2,&dy2);
							/**** get connection vector ***/
						dx3 = x - spur1_x;
						dy3 = y - spur1_y;
							/**** normalize ***************/
						length = sqrt((dx3 * dx3) + (dy3 * dy3));
						dx3 = dx3 / length;
						dy3 = dy3 / length;
                     /**** assign priority *********/
							/*** check for single pixel **/
						if((dx2==0.0) && (dy2==0.0))        /** single point **/
						{	if((dx1==0.0) && (dy1==0.0))     /** #1 is single pt. too*/
							{	if(hit)
									priority = 2.0 * sqrt((double) (r1 * r1) + (r2 * r2)) / (sqrt(2.0)) / ((double) radius);
								else
									priority = 0.0;
							}
							else
							{	dx = dx1 - dx3;
								dy = dy1 - dy3;
								priority = 2.0 * sqrt((dx * dx) + (dy * dy));
							}
						}
						else
						{	if((dx1==0.0) && (dy1==0.0))     /** #1 is single pt. */
							{	dx = dx2 + dx3;
								dy = dy2 + dy3;
								priority = 2.0 * sqrt((dx * dx) + (dy * dy));
							}
							else
							{	dx = dx1 - dx3;
								dy = dy1 - dy3;
								priority = sqrt((dx * dx) + (dy * dy));
								dx = dx2 + dx3;
								dy = dy2 + dy3;
								priority += sqrt((dx * dx) + (dy * dy));
							}
						}

							/**** track spur with best priority ***/
						if(priority < best_priority)
						{	best_priority = priority;
							spur2_x = x;
							spur2_y = y;
							hit = 1;
						}
					}
				}
					/****** increment state ****/
				if(!r1)
					state += 8;
				else if(state < 3)
					state++;
				else if((r2 == r1)  ||  !r2)
					state += 5;
				else
					state++;
			}
			state = 0;
		}
		r2 = 0;
	}


/*  C.  if best priority is < sqrt(2), then return(1) with connections to it. */
	if(best_priority < (double) 1.7)
	{	spur[1][0] = spur2_x;
		spur[1][1] = spur2_y;
		return(1);
	}


/*  D.  else check extension of spur1 up to length radius/2 */
/*      1.  if it connects, then take it as connection & return(1).           */
	if(fabs(dx1) > fabs(dy1))
		divisor = fabs(dx1);
	else
		divisor = fabs(dy1);

	if(divisor > 0.0)
	{	dx1 = dx1 / divisor;
		dy1 = dy1 / divisor;
	
		max =  radius / 2.0;
		if(max < 5)
			max = 5;
		if(radius < max)
			max = radius;
	
		dummy_x = spur1_x + 0.5;
		dummy_y = spur1_y + 0.5;
		for(i=1; i<= max; i++)
		{	dummy_x += dx1;
			x = dummy_x;
			dummy_y += dy1;
			y = dummy_y;
			if((x < 0) || (x > job_stats[1]) || (y < 0) || (y > job_stats[2]))
				break;
			if(connect_touch(spur1_x,spur1_y,x,y,1))
			{	spur[1][0] = x;
				spur[1][1] = y;
				return(1);
			}
		}
	}

/*  E.  else if best priority is < 2 * sqrt(2), return(1) with connections    */
/*      to it.                                                                */
	if(best_priority < (double) 3.1)
	{	spur[1][0] = spur2_x;
		spur[1][1] = spur2_y;
		return(1);
	}




/*  F.  else return(0).                                                       */

	if(show_msg)
		message("couldn't find a good connection within range.");
	return(0);
}




/*******************************************************************************
*       check_spur      (see manual, Internal Functions)                <in>   *
*                                                                              *
*    Purpose:    checks for a spur location.  returns 1 if it is found.        *
*                                                                              *
*    Arguments:    long x,y:   raster coordinates of bit posision              *
*                                                                              *
*    Returned Values: (long) 1 = spur end at location x,y                      *
*                            0 = no spur at location x,y                       *
*                                                                              *
*    Global Variables Modified:  none                                          *
*                                                                              *
*                      j.dabritz  2/27,90                                      *
*******************************************************************************/


long check_spur(x, y)
	long x,y;
{	
	long locale;
	long word, bit;

		/**************** check input *************************************/
	if((x<0) || (x>job_stats[1]) || (y<0) || (y>job_stats[2]) )
		return(0);

		/**************** calculate word and bit **************************/
	word = x >> 5;
	bit = (7 & ~x) | (x & 24);

		/**************** check bit ***************************************/
	if(!(lraster[y][word] & (1 << bit)))
		return(0);

	locale = locale_xy(x,y);
	if(tlocale[locale] & (3 << 8))
		return(1);
	else
		return(0);
}


/*******************************************************************************
*  get_location                                                                *
*          This program sets a location (x,y) given a center, a primary radius *
*  (r1), and a secondary radius(2), and a state (0-7)                          *
*                                                                              *
*  Parameters:  long center_x, center_y;  raster coords of center              *
*               long r1;                  primary radius (out from center)     *
*               long r2;                  secondary radius (from end of primary*
*               long state;               0=up,left, 1=right,up, 2=down,right  *
*                                         3=left,down, 4=up,right, 5=right,down*
*                                         6= down,left, 7=left,up              *
*               long *x, *y;              ptrs to values returned.             *
*  Values returned:  long (0)                                                  *
*******************************************************************************/

long get_location(center_x,center_y,r1,r2,state,x,y)
long center_x,center_y;
long r1,r2;
long state;
long *x, *y;
{
		/*** initialize return values ***/
	*x = center_x;
	*y = center_y;

		/*** add primary radius distance **/
	switch(state % 4) {
	case 0:  *y += r1;
		break;
	case 1:  *x += r1;
		break;
	case 2:  *y -= r1;
		break;
	case 3:  *x -= r1;
		break;
	}

		/*** add secondary radius distance **/
	switch(state) {
	case 0: case 6:  *x -= r2;
		break;
	case 1: case 7:  *y += r2;
		break;
	case 2: case 4:  *x += r2;
		break;
	case 3: case 5:  *y -= r2;
		break;
	}

	return(0);
}








/*******************************************************************************
*  connect_touch                                                               *
*          This program checks a possible line location to see if it touches   *
*  another pixel along its length (except beginning & ending pixels)           *
*                                                                              *
*  Parameters:  long x1,y1;    one endpoint                                    *
*               long x2,y2;    another endpoint.                               *
*               long check_end;  flag, 1=check line end2, 0=don't check end2   *
*  Values returned:  long 0 = line on raster & no touch                        *
*                         1 = line off raster or touching another pixel        *
*******************************************************************************/

long connect_touch(x1,y1,x2,y2,check_end)
long x1,y1;
long x2,y2;
long check_end;
{
	long dx,dy;
	long e,e1,e2;
	long i;
	long x,y;
	long x0,y0;



		/*********   check for on raster ***************************/
	if((x1 < 0) || (x2 < 0) || (x1 > job_stats[1]) || (x2 > job_stats[1]) ||
	   (y1 < 0) || (y2 < 0) || (y1 > job_stats[2]) || (y2 > job_stats[2]))
		return(1);

		/*********   set x0, y0 *******************************************/
	x0 = x1;
	y0 = y1;

		/*********   calc dx and switch points to make it non-negative.   */
	dx = x2 - x1;
	if(dx < 0) 
	{	i = x1;
		x1 = x2;
		x2 = i;
		i = y1;
		y1 = y2;
		y2 = i;
		dx = x2 - x1;
	}
		/*********   calc cy  */
	dy = y2 - y1;

		/*********   set x,y */
	x = x1;
	y = y1;

		/*********   divide situation up into cases ****************/
	if(dy >= 0)              /* line has positive slope */
	{	if(dx > dy)          /* slope is < 1  */
		{	e  = (dy * 2) - dx;
			e1 = (dy - dx) * 2;
			e2 = dy * 2;
			for(i = -1; i != dx; i++, x++)
			{	if((check_end && ((x != x0) || (y != y0))) || (((x != x1) || (y != y1)) && ((x != x2) || (y != y2))))
				{
					if(is_pix_on(x,y))
						return(1);
					if(is_pix_on(x,y-1))
						return(1);
					if(is_pix_on(x,y+1))
						return(1);
				}
				if(e > 0)
				{	if(((i + check_end) != -1) && is_pix_on(x+1,y-1))
						return(1);
					y++;
					if((i+2-check_end) == dx)
						break;
					if(is_pix_on(x,y+1))
						return(1);
					e += e1;
				}
				else
					e += e2;
			}
		}
		else                  /* slope is > 1    */
		{	e  = (dx * 2) - dy;
			e1 = (dx - dy) * 2;
			e2 = dx * 2;
			for(i = -1; i != dy; i++, y++)
			{	if((check_end && ((x != x0) || (y != y0))) || (((x != x1) || (y != y1)) && ((x != x2) || (y != y2))))
				{
					if(is_pix_on(x,y))
						return(1);
					if(is_pix_on(x-1,y))
						return(1);
					if(is_pix_on(x+1,y))
						return(1);
				}
				if(e > 0)
				{	if(((i + check_end) != -1) && is_pix_on(x-1,y+1))
						return(1);
					x++;
					if((i+2-check_end) == dy)
						break;
					if(is_pix_on(x+1,y))
						return(1);
					e += e1;
				}
				else
					e += e2;
			}
		}
	}
	else                     /* slope is negative  */
	{	if(dx > -dy)          /* slope is > -1  */
		{	e  = ((-dy) * 2) - dx;
			e1 = ((-dy) - dx) * 2;
			e2 = (-dy) * 2;
			for(i = -1; i != dx; i++, x++)
			{	if((check_end && ((x != x0) || (y != y0))) || (((x != x1) || (y != y1)) && ((x != x2) || (y != y2))))
				{
					if(is_pix_on(x,y))
						return(1);
					if(is_pix_on(x,y-1))
						return(1);
					if(is_pix_on(x,y+1))
						return(1);
				}
				if(e > 0)
				{	if(((i + check_end) != -1) && is_pix_on(x+1,y+1))
						return(1);
					y--;
					if((i+2-check_end) == dx)
						break;
					if(is_pix_on(x,y-1))
						return(1);
					e += e1;
				}
				else
					e += e2;
			}
		}
		else                  /* slope is < -1   */
		{	e  = (dx * 2) + dy;
			e1 = (dx + dy) * 2;
			e2 = dx * 2;
			for(i = 1; i != dy; i--, y--)
			{	if((check_end && ((x != x0) || (y != y0))) || (((x != x1) || (y != y1)) && ((x != x2) || (y != y2))))
				{
					if(is_pix_on(x,y))
						return(1);
					if(is_pix_on(x-1,y))
						return(1);
					if(is_pix_on(x+1,y))
						return(1);
				}
				if(e > 0)
				{	if(((i + check_end) != 1) && is_pix_on(x-1,y-1))
						return(1);
					x++;
					if((i-2+check_end) == dy)
						break;
					if(is_pix_on(x+1,y))
						return(1);
					e += e1;
				}
				else
					e += e2;
			}
		}
	}

	return(0);
}



/*******************************************************************************
*   is_pix_on(x,y)                                                             *
*       returns 1 if point is on raster & pixel is 'on'.                       *
*       otherwise returns 0.                                                   *
*******************************************************************************/

long is_pix_on(x,y) 
long x,y;
{
	long word, bit;

		/**** check to see if on raster ***/
	if((x < 0) || (x > job_stats[1]) || (y < 0) || (y > job_stats[2]))
		return(0);

		/**** check to see if pixel is on **/
	word = x >> 5;
	bit = (7 & ~x) | (x & 24);
	if(lraster[y][word] & (1 << bit))
		return(1);

		/**** return ***/
	return(0);
}



/*******************************************************************************
*   get_spur_dir                                                               *
*        This program finds the direction of spur ending pixels (up to 3)      *
*   and normalizes a vector pointing in that direction.                        *
*                                                                              *
*   values returned:  long(0);                                                 *
*                                                                              *
*******************************************************************************/

long get_spur_dir(x0,y0,dx_ptr,dy_ptr)
long x0,y0;
double *dx_ptr, *dy_ptr;
{
	long x,y;
	long dir, dir0;
	long i;
	long locale;

	static long move_x[8] = { 0, 1, 1, 1, 0, -1, -1, -1 };
	static long move_y[8] = { 1, 1, 0, -1, -1, -1, 0, 1 };

	double dx, dy;
	double length;


		/************ find first direction back ***************************/
	locale = locale_xy(x0,y0);
	if(!locale)
	{	*dx_ptr = 0.0;
		*dy_ptr = 0.0;
		return(0);
	}
	for(dir0=0; !(tlocale[locale] & (1 << dir0)); dir0++);

		/************ continue back 2 more pixels if within 1 of same dir */
	x = x0 + move_x[dir0];
	y = y0 + move_y[dir0];
	for(i=0, dir=dir0; i<2; i++)
	{	locale = locale_xy(x,y);
		if(!(tlocale[locale] & (1 << 13)))   /** not a node     **/
		{	for(dir=(dir + 6) % 8 ; !(tlocale[locale] & (1<<dir)); dir=(dir+1)%8);
			if((dir == dir0) || (dir == (dir0 + 1) % 8) || (dir == (dir0+7) % 8))
			{	x += move_x[dir];
				y += move_y[dir];
			}
			else
				break;
		}
		else
			break;
	}
	
		/************ get vector ******************************************/
	dx = x0 - x;
	dy = y0 - y;
	
		/************ normalize *******************************************/
	length = sqrt((dx * dx) + (dy * dy));
	if(length != (double) 0.0)
	{	dx = dx / length;
		dy = dy / length;
	}
	*dx_ptr = dx;
	*dy_ptr = dy;

		/************ return **********************************************/
	return(0);
}
