/*******************************************************************************
*       find_vnode2     (see manual, Internal Functions)                <in>   *
*    Searches for a vector node OR probable node location as follows:          *
*       1.  If a node is found within search radius, its  location is set.     *
*       2.  Else if a segment/poly endpoint is found within search radius, its *
*           location is set.                                                   *
*       3.  Else, if vertex common to 3+ vect obj & 2+ diverge, it is set.     *
*       4.  Else, if vertex common to 2 vect obj which diverge, it is set.     *
*       5.  Else, if a vertex is found within r, its location is set.          *
*                                                                              *
*    Arguments:    double x,y,r:  raster coordinates and radius (pixels)       *
*                  double *xr,*yr;      returned values for x,y,r found.       *
*                  long l1, l2;   levels (32-95) range of levels to check.     *
*                                                                              *
*    Returned Values: (long) 1-5 = node type found, as listed above            *
*                            0 = no node found                                 *
*                           -1 = error.                                        *
*                                                                              *
*    Global Variables Modified:  none.                                         *
*                                                                              *
*                      j.dabritz  4/8/91                                       *
*******************************************************************************/

#include "../../incl/globals.h"
#include <math.h>

long count_hits();


long find_vnode2(x,y,r,xr,yr,l1,l2)
double x,y,r;
double *xr,*yr;
long l1,l2;
{	
	INDEX_LIST index_list;
	V_OBJECTS  v_object;
	COORD_LIST coord_list;

	long ret_val;
	long x1,x2,y1,y2;
	long row,col;
	long trial_type;
	long increment;
	long level;
	long v,v1,v2;
	long hits;

	double min_square;
	double dx,dy;
	double dist_squared;
	double r_squared;


		/*********** check for map **************************************/
	if(!job_stats[0])
		return(-1);

		/*********** check input params ******************************/
	if((l1 > l2) || (l1 < 32) || (l2 > 95) || (r < 0.0))
	{	message("find_vnode2:  error, bad input parameter.");
		return(-1);
	}

		/*********** init ret_val ************************************/
	ret_val = 6;

		/*********** set index_array limits of search ****************/
		/*********** find index_array window ****************************/
	x1 = (x - r + 0.5);    /* x min (pixels) */
	x1 = x1 >> 5;          /* x min (index_array coords) */
	x2 = (x + r + 0.5);    /* x max (pixels) */
	x2 = x2 >> 5;          /* x max (index_array coords) */
	y1 = (y - r + 0.5);    /* y min (pixels) */
	y1 = y1 >> 5;          /* y min (index_array coords) */
	y2 = (y + r + 0.5);    /* y max (pixels) */
	y2 = y2 >> 5;          /* y max (index_array coords) */

		/** test for raster coverage *****************************/
	if((x2 < 0)  || (x1 > (job_stats[1] >> 5)))
		return(-1);
	if((y2 < 0)  || (y1 > (job_stats[2] >> 5)))
		return(-1);

		/** limit to range of index_array ***/
	if(x1 < 0) 
		x1 = 0;
	if(y1 < 0)
		y1 = 0;
	if(x2 > (job_stats[1] >> 5))
		x2 = (job_stats[1] >> 5);
	if(y2 > (job_stats[2] >> 5))
		y2 = (job_stats[2] >> 5);
	
		/*********** check for nodes & endpoints *********************/
		/** for each row & col of index array, search objects */
	min_square = r * r + 0.000001;       /** compare squares to save time **/
	for(row = y1; row <= y2; row++)
	{	for(col = x1; col <= x2; col++)
		{
				/** search all indicated objects **/
			index_list.next = index_array[row][col];
			for(; index_list.next;)
			{		/** get index_list record **/
				if(get_indexl(&index_list,index_list.next))
					return(-1);

					/*** check level **/
				level = index_list.type / 10;
				if((level >= l1) && (level <= l2))
				{
      				/*********** get vector object ************************/
   				if(get_vobjects(&v_object,index_list.obj_rec))
      				return(-1);

						/*** check for node or endpoint vertex     ***/
					v1 = v_object.v1;
					if((v_object.type % 10) > 0)
					{	v2 = v_object.v2;     /** segment or poly **/
						trial_type = 2;
					}
					else
					{	v2 = v_object.v1;     /** node **/
						trial_type = 1;
					}
	
					increment = v2 - v1;
	
					for(v = v1 ; ret_val >= trial_type ; v += increment)
					{		/*** read vertex record for vertex         ***/
						if(get_v_coord(&coord_list,v))
							return(-1);
						dx = x - coord_list.x;
						dy = y - coord_list.y;
						dist_squared = (dx * dx) + (dy * dy);
							
							/*** check against min dist ***/
						if(dist_squared < min_square)
						{	min_square = dist_squared;
							*xr = coord_list.x;
							*yr = coord_list.y;
							ret_val = trial_type;
						}
	
						if(v == v2)
							break;
					}
				}
			}
		}
	}

		/*********** if none found, check for repete verticies *******/
	if(ret_val == 6)   /** no node or seg endpoint found **/
	{		/** reset default return value **/
		ret_val = 0;   /* none found */
		r_squared = r * r;                  /** compair squares to save time **/
		for(row = y1; row <= y2; row++)
		{	for(col = x1; col <= x2; col++)
			{
					/** search all indicated objects **/
				index_list.next = index_array[row][col];
				for(; index_list.next;)
				{		/** get index_list record **/
					if(get_indexl(&index_list,index_list.next))
						return(-1);
	
						/*** check level **/
					level = index_list.type / 10;
					if((level >= l1) && (level <= l2))
					{	v1 = index_list.v1;
						v2 = index_list.v2;
						if(v1 == v2)
							increment = 0;
						else if((v2 - v1) > 0)
							increment = 1;
						else
							increment = -1;
	
							/*** check all verticies ***/
						if(increment)
						{	for(v = v1 ; ; v += increment)
							{		/*** read vertex record for vertex         ***/
								if(get_v_coord(&coord_list,v))
									return(-1);

									/*** check dist < r_squared  *****/
								dx = x - coord_list.x;
								dy = y - coord_list.y;
								dist_squared = (dx * dx) + (dy * dy);
								if(dist_squared < r_squared)
								{		/*** count multiple hits     *****/
									hits = count_hits(index_list.next,row,y2,col,x2,l1,l2,coord_list.x,coord_list.y);
									if(hits < 0)
										return(-1);
										/* see if type is as good or same & dist better */
									trial_type = 5 - hits;
									trial_type = (trial_type < 3) ?  3 : trial_type;
									if((trial_type < ret_val) ||
										((trial_type == ret_val) && 
								    	(dist_squared < min_square)))
									{		/*** set new best obj ****/
										min_square = dist_squared;
										*xr = coord_list.x;
										*yr = coord_list.y;
										ret_val = trial_type;
									}
								}
				
								if(v == v2)
									break;
							}
						}
					}
				}
			}
		}
	}
	
		/*********** return ******************************************/
	return(ret_val);
}


/*******************************************************************************
*  count_hits                                                                  *
*     This program counts the number of vector object segments on levels l1-l2 *
*  which have verticies exactly matching x,y.  The search will start at the    *
*  index_list record nx, and will continue through that linked-list and any    *
*  others in index_array cells from row rw through y2 and columns cl through   *
*  x2.  After hits are counted, the vertex is checked for divergence along     *
*  the vector objects.  If no divergence is found, then hits returned will be  *
*  not greater than 1.                                                         *
*                                                                              *
*      Parameters: long nx;  index_list record number of start point.          *
*                  long rw,y2; index_array begin and end row numbers to search *
*                  long cl,x2; index_array begin and end col numbers to search *
*                  long l1,l2; range of level numbers to be searched           *
*                  float x,y;  coordinate to exactly match.                    *
*      Value returned: 0+ = no. objects with a match,  -1=error.               *
*******************************************************************************/
long count_hits(nx,rw,y2,cl,x2,l1,l2,x,y)
long nx,rw,y2,cl,x2,l1,l2;
float x,y;
{
	INDEX_LIST index_list;
	COORD_LIST coord_list;

	long hits;
	long row,col;
	long increment;
	long v1,v2,v;
	long level;
	long hit_list[20];  /** vertex no list for 20 obj **/
	long i,j;

	double unit_vector[20][2][2];
	double dx,dy,d1,d2;
	double dist;


		/*** init hits (number of objects with a matching vertex) ***/
	hits = 0;
	
	for(row = rw; row <= y2; row++)
	{	for(col = cl; col <= x2; col++)
		{
				/** search all indicated objects **/
			if((row != rw) || (col != cl))
				index_list.next = index_array[row][col];
			else
				index_list.next = nx;

			for(; index_list.next;)
			{		/** get index_list record **/
				if(get_indexl(&index_list,index_list.next))
					return(-1);

					/*** check level **/
				level = index_list.type / 10;
				if((level >= l1) && (level <= l2))
				{		/*** check for matching vertex ***/
					v1 = index_list.v1;
					v2 = index_list.v2;

					increment = ((v2 - v1) >= 0) ?    1 : -1;
	
					for(v = v1 ;  ; v += increment)
					{		/*** read vertex record for vertex         ***/
						if(get_v_coord(&coord_list,v))
							return(-1);
						if((coord_list.x == x) && (coord_list.y == y))
						{		/** add object & vertex number to hit_list **/
							if(hits < 20)
								hit_list[hits] = v;
							hits++;
							break;
						}

						if(v == v2)
							break;
					}
				}
			}
		}
	}

		/**** check for divergence from point ****/
	if(hits > 1)
	{	if(hits > 20)
			hits = 20;
			/*** find unit vectors for directions out of verticies ***/
		for(i=0; i<hits; i++)
		{		/** find vertex back **/
			for(v = hit_list[i] -1; v>=0; v--)
			{	if(get_v_coord(&coord_list,v))
					return(-1);
				if((coord_list.x != x) || (coord_list.y != y))
					break;
			}
				/** get unit vector **/
			dx = coord_list.x - x;
			dy = coord_list.y - y;
			dist = sqrt(dx * dx  +  dy * dy);
			if(dist > 0.0)
			{	unit_vector[hits][0][0] = dx / dist;
				unit_vector[hits][0][1] = dy / dist;
			}
			else
				return(-1);

				/** find vertex ahead **/
			for(v = hit_list[i] +1; v<job_stats[18]; v++)
			{	if(get_v_coord(&coord_list,v))
					return(-1);
				if((coord_list.x != x) || (coord_list.y != y))
					break;
			}
				/** get unit vector **/
			dx = coord_list.x - x;
			dy = coord_list.y - y;
			dist = sqrt(dx * dx  +  dy * dy);
			if(dist > 0.0)
			{	unit_vector[hits][1][0] = dx / dist;
				unit_vector[hits][1][1] = dy / dist;
			}
			else
				return(-1);
		}

			/*** compare unit vectors for divergence (dist) > .01 ***/
		for(i=0; i<hits-1; i++)
		{	for(j=i+1; j<hits; j++)
			{		/*** check unit vectors 0:0 & 1:1 ***/
				dx = fabs(unit_vector[i][0][0] - unit_vector[j][0][0]);
				dy = fabs(unit_vector[i][0][1] - unit_vector[j][0][1]);
				d1 = sqrt(dx * dx + dy * dy);
				dx = fabs(unit_vector[i][1][0] - unit_vector[j][1][0]);
				dy = fabs(unit_vector[i][1][1] - unit_vector[j][1][1]);
				d2 = sqrt(dx * dx + dy * dy);
				if((d1 > .01) || (d2 > .01))
				{		/*** check unit vectors 0:1 & 1:0 ***/
					dx = fabs(unit_vector[i][0][0] - unit_vector[j][1][0]);
					dy = fabs(unit_vector[i][0][1] - unit_vector[j][1][1]);
					d1 = sqrt(dx * dx + dy * dy);
					dx = fabs(unit_vector[i][1][0] - unit_vector[j][0][0]);
					dy = fabs(unit_vector[i][1][1] - unit_vector[j][0][1]);
					d2 = sqrt(dx * dx + dy * dy);
					if((d1 > .01) || (d2 > .01))
						return(hits);
				}
			}
		}
			/*** no divergence, return 1 ***/
		return(1);
	}

	return(hits);
}
