/*******************************************************************************
*       find_vpoly      (see manual, Internal Functions)                <in>   *
*    Searches the index_array / index_list or objects files for a poly         *
*    within r pixels of the input x,y location which has the input level.       *
*    If any are found, the record number of the closest one is returned.       *
*                                                                              *
*    Arguments:    float x,y:     raster coordinates of interior point         *
*                  long level:    0-31 map objects (objects rec # returned)    *
*                                 32-95 = vector objects(v_objects rec# ret.)  *
*                                                                              *
*    Returned Values: (long) n = record number of poly (n >= 0), or            *
*                           -1 = no poly found or error.                       *
*                                                                              *
*    Global Variables Modified:  none.                                         *
*                                                                              *
*                      j.dabritz  8/2/89                                       *
*******************************************************************************/

#include "../../incl/globals.h"
#include <math.h>

struct {
	double x, y;
	double dist;
	double ext_dist;
	long p;
	} 		interior, exterior;

double cent_x, cent_y;
long cx, cy;             /** current index_array coordinates            **/


long find_vpoly(x,y,level)
float x,y;
long level;
{	
	long check_cell();

	long cell_x, cell_y;     /** cell coordinates of starting point         **/
	long x_max, y_max;       /** max cell values of index_array             **/
	long x_inc, y_inc;       /** increment values for searching index array **/
	long p;                  /** polygon number (returned)                  **/

		/************ check for map & index_list ***********************/
	if((!job_stats[0]) || (!job_stats[15]))
		return(-1);

		/*********** check for file type needed *************************/
	if((level >= 0) && (level < 32))  /* objects file record wanted   */
	{	if(!job_stats[7])
			return(-1);
		if(job_stats[49])        /*  map needs to be re-indexed */
		{	if(index_map())      /*  1 = index failed           */
				return(-1);
		}
	}
	else if((level >= 32) && (level < 96))  /* v_objects file record wanted */
	{	if(job_stats[15] == 0)
			return(-1);
	}
	else              /* illegal level, abort          */
		return(-1);

		/************ find index_array coords of input point ***********/
	cell_x = (x + .5);
	cell_x = cell_x >> 5;
	cell_y = (y + .5);
	cell_y = cell_y >> 5;

		/************ set up index_array cell search pattern variables */
	x_max = job_stats[1] >> 5;
	y_max = job_stats[2] >> 5;
	cx = cell_x;
	cy = cell_y;
	x_inc = -1;
	y_inc = -1;

		/************ search array until poly found or no cells left ***/
		/** search pattern starts at cell containing point.  then goes 
			down along a single cell column.  When the bottom of the map
			is reached, the search goes up from the cell above the starting
			cell.  When a column is finished the column to the left is 
			searched in the same manner.  When the left edge of the map
			is searched, the next column searched (same manner) is the one
			to the right of the initial cell.                   ***/
	p = -1;  /* default polygon number (flag for not found)  **/
	interior.p = -1;
	exterior.p = -1;
	cent_x = x;           /*  for passing in global space as a double */
	cent_y = y;           /*  for passing in global space as a double */

	for(;; cy += y_inc)
	{		/** see if cell is on map **/
		if(y_inc < 0)   
		{	if(cy < 0)   /** if off bottom of index_array, start up **/
			{	cy = cell_y + 1;
				y_inc = 1;
			}
		}
		if(y_inc > 0)
		{	if(cy > y_max)  /** if off top, move over & start down **/
			{	cx += x_inc;
				cy = cell_y;
				y_inc = -1;
			}
		}
		if(x_inc < 0)      /** if off left, move back & start right **/
		{	if(cx < 0)
			{	cx = cell_x + 1;
				x_inc = 1;
			}
		}
		if(x_inc > 0)
		{	if(cx > x_max)  /** if off right, array is all searched, done **/
				break;        /** all cells have been searched */
		}
			/*** if cell is on map, then search it. ****/
		if((cx >= 0) && (cx <= x_max) && (cy >=0) && (cy <= y_max))
			p = check_cell(level);   /* returns polygon number */

		if(p == -2)
			return(-1);  /* not inside any polygons */

		if(p >=0)
			break;       /* found inside polygon    */
	}

	return(p);
}


/*******************************************************************************
*        check_cell                                                            *
*            this program checks one index_array cell for polygon verticies.   *
*        If there are polygon verticies, then it finds the closest interior    *
*        polygon & returns its record number.  If no polygons are found, it    *
*        returns -1,  if an exterior polygon is closest, then -2 is returned.  *
*                                                                              *
*******************************************************************************/

long check_cell(level)
	long level;
{
	OBJECTS object;
	INDEX_LIST index_list;
	COORD_LIST coord_list;

	long i, v;
	long go_on;

	double xmin,xmax,ymin,ymax;         /* limits of index_array cell (pix) */
	double vx, vy;                      /* vertex x & y (pixels)            */
	double last_vx, last_vy;            /* last vertex x & y                */
	double x1,x2,y1,y2;                 /* clipped vertex points            */
	double dx,dy;                       /* differences */
	double length;                      /* vector length */
	double array[2][3];                 /* for intersection solutions       */
	double dist;                        /* dist from vertex pair to point   */
	double ext_dist;                    /* dist that vector must be extended
											to be perpendicularly adjacent 
											to the point                    */
	double intersect_x, intersect_y;    /* raster (pixel) coords of closest
											point of clipped segment to pt. */



		/*********** find limits of cell window ****************************/
	xmin = cx * 32.0 - 0.5;
	xmax = xmin + 32.0;
	ymin = cy * 32.0 - 0.5;
	ymax = ymin + 32.0;

		/*********** search for polygon verticies within cell **************/
	index_list.next = index_array[cy][cx];
	for(; index_list.next;)
	{		/** get index_list record **/

		if(get_indexl(&index_list,index_list.next))
			return(-1);
		  /*** check for object type == polygon (seg for map type) ***/
		go_on = 1;
		if(level < 32)                /* objects file records */
		{	if(index_list.type != 1)
				go_on = 0;
			else if(level > 0)
			{	if(get_objects(&object,index_list.obj_rec))
					go_on = 0;
				else
				{	if(!(object.i12[4] & (1 << level)))
						go_on = 0;
				}
			}
		}
		else            /* v_objects file records */
		{
			if((index_list.type/10 != level) || ((index_list.type % 10) != 2))
				go_on = 0;
		}

		if(go_on)
		{
				/**** get initial vertex ****/
			v = index_list.v1;
			if(level < 32)    /* objects file record wanted   */
			{	if(get_coordlst(&coord_list,v))
					return(-1);
			}
			else              /* v_objects file record wanted */
			{	if(get_v_coord(&coord_list,v))
					return(-1);
			}
			last_vx = coord_list.x;
			last_vy = coord_list.y;
	
				/**** test every side for distance & extension_dist **/
			i = (index_list.v1 < index_list.v2)  ?  1 : -1;
			for(v = index_list.v1 + i;   ; v += i)
			{	/*** get next vertex coords ***/
				if(level < 32)    /* objects file record wanted   */
				{	if(get_coordlst(&coord_list,v))
						return(-1);
				}
				else              /* v_objects file record wanted */
				{	if(get_v_coord(&coord_list,v))
						return(-1);
				}
				vx = coord_list.x;
				vy = coord_list.y;
	
					/*** clip verticies to cell window ***/
				x1 = last_vx;
				y1 = last_vy;
				x2 = vx;
				y2 = vy;
				if(!clipd(&x1,&y1,&x2,&y2,xmin,xmax,ymin,ymax))
				{
						/*** if not pts not equal, then get dist to polygon **/
					if((x1 != x2)  ||  (y1 != y2))
					{	dx = x2 - x1;
						dy = y2 - y1;
							/** normalize vector **/
						length = sqrt( dx*dx + dy*dy );
						dx = dx / length;
						dy = dy / length;
	
						if((dx != 0.0) || (dy != 0.0))
						{	array[0][0] = dx;
							array[0][1] = dy;
							array[0][2] = cent_x - x1;
							array[1][0] = dy;
							array[1][1] = -dx;
							array[1][2] = cent_y - y1;
							if(!mat_solve(2,3,array))
							{		/*** set dist & extension distance      **/
								if((array[0][2] >= 0.0) && (array[0][2] <= length))
								{		/** pt is perpendicular to segment **/
									dist = fabs(array[1][2]);
									ext_dist = 0.0;
									intersect_x = x1 + array[0][2] * dx;
									intersect_y = y1 + array[0][2] * dy;
								}
								else    /** point is off one end of segment **/
								{	if(array[0][2] > length)  /** past end **/
									{	dx = cent_x - x2;
										dy = cent_y - y2;
										dist = sqrt( dx*dx + dy*dy );
										ext_dist = array[0][2] - length;
										intersect_x = x2;
										intersect_y = y2;
									}
									else                      /** befor begining **/
									{	dx = cent_x - x1;
										dy = cent_y - y1;
										dist = sqrt( dx*dx + dy*dy );
										ext_dist = -array[0][2];
										intersect_x = x1;
										intersect_y = y1;
									}
								}
	
									/** check to see if closer than previous 
										polys of same type (interior/exterior) **/
									/** note fuzzy tolerance of .001 pixel **/
								if(level < 32)  /** map segment/poly verticies **/
								{		/** check int set, (the dist &ext are =) */
									if((dist < interior.dist-.001)  || 
								 	((dist < interior.dist + .001) &&
								  	(ext_dist < interior.ext_dist)) ||
								  	(interior.p == -1) )
									{		/** get segment object record **/
										get_objects(&object,index_list.obj_rec);
										if(array[1][2] > 0.0)  /** pt on rt **/
										{	interior.p = object.i12[7];
											exterior.p = object.i12[8];
										}
										else                   /** pt on lt **/
										{	interior.p = object.i12[8];
											exterior.p = object.i12[7];
										}
										interior.x = intersect_x;
										interior.y = intersect_y;
										interior.dist = dist;
										interior.ext_dist = ext_dist;
										exterior.x = intersect_x;
										exterior.y = intersect_y;
										exterior.dist = dist;
										exterior.ext_dist = ext_dist;
									}
								}
								else  /** level is import or vector (not map) **/
									/**  polygons listed seperate from segs &
										each other (but always on rt side) **/
								{	if(array[1][2] > 0.0)  /* rt = inside polygon */
									{	if((dist < interior.dist-.001)  || 
								 		((dist < interior.dist + .001) &&
								  		(ext_dist < interior.ext_dist)) ||
								  		(interior.p == -1) )
										{	interior.p = index_list.obj_rec;
											interior.x = intersect_x;
											interior.y = intersect_y;
											interior.dist = dist;
											interior.ext_dist = ext_dist;
										}
									}
									else  /** lt = outside polygon **/
									{	if((dist < exterior.dist-.001)  || 
								 		((dist < exterior.dist + .001) &&
								  		(ext_dist < exterior.ext_dist)) ||
								  		(exterior.p == -1) )
										{	exterior.p = index_list.obj_rec;
											exterior.x = intersect_x;
											exterior.y = intersect_y;
											exterior.dist = dist;
											exterior.ext_dist = ext_dist;
										}
									}
								}
							}
						}
					}
				}
				
				/*** check to see if verticies done ***/
				if(v == index_list.v2)
					break;
	
					/*** reset last_v coords ***/
				last_vx = vx;
				last_vy = vy;
			}
		}
	}
		
		/*********** return closest interior polygon ( or -1 ) *************/
		/** note fuzzy tolerance of .001 pixel throughout **/
	if((interior.p == -1)  &&  (exterior.p == -1))
		return(-1);      /** none found **/

	if((interior.p == -1)  &&  (exterior.p >= 0))
		return(-2);      /** found exterior but no interior **/

	if(interior.dist <= exterior.dist - .001)  /** closer to interior **/
		return(interior.p);    /** this happens when topology is incomplete **/

	if((interior.dist <= exterior.dist + .001)  /** about equal dist **/
    ||  (exterior.p == -1))                     /** no exterior poly found **/
	{		/** check to see if the interior and exterior are the same poly **/
		if((interior.p != exterior.p)  /** exterior different poly than int **/
       || (exterior.p == -1))        /** no exterior poly found **/
			return(interior.p);  /** note, this is the normal case **/

		  /** so, they must be the same polygon (the very rare case) **/
		if(interior.ext_dist < exterior.ext_dist + .001)  /* ext_dist decides */
			return(interior.p);
	}

		/** in all other cases, the exterior polygon rules **/
	return(-2);       /** polys found, but none interior **/
}
