/*******************************************************************************
*       geo_to_stpln    (see manual, Internal Functions)                <in>   *
*                 This program converts geodetic coordinates (lon,lat in       *
*                 decimal degrees E and N) to state plane coordinates (E,N).   *
*                                                                              *
*                 No global or file data is changed.  The state plane          *
*                 parameters in register.h are used for the conversion.        *
*                      stpln_params[0] -  zone type:  1 = lambert              *
*                           (projection type)         2 = transverse mercator  *
*                                                     3 = Alaska zones 2-9     *
*                                                     4 = Alaska zone 1        *
*                                                     5 = azmthl equdist (Guam)*
*                      stpln_params[1]-[15] - projection parameters.  The      *
*                          number of parameters vary with the projection type. *
*                                                                              *
*                 Note:  These values and the formulae to use them were taken  *
*                 from"STATE PLANE COORDINATES BY AUTOMATIC DATA PROCESSING"  *
*                 by Charles N Claire, Geodesist (retired)                     *
*                 published by the Department of Commerce, Environmental       *
*                 Science Services Administration, Coast and Geodetic Survey,  *
*                 Publication 62-4                                             *
*                                                                              *
*                 Returned values:  0 if conversion successful, 1 if not.      *
*                                                                              *
*                 C.Jonson  june/89                                            *
*******************************************************************************/

#include"../../incl/globals.h"
#include <math.h>

long geo_to_stpln(xin, yin, xout, yout)
double xin, yin, *xout, *yout;     /*  lon,lat (input), E, N (output) */
{
	long projection;
	double C, Cm, cm_lambda;
	double s, R;
	double fi, fi_cos, fi_min, fi_sec, fi_sin, fi_sin2, fi_tan;
	double fi_cos2, fi_cos4, fi_cos6;
	double fi1, fi1_sin, fi1_tan, fi1_rad;
	double fi2, fi2_cos, fi2_sin, fi2_min, fi2_sec, fi2_rad, fi_seconds;
	double *l, *T;
	double s1, sm, s1tmp;
	double theta_sin2, theta_sec, theta_rad, theta_sin,  lambda;
	double lambda_sec, tlamb, lambda0, Blambda;
	double x, y;
	double B, D, F, G, e, P, Q, u, v, mu, BuC;
	double lat,lon;

/*  Switch sign for lambda  */
/*  code expects west longitude as positive between 0 - 180 */
	for(; xin < 0.0; xin += 360);    /*  make positive */
	for(; xin > 0.0; xin -= 360);    /*  make negative 0 to -360 */
	xin = -(xin);                    /*  switch to W longitude */

/*  set projection code    */
	projection = stpln_params[0];

/*  check for a legal projection code  */
/*      code     :  1 = lambert              */
/*                  2 = transverse mercator  */
/*                  3 = Alaska zones 2-9     */
/*                  4 = Alaska zone 1        */
/*                  5 = azmthl equdist (Guam)*/
	if((projection<1) || (projection>5))
		return(1);
/*  check for legitimate lat/lon  */
	lat = yin;
	lon = xin;
	if(state_plane_zone < 5000)    /*  lower 48  */
	{	if((lat < 23.0) || (lat > 50.0))
			return(1);
		if((lon < 65.0) || (lon > 130.0))
			return(1);
	}
	else if(state_plane_zone < 5100)    /* alaska */
	{	if((lat < 54.0) || (lat > 73.0))
			return(1);
		if((lon < 130.0) || (lon > 190.0))
			return(1);
	}
	else if(state_plane_zone < 5200)    /* hawaii */
	{	if((lat < 15.0) || (lat > 30.0))
			return(1);
		if((lon < 150.0) || (lon > 180.0))
			return(1);
	}
	else if(state_plane_zone < 5300)    /* puerto rico & virgin is. */
	{	if((lat < 15.0) || (lat > 20.0))
			return(1);
		if((lon < 60.0) || (lon > 70.0))
			return(1);
	}
	else if(state_plane_zone < 5400)    /* samoa  */
	{	if((lat < -20.0) || (lat > -5.0))
			return(1);
		if((lon < 165.0) || (lon > 175.0))
			return(1);
	}
	else if(state_plane_zone == 9999)    /* guam */
	{	xin = 360.0 - xin;       /**  make xin E longitude again */
		if((lat < 10.0) || (lat > 20.0))
			return(1);
		if((lon < 140.0) || (lon > 150.0))
			return(1);
	}
	else
		return(1);

/*  Load up fi and lambda  */
	lambda = xin/ 180.0 * M_PI;
	fi = yin / 180.0 * M_PI;

/* switch by projection code and convert data   */
	switch (projection)   
	{
	case 1:				/*   lambert                 */
		l = stpln_params;	/** Done for consistency with book **/
		fi_cos = cos(fi);
		fi_min = (double)(long)(fabs(yin) * 60.0);
		fi_sec = ( fabs(yin) * 60.0 - fi_min) * 60.0;

		s = 101.2794065 * (60.0 * (l[7] - fi_min)  +l[8] - fi_sec + 
		(1052.893882 - (4.483344 - (0.02352 * fi_cos * fi_cos) ) *
		fi_cos * fi_cos) * sin(fi) * fi_cos );

		s1 = s/100000000.0;
		R = l[3] + s * l[5] * (1 + pow(s1, 2.0) *
		(l[9] - s1 * l[10] + pow( s1, 2.0) * l[11]) );

		if( !strncmp(header.stplzn, "5010", 4) )
		{	xin = -(xin);
			theta_sec = l[6] * (xin - 662400.0);
		}
		else
			theta_sec = l[6] * (l[2] - xin * 60.0  * 60);

		theta_rad = theta_sec/3600.0 / 180.0 * M_PI;
		theta_sin = sin( theta_rad );
		theta_sin2 = sin( theta_rad / 2.0);

		x = l[1] + R * theta_sin;
		y = l[4] - R + 2 * R * pow( theta_sin2, 2.0) ;

		break;
/***********************************************************************/
	case 2:				/*   transverse mercator     */
		T = stpln_params;	/** Done for consistency with book **/
		fi_seconds = yin * 3600.0;
		fi_cos = cos(fi);
		fi_sin = sin(fi);
		fi_tan = tan(fi);
		fi_min = (double)(long)(fabs(yin) * 60.0);
		fi_sec = ( fabs(yin) * 60.0 - fi_min) * 60.0;

		lambda_sec = xin * 60.0 * 60.0;
		tlamb = (T[2] - lambda_sec)/10000.0; 
		s1tmp = T[2] - lambda_sec - 3.9174 * pow(tlamb, 3.0);
		s1 = (30.92241724 * fi_cos) / pow( (1.0 - 0.006768658 * 
			pow( fi_sin, 2.0 ) ), 0.5 )  * s1tmp; 

		sm = s1 + 4.0831 * pow( s1/100000.0, 3.0 );

		x = T[1] + 3.28083333 * sm * T[5] + 
			pow( (3.28083333 * sm * T[5])/100000, 3.0 ) * T[6];

		fi1 = fi_seconds + ((25.52381/10000000000.0) * pow(sm, 2.0)) * 
			((pow((1 - 0.0067686580 * pow(fi_sin, 2.0)), 2.0) ) *	
				fi_tan);
		fi1_rad = fi1/3600.0/180.0 * M_PI;
		fi1_sin = sin(fi1_rad);
		fi1_tan = tan(fi1_rad);

		fi2 = fi_seconds + 25.52381/10000000000.0 * pow(sm, 2.0) * 
		pow( (1.0 - 0.0067686580 * pow(fi1_sin, 2.0)), 2.0 ) * fi1_tan;

		fi2_rad = fi2/3600.0/180.0 * M_PI;
		fi2_cos = cos(fi2_rad);
		fi2_sin = sin(fi2_rad);

		fi2_min = (long)(fi2/60.0);
		fi2_sec = fi2 - fi2_min * 60.0;

		y = 101.2794065 * T[5] * (60.0 * (fi2_min - T[3]) + fi2_sec - 
			T[4] - (1052.893882 - (4.483344 - 0.02352 * 
				pow(fi2_cos, 2.0)) * pow(fi2_cos, 2.0) ) * 
					(fi2_sin * fi2_cos));

		break;
/***********************************************************************/
	case 3:				/*   Alaska zones 2-9        */
		C = stpln_params[1];  /** Done for consistency with book **/
		Cm = stpln_params[2]; /** Done for consistency with book **/

		fi_seconds = yin * 3600.0;
		fi_cos = cos(fi);
		fi_sin = sin(fi);
		fi_tan = tan(fi);
		fi_min = (double)(long)(fabs(yin) * 60.0);
		fi_sec = ( fabs(yin) * 60 - fi_min) * 60.0;
		fi_cos2 = pow(fi_cos, 2.0);
		fi_cos4 = pow(fi_cos, 4.0);
		fi_cos6 = pow(fi_cos, 6.0);

		lambda_sec = xin * 60.0 * 60.0;

		cm_lambda = (Cm - lambda_sec)/10000.0;

		x = C + ((1017862.150 * fi_cos) / pow((1 + .0068147849 *
			fi_cos2), 0.5)) * cm_lambda *
			(1 - (3.91740509/10000.0) * pow(cm_lambda, 2.0) *
			(1 - 2 * fi_cos2 - (.681478/100.0) * fi_cos4) + 
			(4.60382/100000000.0) * pow(cm_lambda, 4.0) * (1.0 - 
			20.0 * fi_cos2 + 23.6047 * fi_cos4 + .4907 * fi_cos6 ));
		
		y = 101.269278503 * (fi_seconds - 193900.05442 - (1052.893943 -
			4.483386 * fi_cos2 + 2.3559/100.0 * fi_cos4)
			* pow( (1.0 - fi_cos2), 0.5)  * fi_cos) +
			(24673.67480 * (pow((1 - fi_cos2), 0.5) * fi_cos)) / 
			pow((1 + .0068147849 * fi_cos2), 0.5) *
			pow(cm_lambda, 2.0) * (1.0 + 1.958703/10000.0 *
			pow(cm_lambda, 2.0) * (-1.0 + 6 * fi_cos2 +
			(6.133306/100) * fi_cos4 + 1.8577/10000.0 * fi_cos6) +
			1.5346/100000000.0 * pow(cm_lambda, 4.0) * (1.0 - 60.0 *
			fi_cos2 + 117.75 * fi_cos4 + 4.089 * fi_cos6));

		break;
/***********************************************************************/
	case 4:				/*   Alaska zone 1           */
		B = stpln_params[2];
		C = stpln_params[3];
		D = stpln_params[4];
		F = stpln_params[5];
		G = stpln_params[6];
		e = stpln_params[9];
		fi = yin/180.0 * M_PI;
		fi_sin = sin(fi);
		lambda = xin/180.0 * M_PI;			/*  rads  */
		lambda0 = stpln_params[15] / 180.0 * M_PI;	/*  rads  */

		mu = log(tan( M_PI/4.0 + fi/2.0)) - e/2.0 * 
			log( (1.0 + e * fi_sin)/(1.0 - e * fi_sin) );
		BuC = B * mu + C;
		Blambda = B * (lambda - lambda0);
		P = (pow(M_E, BuC) - pow(M_E, -(BuC)) ) / 2.0;
		Q = (pow(M_E, BuC) + pow(M_E, -(BuC)) ) / 2.0;
		u = D * atan( (G * P + F * sin(Blambda) ) /
			cos(Blambda));
		v = D/2.0 * log((Q + F * P - G * sin( Blambda) ) /
			(Q - F * P + G * sin(Blambda) ) );

		x = 3.28083333333 * (-0.6 * u + 0.8 * v + 5000000.0);
		y = 3.28083333333 * (0.8 * u + 0.6 * v - 5000000.0);
		break;
/***********************************************************************/
	case 5:				/*   azmthl equdist (Guam)   */
/***  No stpln_params are used as all values for Guam are hardcoded  */
		fi_seconds = yin * 3600.0;
		fi_cos = cos(fi);
		fi_sin = sin(fi);
		fi_tan = tan(fi);
		fi_sin2 = pow(fi_sin, 2.0);
		fi_cos2 = pow(fi_cos, 2.0);
		fi_cos4 = pow(fi_cos, 4.0);

		lambda_sec = xin * 60.0 * 60.0;
	
		x = 50000.0 + (30.92241724 * (lambda_sec - 521095.50254) *
			fi_cos) / pow( (1.0 - .0067686580 * fi_sin2), 0.5);

		y = 50000.0 + 30.87002482 * (fi_seconds - 48263.2837578 -
			(1052.893943 - 4.483386 * fi_cos2 + .023559 * fi_cos4)
			* fi_sin * fi_cos) + pow(((x - 50000.0)/10000.0), 2.0)
			* (fi_tan * pow( (1.0 - .0067686580 * fi_sin2), 0.5)) /
			.127564128 ;
		break;
	}

	*xout = x;
	*yout = y;

	return(0);
}
