/*******************************************************************************
*     get_parms                                                                *
*        This program accepts 2 or more points (both in true and trial coords) *
*    as input and calculates the parameters for an rms best fit translation-   *
*    rotation-change of scale transformation.  The parameters (a,b,c,d) are    *
*    used in the transformation:    xout = a*x - b*y + c                       *
*                                   yout = b*x + a*y + d                       *
*                                                                              *
*    parameters:  double params[4]:    (a, b, c, d, the returned parameters)   *
*                 double true[m][2]:   the true control locations ([0] = x)    *
*                 double trial[m][2]:  the measured control locations          *
*                 long m:              the number of points                    *
*                                                                              *
*    Globals modified:   none                                                  *
*                                                                              *
*    Returned values:  0 = normal completion                                   *
*                      1 = failure (not enough separate points etc.)           *
*                                                                              *
*    Method taken from Example 8.8 (p 201) of Observations and Least Squares   *
*    by  Edward M Mikhail 1976 pub by Thomas Y Crowell Co. Inc.                *
*                                                                              *
*    J.Dabritz,  Aug 4, 1989                                                   *
*******************************************************************************/
   


long get_params(params, true, trial, m)
	double params[4];             /*  for passing back 4 transform parameters */
	double true[][2];             /*  contain m control pts ([m][2])          */
	double trial[][2];            /*  contain m trial control pts ([m][2])    */
	long m;                       /*  number of points                        */
{
	double matrix[4][5];
	double a,b,c,d,e,f,g;
	
	long i;

		/*** initialize accumulation variables ***/
	a=b=c=d=e=f=g= 0.0;

		/*** accumulate values for matrix **/
	for(i=0; i<m; i++)
	{		/****  a = sum of trial  x*x + y*y             ****/
		a += trial[i][0] * trial[i][0] + trial[i][1] * trial[i][1];
			/****  b = sum of trial x's                    ****/
		b += trial[i][0];
			/****  c = sum of trial y's                    ****/
		c += trial[i][1];
			/****  d = sum trialx*truex + trialy*truey     ****/
		d += trial[i][0] * true[i][0] + trial[i][1] * true[i][1];
			/****  e = sum trialx*truey - trialy*truex     ****/
		e += trial[i][0] * true[i][1] - trial[i][1] * true[i][0];
			/****  f = sum of true x's                     ****/
		f += true[i][0];
			/****  g = sum of true y's                     ****/
		g += true[i][1];
	}
/*       note, matrix is symmetric with coefficients tacked on to the right: */
/*			(	a	0	b	c	  d  )                               */
/*			(	0	a  -c	b 	  e  )                               */
/*			(	b  -c	m	0	  f  )                               */
/*			(	c	b	0	m	  g  )                               */

		/*** set all matrix values (symetric 4 X 4 plus a column for solution)*/
	matrix[0][1] = matrix[1][0] = matrix[2][3] = matrix[3][2] = 0.0;
	matrix[0][0] = matrix[1][1] = a;
	matrix[0][2] = matrix[2][0] = matrix[1][3] = matrix[3][1] = b;
	matrix[0][3] = matrix[3][0] = c;
	matrix[1][2] = matrix[2][1] = -c;
	matrix[2][2] = matrix[3][3] = m;
	matrix[0][4] = d;
	matrix[1][4] = e;
	matrix[2][4] = f;
	matrix[3][4] = g;

		/*** solve for parameters ***/
	if(mat_solve(4,5,matrix))
		return(1);                /** no unique solution **/

		/*** set parameters in pass back array **/
	params[0] = matrix[0][4];   /*  transformation formula coefficient a  */
	params[1] = matrix[1][4];   /*  transformation formula coefficient b  */
	params[2] = matrix[2][4];   /*  transformation formula coefficient c  */
	params[3] = matrix[3][4];   /*  transformation formula coefficient d  */

		/*** return ***/
	return(0);
}
