/*******************************************************************************
*     get_units                                                                *
*          This program calculates and fill the unit_factor array for as       *
*     many of the designated output/input unit type as possible.  The          *
*     possibilities are:                                                       *
*         (unit_factor[10][2])  (pixel to other units conversion factors)      *
*                               0 = raster pixels[0],  square raster pixels[1] *
*                               1 = map in.[0],  square inches[1]              *
*                               2 = ground ft[0],  ground acres[1]             *
*                               3 = ground mi[0],  square miles[1]             *
*                               4 = map cm[0],  square cm[1]                   *
*                               5 = ground meters[0],  square meters[1]        *
*                               6 = ground km[0],  square km[1]                *
*                               7 = - unused -                                 *
*                               8 = - unused -                                 *
*                               9 = - unused -                                 *
*                                                                              *
*     only the unit_factor global is changed, no files or other globals.       *
*                                                                              *
*     Returns:   (long) 0 = all factors set.                                   *
*                       1 = some factors not set.                              *
*                                                                              *
*           J.Dabritz  8/19/89                                                 *
*******************************************************************************/

#include "../../incl/globals.h"
#include <math.h>

long get_units()
{
	long i;

	double raster_dx, raster_dy, raster_ysum;
	double raster_a, raster_p;
	double utm_dx, utm_dy, utm_ysum;
	double utm_a, utm_p;


		/*************** initialize unit_factor global ******************/
	unit_factor[0][0] = 1.0;   /** pixels to pixels conversion **/
	unit_factor[0][1] = 1.0;   /** sq pixels to sq pixels conversion **/
	for(i=1; i<10; i++)
	{	unit_factor[i][0] = 0.0;
		unit_factor[i][1] = 0.0;
	}

		/*************** check for map **********************************/
	if(!job_stats[0])
		return(1);

		/*************** check for raster registration ******************/
	if(!registered(0))
		return(1);

		/*************** check for utm registration *********************/
	if(!registered(2))
		return(1);

		/*************** calculate ground factors ***********************/
		/** first calculate raster and utm perimeter & area **/
	raster_a = 0.0;
	raster_p = 0.0;
	utm_a = 0.0;
	utm_p = 0.0;
	for(i=0; i<4; i++)
	{	raster_dx = corner[0][(i+1) % 4][0]  -  corner[0][i][0];
		raster_dy = corner[0][(i+1) % 4][1]  -  corner[0][i][1];
		raster_ysum = corner[0][(i+1) % 4][1]  +  corner[0][i][1];
		
		raster_a += (raster_dx * raster_ysum / 2.0);
		raster_p += sqrt(raster_dx * raster_dx  +  raster_dy * raster_dy);

		utm_dx = corner[2][(i+1) % 4][0]  -  corner[2][i][0];
		utm_dy = corner[2][(i+1) % 4][1]  -  corner[2][i][1];
		utm_ysum = corner[2][(i+1) % 4][1]  +  corner[2][i][1];
		
		utm_a += (utm_dx * utm_ysum / 2.0);
		utm_p += sqrt(utm_dx * utm_dx  +  utm_dy * utm_dy);
	}
		/** convert to ground factors **/
	if((raster_a==0.0) || (raster_p == 0.0) || (utm_a == 0.0) || (utm_p == 0.0))
		return(1);

	unit_factor[5][0] = utm_p / raster_p;    /** pixels to meters **/
	unit_factor[5][1] = utm_a / raster_a;    /** sq pixels to sq meters **/

	unit_factor[6][0] = unit_factor[5][0] / 1000;   /** pixels to km **/
	unit_factor[6][1] = unit_factor[5][1] / 1000000;   /** pixels to sq km **/

	unit_factor[2][0] = unit_factor[5][0] * 3937.0 / 1200.0; /* pix to ft */
	unit_factor[2][1] = unit_factor[5][1] * 3937.0 * 3937.0 / 1200.0 / 1200.0
												  /	43560.0; /* pix to acres */

	unit_factor[3][0] = unit_factor[2][0] / 5280.0;   /* pix to miles */
	unit_factor[3][1] = unit_factor[2][1] / 640.0;    /* pix to sq mi */
	
		/*************** check for map_scale ****************************/
	if(map_scale == 0.0)
		return(1);

		/*************** calculate map factors **************************/
	unit_factor[1][0] = unit_factor[2][0] * 12.0 / map_scale;  /* pix > in */
	unit_factor[1][1] = unit_factor[2][1] * 43560.0 * 12.0 * 12.0
						/ map_scale / map_scale;        /* sq pix > sq in. */

	unit_factor[4][0] = unit_factor[5][0] * 100.0 / map_scale; /* pix > cm */
	unit_factor[4][1] = unit_factor[5][1] * 100.0 * 100.0
						/ map_scale / map_scale;      /* sq pix > sq cm */

		/*************** return *****************************************/
	return(0);
}
