/*******************************************************************************
*       index_vobj      (see manual, Internal Functions)                <in>   *
*                                                                              *
*        Sets up the index_array & index_list files for the objects in the     *
*    v_objects file.  The index_array has a row and column                     *
*    for every 32 rows and columns in the raster.  0,0 is in the lower         *
*    left corner of each.  The index array has a single index value which      *
*    is -1 or the index_list file record number of the first object which is   *
*    within/accross it.  These index_list records form a linked-list for each  *
*    index_array cell.                                                         *
*                                                                              *
*    Arguments:    long begin_rec     beginning record in the v_objects file   *
*                                     from which to index.  All objects will   *
*                                     be indexed from the input record number  *
*                                     to the end of the file (job_stats[17]).  *
*                                                                              *
*    Returned Values: (long) 0 = ok,  1 = failure.                             *
*                                                                              *
*    Global Variables Modified:  none, the job_stats, index_array, and         *
*    index_list files are changed.                                             *
*                      j.dabritz  9/18/89                                      *
*******************************************************************************/

#include "../../incl/globals.h"
#include <math.h>

long index_vobj(begin_rec)
	long begin_rec;
{
	V_OBJECTS v_object;
	COORD_LIST coord_list;
	INDEX_LIST index_list;
	char string[256];
	long i,o;
	long long_x, long_y;
	long row, col;
	long ix1, iy1, ix2, iy2;
	long iystep, ixstep;
	long vertex;
	long increment;

	double x1,y1,x2,y2;
	double dx,dy;
	double slope;
	double xstep, ystep;
	double xenter, yenter;
	double xexit, yexit;

		/********* check for map ****************************************/
	if(!job_stats[0])
	{	message("abort, index_vobj, location 1");
		return(1);
	}

		/********* check input parameter ********************************/
	if(begin_rec < 0)
	{	message("Cannot index v_object records < zero.  Aborting");
		return(1);
	}

		/********* write a dummy first record ***************************/
	if(put_indexl(&index_list,0))
	{	message("abort, index_vobj, location 2");
		return(1);
	}
	
		/********* put objects into new file ****************************/
	for(o = begin_rec; o < job_stats[17]; o++)
	{
		if(!((o-begin_rec) % 100))
		{	sprintf(string,"indexing v_object %d",o);
			message(string);
		}

			/** check for user abort **/
		if(abort_cmd())
			return(1);

		if(get_vobjects(&v_object,o))
		{	message("abort, index_vobj, location 3");
			return(1);
		}

		if((v_object.type % 10) == 0)        /* object is a node */
		{		/***** index a node ********/
				/* find index row and column of node */
			if(get_v_coord(&coord_list,v_object.v1))
			{	message("abort, index_vobj, location 4");
				return(1);
			}
			long_x = coord_list.x + 0.5;
			long_y = coord_list.y + 0.5;
			row = long_y >> 5;
			col = long_x >> 5;
				/* set index_list data */
			index_list.type = v_object.type;
			index_list.obj_rec = o;
			index_list.v1 = v_object.v1;
			index_list.v2 = v_object.v2;
			if((row >= 0) && (row <= (job_stats[2] >> 5)) &&
			   (col >= 0) && (col <= (job_stats[1] >> 5)))
			{	index_list.next = index_array[row][col];
					/* insert object into linked list */      
				index_array[row][col] = job_stats[15];
				if(put_indexl(&index_list,job_stats[15]))
				{	message("abort, index_vobj, location 5");
					return(1);
				}
			}
		}
		else                            /* object is a segment or area */
		{		/***** index a segment ********/
				/* set index_list data */
			index_list.type = v_object.type;
			index_list.obj_rec = o;
			index_list.v1 = v_object.v1;

				/***** set beginning vertex data ***/
			if(get_v_coord(&coord_list,v_object.v1))
			{	message("abort, index_vobj, location 6");
				return(1);
			}
			x2 = (coord_list.x + 0.5) * 0.03125; /* fractional cell x */
			y2 = (coord_list.y + 0.5) * 0.03125; /* fractional cell y */

				/***** index all verticies in the segment   */
			increment = (v_object.v1 < v_object.v2) ?   1 : -1;
			for(vertex = v_object.v1;  ; vertex += increment)
			{		/*** get vertex coords **/
					/**** increment pt 1 *******/
				x1 = x2;
				y1 = y2;
				ix1 = x1;                       /* integer cell x    */
				iy1 = y1;                       /* integer cell y    */
				if(get_v_coord(&coord_list,vertex))
				{	message("abort, index_vobj, location 7");
					return(1);
				}
				x2 = (coord_list.x + 0.5) * 0.03125; /* fractional cell x */
				y2 = (coord_list.y + 0.5) * 0.03125; /* fractional cell y */
				ix2 = x2;                            /* integer cell x    */
				iy2 = y2;                            /* integer cell y    */

					/*** reset ending vertex record pointer **/
				index_list.v2 = vertex;

					/*** if in a new grid cell then store index_list record */
				if((ix1 != ix2) || (iy1 != iy2))
				{	dx = x2 - x1;        /*  change in cell coordinates */
					dy = y2 - y1;
						/*** check for more vertical or horizontal vector */
					if(fabs(dy) > fabs(dx))       /* more vertical than hz */
					{		/* calc y slope */
						slope = dx / dy;

							/* set variables for storing by row */
						if(y2 > y1)
							ystep = 1.0;
						else
							ystep = -1.0;
						xstep = ystep * slope;
						if(xstep > 0.0)
							ixstep = 1;
						else
							ixstep = -1;

							/** set enter and exit variables for first col */
						xenter = x1;
						yenter = y1;
						if(y2 > y1)
							yexit = floor(yenter + ystep);
						else
							yexit = floor(yenter);
						xexit = xenter + (yexit - yenter) * slope;


							/*** store path, one row at a time **/
						for(; (iy1 != iy2) || (ix1 != ix2) ; )
						{		/*** store cells for one row **/
							for(;; ix1 += ixstep)
							{		/** exit befor storing last cell **/
								if((ix1 == ix2) && (iy1 == iy2))  
									break;
								  /** store cell record **/
								if((iy1 >= 0) && (iy1 <= (job_stats[2] >> 5)) &&
								   (ix1 >= 0) && (ix1 <= (job_stats[1] >> 5)))
								{	index_list.next = index_array[iy1][ix1];
									index_array[iy1][ix1] = job_stats[15];
									if(put_indexl(&index_list,job_stats[15]))
									{	message("abort, index_vobj, location 8");
										return(1);
									}
								}
									/*** reset first vertex **/
								index_list.v1 = index_list.v2-increment;
									/*** check for end of vector **/
								if(ix1 == (long) xexit)
									break;
							}
								/*** calc xenter, xexit for next row **/
							xenter = xexit;
							xexit = xenter + xstep;
							if(iy1 != iy2)
								iy1 += ystep;
						}
					}
					else                          /* more horizontal that vt */
					{		/* calc x slope */
						slope = dy / dx;

							/* set variables for storing by col */
						if(x2 > x1)
							xstep = 1.0;
						else
							xstep = -1.0;
						ystep = xstep * slope;
						if(ystep > 0.0)
							iystep = 1;
						else
							iystep = -1;

							/** set enter and exit variables for first row */
						yenter = y1;
						xenter = x1;
						if(x2 > x1)
							xexit = floor(xenter + xstep);
						else
							xexit = floor(xenter);
						yexit = yenter + (xexit - xenter) * slope;

							/*** store path, one col at a time **/
						for(; (iy1 != iy2) || (ix1 != ix2) ; )
						{		/*** store cells for one col **/
							for(;; iy1 += iystep)
							{		/** exit befor storing last cell **/
								if((iy1 == iy2) && (ix1 == ix2))  
									break;
								  /** store cell record **/
								if((iy1 >= 0) && (iy1 <= (job_stats[2] >> 5)) &&
								   (ix1 >= 0) && (ix1 <= (job_stats[1] >> 5)))
								{	index_list.next = index_array[iy1][ix1];
									index_array[iy1][ix1] = job_stats[15];
									if(put_indexl(&index_list,job_stats[15]))
									{	message("abort, index_vobj, location 9");
										return(1);
									}
								}
									/*** reset first vertex **/
								index_list.v1 = index_list.v2-increment;
									/*** check for end of vector **/
								if(iy1 == (long) yexit)
									break;
							}
								/*** calc yenter, yexit for next col **/
							yenter = yexit;
							yexit = yenter + ystep;
							if(ix1 != ix2)
								ix1 += xstep;
						}
					}
				}
				if(vertex == v_object.v2)
					break;
			}
				/*** store index_list record for final vertex **/
			if((iy1 >= 0) && (iy1 <= (job_stats[2] >> 5)) &&
			   (ix1 >= 0) && (ix1 <= (job_stats[1] >> 5)))
			{	index_list.next = index_array[iy1][ix1];
				index_array[iy1][ix1] = job_stats[15];
				if(put_indexl(&index_list,job_stats[15]))
				{	message("abort, index_vobj, location 11");
					return(1);
				}
			}
				/*** reset first vertex **/
			index_list.v1 = index_list.v2;
		}
	}

		/********* set need-to-index flag to 0 **************************/
	job_stats[49] = 0;


		/********* return ***********************************************/
	message("  indexing finished");

	return(0);
}
