/*******************************************************************************
*       locale_ptr      (see manual, Internal Functions)                <in>   *
*                                                                              *
*    Purpose:    Checks the raster at the input location and returns a         *
*    'locale' variable based on the 8 surrounding bits.  The returned          *
*    variable is a long in which only the lowest 8 bits are used.  These       *
*    based on the surrounding bits as follows ( X marks center ):              *
*                                                                              *
*                                     7 0 1                                    *
*                                     6 X 2                                    *
*                                     5 4 3                                    *
*                                                                              *
*    Arguments:    long *word_ptr;   pointer to word of raster to check.       *
*                  long bit;         bit number (0-31) on word to check around.*
*    NOTE: 1. THERE IS NO PARAMETER CHECKING IN THIS ROUTINE.  THE POINTER     *
*             AND BIT NUMBER ARE ASSUMED TO BE LEGITIMATE.                     *
*          2. THE BITS OF WORDS ON 80386 MACHINES ARE NUMBERED AS FOLLOWS:     *
*                             1 1 1 1 1 1     2 2 2 2 1 1 1 1 3 3 2 2 2 2 2 2  *
*             7 6 5 4 3 2 1 0 5 4 3 2 1 0 9 8 3 2 1 0 9 8 7 6 1 0 9 8 7 6 5 4  *
*          3. THERE IS NO SPECIAL PROCESSING DUE TO MAP EDGES (SIDES OR        *
*             TOP OR BOTTOM).  IT IS ASSUMED THAT THESE AREAS ARE AVAILABLE    *
*             IN RAM TO CHECK.                                                 *
*          (1 & 3 above are assumed to allow faster processing in this         *
*          intensively used routine.)                                          *
*                                                                              *
*    Returned Values: (long) locale (as defined above) ( 0 - 255 ).            *
*                                                                              *
*    Global Variables Modified:  none                                          *
*                                                                              *
*                      j.dabritz  6/29/89                                      *
*******************************************************************************/
/* Modified 6/92 by DBS@CWU for big endian support */
/*   THE BITS OF WORDS ON Big Endian MACHINES ARE NUMBERED AS FOLLOWS:     *
*         3 3 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1                      *
*         1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0  *
*/
#include "../../incl/globals.h"


/*****
#ifndef BIG_ENDIAN 
*****/
  /***** first here is the original Little Endian code *****/


long locale_ptr(word_ptr, bit)
	register long *word_ptr;
	register long bit;
{	
	static long lt_bit;        /*  bit number of bit to the left of center  */
	static long rt_bit;        /*  bit number of bit to the right of center */
	static long locale;        /*  returned variable                        */


		/***************** calc the locale variable **************************/
	switch(bit & 7)   {
	case 0:               /************* on rt edge of byte  */
		lt_bit = bit + 1;
		if(bit == 24)     /************* on rt edge of word  */
		{	rt_bit = 7;
			locale = ((*word_ptr >> lt_bit) & 1) << 6;
			word_ptr++;
			locale |= ((*word_ptr >> rt_bit) & 1) << 2;
			word_ptr -= raster_wincrement;
			locale |= ((*word_ptr >> rt_bit) & 1) << 3;
			word_ptr--;
			locale |= ((*word_ptr >>    bit) & 3) << 4; /* bits 5 4 together */
			word_ptr += raster_wincrement + raster_wincrement;
			locale |= ((*word_ptr >> lt_bit) & 1) << 7;
			locale |= ((*word_ptr >>    bit) & 1)     ;
			word_ptr++;
			locale |= ((*word_ptr >> rt_bit) & 1) << 1;
		}
		else              /************* not on rt edge of word */
		{	rt_bit = bit + 15;
			locale = ((*word_ptr >> rt_bit) & 1) << 2;
			locale |= ((*word_ptr >> lt_bit) & 1) << 6;
			word_ptr -= raster_wincrement;
			locale |= ((*word_ptr >>    bit) & 3) << 4; /* bits 5 4 together */
			locale |= ((*word_ptr >> rt_bit) & 1) << 3;
			word_ptr += raster_wincrement + raster_wincrement;
			locale |= ((*word_ptr >> lt_bit) & 1) << 7;
			locale |= ((*word_ptr >>    bit) & 1)     ;
			locale |= ((*word_ptr >> rt_bit) & 1) << 1;
		}
		break;
	case 7:               /************* on lt edge of byte  */
		rt_bit = bit - 1;
		if(bit == 7)      /************* on lt edge of word  */
		{	lt_bit = 24;
			locale = ((*word_ptr >> rt_bit) & 1) << 2;
			word_ptr--;
			locale |= ((*word_ptr >> lt_bit) & 1) << 6;
			word_ptr -= raster_wincrement;
			locale |= ((*word_ptr >> lt_bit) & 1) << 5;
			word_ptr++;
			locale |= ((*word_ptr >> rt_bit) & 3) << 3;   /* bits 4 3 together*/
			word_ptr += raster_wincrement + raster_wincrement;
			locale |= ((*word_ptr >>    bit) & 1)     ;
			locale |= ((*word_ptr >> rt_bit) & 1) << 1;
			word_ptr--;
			locale |= ((*word_ptr >> lt_bit) & 1) << 7;
		}
		else              /************* not on lt edge of word */
		{	lt_bit = bit - 15;
			locale = ((*word_ptr >> rt_bit) & 1) << 2;
			locale |= ((*word_ptr >> lt_bit) & 1) << 6;
			word_ptr -= raster_wincrement;
			locale |= ((*word_ptr >> lt_bit) & 1) << 5;
			locale |= ((*word_ptr >> rt_bit) & 3) << 3;   /* bits 4 3 together*/
			word_ptr += raster_wincrement + raster_wincrement;
			locale |= ((*word_ptr >> lt_bit) & 1) << 7;
			locale |= ((*word_ptr >>    bit) & 1)     ;
			locale |= ((*word_ptr >> rt_bit) & 1) << 1;
		}
		break;
	default:              /************* in middle of byte   */
		lt_bit = bit + 1;
		rt_bit = bit - 1;
		locale = ((*word_ptr >> rt_bit) & 1) << 2;
		locale |= ((*word_ptr >> lt_bit) & 1) << 6;
		word_ptr -= raster_wincrement;
		locale |= ((*word_ptr >> rt_bit) & 7) << 3;   /*  bits 5 4 3 together */
		word_ptr += raster_wincrement + raster_wincrement;
		locale |= ((*word_ptr >> lt_bit) & 1) << 7;
		locale |= ((*word_ptr >>    bit) & 1)     ;
		locale |= ((*word_ptr >> rt_bit) & 1) << 1;
		break;
	}

		/***************** return **************************************/
	return(locale);
}

