/*******************************************************************************
*       not_thinned     (see manual, Command Ref., Internal Functions) <in>    *
*               Description:  checks to see if the raster is thinned.          *
*                                                                              *
*               Parameters:  
*                            long thin_type;  thinning type code:              *
*                               0 : non-destructive, leave spur ends, remove   *
*                                   square corners & stairsteps.               *
*                               1 : non-destructive, leave spur ends, leave    *
*                                   square corners & stairsteps.               *
*                               2:  DESTRUCTIVE, remove all pixels which are   *
*                                   not part of a solid interior.              *
*                                                                              *
*               Globals changed:  none                                         *
*                                                                              *
*               Returns zero if raster thinned, 1 if not thinned, -1=error.    *
*               J.Dabritz   7/6/90                                             *
*******************************************************************************/

#include <memory.h>
#include "../../incl/globals.h"
#include "../../incl/in.h"

long not_thinned(thin_type)
  long thin_type;
{
  long *row_done_array; /*** pointer to array of flags, one for each row      */
  long *done_ptr;       /*** pointer to done flag for current row             */
  
  long *old_row_array;  /*** pointer to array of 2 old raster rows            */
  long *old_row0_ptr;   /*** pointer to old current row                       */
  long *old_row1_ptr;   /*** pointer to old row above current row             */

  long *row_ptr;        /*** pointer to current raster row                    */
  long *end_raster_ptr; /*** pointer just beyond end of raster (last row)     */

  long new_old_incr0;   /*** difference (increment) from new to old row0 ptrs */
  long new_old_incr1;   /*** difference (increment) from new to old row1 ptrs */

  long *ptr, *end_ptr;  /*** pointers for temporary use, initializing etc.    */

  long *word_ptr;       /*** pointer to current word in raster row            */
  long *end_row_ptr;    /*** pointer just beyond end of current raster row    */
  long word;            /*** resolved word_ptr value                          */

  long *ptr0;           /*** pointer to word in old row 0 (current row)       */
  long *ptr1;           /*** pointer to word above in old array               */

    /***  array of all bit numbers of a word in RAM order  */
  static long bits[32] = { 7,6,5,4,3,2,1,0,  15,14,13,12,11,10,9,8,
    23,22,21,20,19,18,17,16,  31,30,29,28,27,26,25,24 };
  long *bit_ptr;        /*** pointer to current bit number                    */
  long *end_bit_ptr;    /*** pointer just beyond bits array                   */

  long locale;          /*** locale variable coding 8 surrounding bits        */
  long old_locale;      /*** same for 'old' raster of previous pass           */
  long key;             /*** value of tlocale array for current pixel         */
  long old_key;         /*** value of tlocale array on the last-pass raster   */
  long thin_mask;       /*** mask for checking value of 'key' for thinning    */
  long directional_check; /* mask for bit flag for directional thinning       */

  long count;           /*** count of pixels thinned on a single pass         */

  long bit;             /*** bit number (0-31) of pixel in question           */
  long lt_bit;          /*** bit number (0-31) of bit to the left             */
  long rt_bit;          /*** bit number (0-31) of bit tot the right           */


    /***************** check parameters ********************************/
  if((thin_type < 0)  ||  (thin_type > 2))
  {  message("illegal not_thinned parameters.  not_thinned aborted.");
    return(1);
  }

    /***************** check for map present ***************************/
  if(!job_stats[0])
  {  message("No map current acquired.  not_thinned aborted.");
    return(1);
  }

    /***************** set up row tracking array ***********************/
  row_done_array = (long *) calloc_ltp(job_stats[2]+1, 4);
  if(row_done_array <= 0)
  {  message("not_thinned aborted.  Not enough RAM space available.");
    return(1);
  }


    /***************** set up old row array ****************************/
  old_row_array = (long *) calloc_ltp(3 + 2 * raster_wincrement,4);
  if(old_row_array <= 0)
  {  message("not_thinned aborted.  Not enough RAM space available.");
    free_ltp((char *) row_done_array);
    return(1);
  }
  old_row0_ptr = old_row_array + 1;                 /*** space for 1 row  */
  old_row1_ptr = old_row0_ptr + raster_wincrement + 1; /*space for 1 more row */


    /***************** set thin and directional_check bit masks ********/
  thin_mask = 1 << (15 + thin_type);
  directional_check = 1 << 18;


    /***************** thin by loop ************************************/
  end_raster_ptr = (long *) (lraster[0] - 1);
  end_bit_ptr = bits + 32;
  count = 0;
    /***  init done pointer */
  done_ptr = row_done_array + job_stats[2];
    /***  reset old row above to zeros (empty is assumed around raster) */
  end_ptr = old_row0_ptr + raster_wincrement;
  for(ptr = old_row0_ptr;  ptr < end_ptr  ;  ptr++)
    *ptr = 0;
    /***  set pointer to row            */
  row_ptr = (long *) (lraster[job_stats[2]]);
    /***  thin each row, top to bottom  */
  for( ; (row_ptr > end_raster_ptr) && !count ; row_ptr -= raster_wincrement, done_ptr--)
  {    /*** reset old row pointers */
    ptr = old_row1_ptr;
    old_row1_ptr = old_row0_ptr;
    old_row0_ptr = ptr;
      /*** load current row into old_row0 */
    memcpy((char *) old_row0_ptr, (char *) row_ptr, raster_bincrement);
      /*** check for user abort ***/
    if(abort_cmd())
      return(1);
      /*** calc new increment from current to old row pointers */
    new_old_incr0 = old_row0_ptr - row_ptr;
    new_old_incr1 = old_row1_ptr - row_ptr;
      /*** check for row complete ******************/
    if(!(*done_ptr))
    {    /***  set default row done flag  */
      *done_ptr = 1;
        /***************** thin by word, left to right *************/
      end_row_ptr = row_ptr + raster_wincrement;
      for(word_ptr = row_ptr; word_ptr < end_row_ptr; word_ptr++)
      {    /***  check for any on bits in word  */
        if(*word_ptr)
        {   /*** set word variable (for repeted use, faster than *word_ptr) */
          word = *word_ptr;
            /*** thin word by bit, left to right ******/
          for(bit_ptr = bits; bit_ptr < end_bit_ptr; bit_ptr++)
          {    /***  check for 'on' bit  */
            if(word & (1 << *bit_ptr))
            {    /***************** calc new locale key *********/
              locale = locale_ptr(word_ptr,*bit_ptr);
              key = tlocale[locale];
                /*** calc locale for old raster at this location  */
              old_locale = locale & 60;   /* clear bits 6,7,0,1  */
              ptr0 = word_ptr + new_old_incr0;
              ptr1 = word_ptr + new_old_incr1;
              bit = *bit_ptr;
                /******* calc the old_locale variable *********/
              switch(bit & 7)   {
              case 0:             /*** on rt edge of byte  */
                lt_bit = bit + 1;
                if(bit == 24)     /*** on rt edge of word  */
                {  rt_bit = 7;
                  old_locale |= ((*ptr0 >> lt_bit) & 1) << 6;
                  old_locale |= ((*ptr1 >> lt_bit) & 1) << 7;
                  old_locale |= ((*ptr1 >>    bit) & 1)     ;
                  ptr1++;
                  old_locale |= ((*ptr1 >> rt_bit) & 1) << 1;
                }
                else              /*** not on rt edge of word */
                {  rt_bit = bit + 15;
                  old_locale |= ((*ptr0 >> lt_bit) & 1) << 6;
                  old_locale |= ((*ptr1 >> lt_bit) & 1) << 7;
                  old_locale |= ((*ptr1 >>    bit) & 1)     ;
                  old_locale |= ((*ptr1 >> rt_bit) & 1) << 1;
                }
                break;
              case 7:             /*** on lt edge of byte  */
                rt_bit = bit - 1;
                if(bit == 7)      /*** on lt edge of word  */
                {  lt_bit = 24;
                  ptr0--;
                  old_locale |= ((*ptr0 >> lt_bit) & 1) << 6;
                  old_locale |= ((*ptr1 >>    bit) & 1)     ;
                  old_locale |= ((*ptr1 >> rt_bit) & 1) << 1;
                  ptr1--;
                  old_locale |= ((*ptr1 >> lt_bit) & 1) << 7;
                }
                else              /*** not on lt edge of word */
                {  lt_bit = bit - 15;
                  old_locale |= ((*ptr0 >> lt_bit) & 1) << 6;
                  old_locale |= ((*ptr1 >> lt_bit) & 1) << 7;
                  old_locale |= ((*ptr1 >>    bit) & 1)     ;
                  old_locale |= ((*ptr1 >> rt_bit) & 1) << 1;
                }
                break;
              default:              /*** in middle of byte   */
                lt_bit = bit + 1;
                rt_bit = bit - 1;
                old_locale |= ((*ptr0 >> lt_bit) & 1) << 6;
                old_locale |= ((*ptr1 >> lt_bit) & 1) << 7;
                old_locale |= ((*ptr1 >>    bit) & 1)     ;
                old_locale |= ((*ptr1 >> rt_bit) & 1) << 1;
                break;
              }
              old_key = tlocale[old_locale];

                /*************  check for thinning   ***********/
              if(old_locale)
              {  if((old_key & directional_check) && !(thin_type & 2))
                {  if((locale == (255 & (old_key>> 19))) ||
                       ((old_key & thin_mask) && (key & thin_mask)))
                  {    /*** thin it ***/
                    count++;
                  }
                  else
                    *done_ptr = *done_ptr & (old_key >> 17) & 1;
                }
                else
                {  if(old_key & thin_mask)
                  {  if(key & thin_mask)
                    {    /*** thin it ***/
                      count++;
                    }
                    else
                      *done_ptr = *done_ptr & (old_key >> 17) & 1;
                  }
                  else
                    *done_ptr = *done_ptr & (old_key >> 17) & 1;
                }
              }
              else
              if(!(map_type & 1))             /*  not a point map  */
              {    /*** thin it ***/
                count++;
              }
            }
          }
        }
      }
    }
  }


    /***************** free malloc'd space *****************************/
  free_ltp((char *) row_done_array);
  free_ltp((char *) old_row_array);


    /***************** return ******************************************/
  if(count)
    return(1);
  else
    return(0);
}
