/*******************************************************************************
*       reg_raster      (see manual, Internal Functions)                <in>   *
*    This program uses the raster size and another coordinate system           *
*    registration to calculate a raster registration of the same shape.        *
*    This allows automatic raster registration (unless it already exists)      *
*    upon the first geodetic or digitizer registration.                        *
*                                                                              *
*    Arguments:    long  ct:  coord_type (same as in registration file)        *
*                                                                              *
*    Returned Values: (long) 0 = raster (coord_type 0) registered              *
*                            1 = failure                                       *
*                                                                              *
*    Global Variables Modified:  corner[0][][]                                 *
*                                                                              *
*                      j.dabritz  8/21/89                                      *
*******************************************************************************/

#include "../../incl/globals.h"
#include <math.h>

long reg_raster(ct)
long ct;
{	
	long i,j;

	double dx,dy;
	double hw_ratio, hw_raster;
	double ht, wd, margin;
	double params[4];
	double x,y;
	double minmax[2][2];
	double a,da,lon_avg;


		/************* check for map ***********************************/
	if(!job_stats[0])
		return(1);

		/************* check for legitimate parameters *****************/
	if((ct < 1)  ||  (ct > 9))
		return(1);
	if(!registered(ct))
		return(1);

		/************* check for raster already registered *************/
	if(registered(0))
		return(0);

		/************* calc a N pointed window if possible *************/
	if((ct == 2) || (ct == 3))      /** utm or state plane registration   **/
	{	if(registered(1))           /** lat/lon registered **/
		{		/*** move window to center of utm zone **/
			check_llref();
			a = lon_ref < 0.0 ?  -lon_ref : lon_ref;
			a = a / 6.0;
			i = a;
			a = 6.0 * i + 3.0;
			a = lon_ref < 0.0 ?  -a : a;   /** a is in center of utm zone **/

			lon_avg = 0.0;
			for(i=0; i<4; i++)
				lon_avg += corner[1][i][0];
			lon_avg = lon_avg / 4.0;       /** avg longitude of registration **/

			da = a - lon_avg;              /** da is offset from center of zone **/

			for(i=0; i<4; i++)
			{	x = corner[ct][i][0];
				y = corner[ct][i][1];
				transform(&x,&y,ct,1);      /** change to lat/lon coords **/
				x = x + da;                 /** center in utm zone       **/
				transform(&x,&y,1,2);       /** change to utm coords     **/
				corner[21][i][0] = x;
				corner[21][i][1] = y;
			}
		}
		else
		{	for(i=0; i<4; i++)
			{	corner[21][i][0] = corner[ct][i][0];
				corner[21][i][1] = corner[ct][i][1];
			}
		}
	}
	else
	{	corner[21][i][0] = corner[ct][i][0];
		corner[21][i][1] = corner[ct][i][1];
	}

		/************* calc ht & width ratio ***************************/
	minmax[0][0] = minmax[1][0] = corner[21][0][0];
	minmax[0][1] = minmax[1][1] = corner[21][0][1];
	for(i=1; i<4; i++)
	{	if(corner[21][i][0] < minmax[0][0])
			minmax[0][0] = corner[21][i][0];
		if(corner[21][i][0] > minmax[1][0])
			minmax[1][0] = corner[21][i][0];
		if(corner[21][i][1] < minmax[0][1])
			minmax[0][1] = corner[21][i][1];
		if(corner[21][i][1] > minmax[1][1])
			minmax[1][1] = corner[21][i][1];
	}
	dx = minmax[1][0] - minmax[0][0];
	dy = minmax[1][1] - minmax[0][1];
	if((dx == 0.0)  ||  (dy == 0.0))
		return(1);
	
	hw_ratio = dy/dx;

		/************* put on trial corners with 5% margin *************/
	if(job_stats[1] < job_stats[2])
		margin = 0.05 * job_stats[1];
	else
		margin = 0.05 * job_stats[2];

		/************* find hw_raster ratio, without margin area ********/
	hw_raster = ((double) job_stats[2] - 2.0 * margin) / ((double) job_stats[1] - 2.0 * margin);


		/************* set ht & width of map area on raster ***********/
	if(hw_ratio > hw_raster) /* trial is taller & skinner than raster */
	{	ht = job_stats[2] - 2.0 * margin;
		wd = ht / hw_ratio;
	}
	else                         /* trial is shorter & fatter than raster */
	{	wd = job_stats[1] - 2.0 * margin;
		ht = wd * hw_ratio;
	}

	corner[8][0][0] = corner[8][1][0] = margin;
	corner[8][0][1] = corner[8][3][1] = margin;
	corner[8][2][0] = corner[8][3][0] = margin + wd;
	corner[8][1][1] = corner[8][2][1] = margin + ht;

	corner[9][0][0] = corner[9][1][0] = minmax[0][0];
	corner[9][0][1] = corner[9][3][1] = minmax[0][1];
	corner[9][2][0] = corner[9][3][0] = minmax[1][0];
	corner[9][1][1] = corner[9][2][1] = minmax[1][1];
	
		/************* calc rotation/trans/change scale parameters *****/
	if(get_params(params, corner[8], corner[9], 4))
		return(1);

		/************* calc corners of utm centered system *******/
	for(i=0; i<4; i++)
	{	x = corner[21][i][0];
		y = corner[21][i][1];
		corner[9][i][0] = params[0] * x  -  params[1] * y  +  params[2];
		corner[9][i][1] = params[1] * x  +  params[0] * y  +  params[3];
	}

		/************* move to raster reg  ********************************/
	for(i=0; i<4; i++)
	{	j = corner[9][i][0] + 0.5;
		corner[9][i][0] = j;
		j = corner[9][i][1] + 0.5;
		corner[9][i][1] = j;
	}

		/************* register raster ************************************/
	j = register_ltp(0,0,corner[9],4,(char) 0);

		/************* reset temp registers to 0 **************************/
	for(i=0; i<4; i++)
	{	corner[9][i][0] = 0.0;
		corner[9][i][1] = 0.0;
		corner[21][i][0] = 0.0;
		corner[21][i][1] = 0.0;
	}

		/************* return *********************************************/
	return(j);
}
