/*******************************************************************************
*       snap_vertex     (see manual, Internal Functions)                <in>   *
*    Searches the index_array / index_list or objects files for a vertex       *
*    or segment within r pixels of the input x,y location which is on levels   *
*    specified.    If a side is found, x & y are moved to the closest point,   *
*    then the adjacent verticies are checked to see if they are within r and   *
*    x & y are moved again accordingly.  Otherwise the distance to the nearest *
*    vertex/node is check and x & y are moved if within r of it.  The return   *
*    value is set to 1 if the input (x & y) point is moved.                    *
*                                                                              *
*    Arguments:    float x,y:     raster coordinates (pixels)                  *
*                  long level1:   lower bound of levels to snap to             *
*                  long level2:   upper bound of levels to snap to             *
*                  double r:      snap radius (pixels)                         *
*                  long p:        vetext preference code:                      *
*                                 0 = no preference                            *
*                                 1 = snap preferentially to verticies         *
*                                 2 = snap only to nodes/ends of segments      *
*                                 3 = - same as 2 -                            *
*                                                                              *
*    Returned Values: (long) 1 = level object found, input coordinates moved.  *
*                            0 = no object found within r, nothing moved.      *
*                           -1 = error.                                        *
*                                                                              *
*    Global Variables Modified:  none.                                         *
*                                                                              *
*                      j.dabritz  10/31/89                                     *
*******************************************************************************/

#include "../../incl/globals.h"
#include <math.h>

long snap_vertex(x_in,y_in,level1,level2,r,p)
float *x_in,*y_in;
long level1,level2;
double r;
long p;
{	
	INDEX_LIST index_list;
	COORD_LIST coord_list;
	V_OBJECTS  v_object;

	long level;
	long row, col;
	long x1, y1, x2, y2;
	long i, v;
	long v_begin, v_end;
	long return_val;

	double x,y;
	double x_out, y_out;
	double vx, vy;
	double last_vx, last_vy;
	double array[2][3];
	double dx, dy;
	double dist_squared, min_square;
	double r2;
	double r_squared;
	double dist_squared1, dist_squared2;
	
		/*** Initialize automatics ***/
	last_vx = last_vy = 0.0;

		/*********** check for map **************************************/
	if(!job_stats[0])
		return(-1);

		/*********** check for file type needed *************************/
	if((level1 < 32)  ||  (level2 < 32))  /* objects file record wanted   */
		return(0);	/**  can only snap to an inport level **/
	else if((level1 <= 95)  &&  (level2 <= 95))
	{	if(job_stats[15] == 0)
			return(0);
		if(level1 > level2)     /* reverse levels if necessary  */
		{	level = level1;
			level1 = level2;
			level2 = level1;
		}
	}
	else              /* illegal level, abort          */
		return(-1);

		/*********** check r ********************************************/
	if(r < 0.0)
		return(-1);

		/*********** set up output coords *******************************/
	x = *x_in;
	y = *y_in;
	return_val = 0;

		/*********** find index_array window ****************************/
	r2 = (x - r + 0.5);    /* x min (pixels) */
	x1 = r2;               /* x min (pixels) */
	x1 = x1 >> 5;          /* x min (index_array coords) */
	r2 = (x + r + 0.5);    /* x max (pixels) */
	x2 = r2;               /* x max (pixels) */
	x2 = x2 >> 5;          /* x max (index_array coords) */
	r2 = (y - r + 0.5);    /* y min (pixels) */
	y1 = r2;               /* y min (pixels) */
	y1 = y1 >> 5;          /* y min (index_array coords) */
	r2 = (y + r + 0.5);    /* y max (pixels) */
	y2 = r2;               /* y max (pixels) */
	y2 = y2 >> 5;          /* y max (index_array coords) */
		/** test for raster coverage *****************************/
	if((x2 < 0)  || (x1 > (job_stats[1] >> 5)))
		return(-1);
	if((y2 < 0)  || (y1 > (job_stats[2] >> 5)))
		return(-1);
		/** limit to range of index_array ***/
	if(x1 < 0) 
		x1 = 0;
	if(y1 < 0)
		y1 = 0;
	if(x2 > (job_stats[1] >> 5))
		x2 = (job_stats[1] >> 5);
	if(y2 > (job_stats[2] >> 5))
		y2 = (job_stats[2] >> 5);
	
		/*********** search for segment within r dist **********************/
		/** for each row & col of index array, search objects */
	min_square = (r * r) + .01;       
	r_squared = r * r;
	for(row = y1; row <= y2; row++)
	{	for(col = x1; col <= x2; col++)
		{		/** search all indicated objects **/
			index_list.next = index_array[row][col];
			for(; index_list.next;)
			{		/** get index_list record **/
				if(get_indexl(&index_list,index_list.next))
					return(-1);
				  /*** check for object type == segment ***/
				level = index_list.type / 10;
				if((level >= level1)  &&  (level <= level2))
				{	if((index_list.type % 10) <= 1)    /** node or segment **/
					{		/**** test every vertex & side for distance **/
						i = (index_list.v1 < index_list.v2)  ?  1 : -1;
						if(p < 2)
						{	v_begin = index_list.v1;
							v_end = index_list.v2;
						}
						else
						{	if(get_vobjects(&v_object,index_list.obj_rec))
								return(-1);
							v_begin = v_object.v1;
							v_end = v_object.v2;
						}

						for(v = v_begin; ; )
						{		/*** get vertex coords ***/
							if(get_v_coord(&coord_list,v))
								return(-1);

							vx = coord_list.x;
							vy = coord_list.y;
					
								/*** calculate distance to input point **/
							dx = x - vx;
							dy = y - vy;
							dist_squared = (dx * dx) + (dy * dy);
								
								/*** check against min dist ***/
							if(dist_squared < min_square)
							{	min_square = dist_squared;
								x_out = vx;
								y_out = vy;
								return_val = 1;
							}
						
								/* if not first point, then get dist to vector **/
							if((p < 2) && (v != v_begin))
							{	dx = vx - last_vx;
								dy = vy - last_vy;
								if((dx != 0.0) || (dy != 0.0))
								{	array[0][0] = dx;
									array[0][1] = dy;
									array[0][2] = x - last_vx;
									array[1][0] = dy;
									array[1][1] = -dx;
									array[1][2] = y - last_vy;
									if(!mat_solve(2,3,array))
									{		/*** see if pt perp. off side of vector **/
										if((array[0][2] > 0.0) && (array[0][2] < 1.0))
										{	dist_squared = array[1][2] * array[1][2] * 
                                     		((dx * dx) + (dy * dy));
												/*** check against min dist ***/
											if(dist_squared < min_square)
											{	min_square = dist_squared;
												return_val = 1;
												x_out = last_vx + array[0][2] * dx;
												y_out = last_vy + array[0][2] * dy;

													/** if vertex preference is 1 then    **/
													/** check dist to both ends of segment */
												if(p)
												{	dx = x_out - last_vx;
													dy = y_out - last_vy;
													dist_squared1 = (dx*dx) + (dy*dy);
													dx = x_out - vx;
													dy = y_out - vy;
													dist_squared2 = (dx*dx) + (dy*dy);
													
													if(dist_squared1 < dist_squared2)
													{	if(dist_squared1 < r_squared )
														{	x_out = last_vx;
															y_out = last_vy;
														}
													}
													else
													{	if(dist_squared2 < r_squared )
														{	x_out = vx;
															y_out = vy;
														}
													}
												}
											}
										}
									}
								}
							}
				
								/*** check to see if verticies done ***/
							if(v == v_end)
								break;

								/*** reset last_v coords ***/
							last_vx = vx;
							last_vy = vy;

								/*** increment vertex **/
							if(p < 2)
						   	v += i;
							else
								v = v_end;
		
						}
					}
				}
			}
		}
	}

		/*********** return closest segment record number or -1 ************/
	if(return_val)
	{	*x_in = x_out;
		*y_in = y_out;
	}

	return(return_val);
}
