/*******************************************************************************
*       stpln_to_geo    (see manual, Internal Functions)                <in>   *
*                 This program converts state plane coordinates to geodetic    *
*                 (longitude, latitude) coordinates in decimal degrees.        *
*                                                                              *
*                 No global or file data is changed.  The state plane          *
*                 parameters in register.h are used for the conversion.        *
*                      stpln_params[0] -  zone type:  1 = lambert              *
*                           (projection type)         2 = transverse mercator  *
*                                                     3 = Alaska zones 2-9     *
*                                                     4 = Alaska zone 1        *
*                                                     5 = azmthl equdist (Guam)*
*                      stpln_params[1]-[15] - projection parameters.  The      *
*                          number of parameters vary with the projection type. *
*                                                                              *
*                 Note:  These values and the formulae to use them were taken  *
*                 from"STATE PLANE COORDINATES BY AUTOMATIC DATA PROCESSING"  *
*                 by Charles N Claire, Geodesist (retired)                     *
*                 published by the Department of Commerce, Environmental       *
*                 Science Services Administration, Coast and Geodetic Survey,  *
*                 Publication 62-4                                             *
*                                                                              *
*                 Returned values:  0 (if zone is found)  or -1 (if zone not   *
*                                   is not found)                              *
*                                                                              *
*                 j.dabritz  5/11/89                                           *
*******************************************************************************/

#include"../../incl/globals.h"
#include <math.h>

long stpln_to_geo(xin, yin, xout, yout)
double xin, yin, *xout, *yout;  /* registers for E,N(input), lon,lat (output) */
{
	long projection;
	double *l, *t;

	double a;
	double b;
	double alpha;
	double c, cm;
	double d;
	double dlambdaa, dlambda1;
	double e, e2;
	double epsilon;
	double f;
	double fi, fi_min, fi_min_min, fi_min_sec, fi_rad, fi_sec;
	double fi1, fi1_sec;
	double fi2, fi2_sec;
	double g;
	double h;
	double i;
	double kc;
	double lambda;
	double lambda0, lambda_sec;
	double latitude, longitude;
	double mu;
	double pcf1, pcf2, pcf4;
	double pi;
	double r;
	double s, sa, sg1, sm, s1, s1t, s2, s2t, s3, s3t;
	double theta, theta_sec;
	double u;
	double v;
	double w, w_min, w_min_sec, w_sec;
	double w1, w1_sec, w2, w2_sec, w3, w3_sec;
	double x, xr;
	double y;

	
		/*  set projection code    */
	projection = stpln_params[0];

		/*  check for a legal projection code  */
        /*      code     :  1 = lambert              */
        /*                  2 = transverse mercator  */
        /*                  3 = Alaska zones 2-9     */
        /*                  4 = Alaska zone 1        */
        /*                  5 = azmthl equdist (Guam)*/
	if((projection<1) || (projection>5))
		return -1;

		/*  equivalance the variable array l[] with  stpln_params[] so that
			the work below will visually match the formulae from the book  */
	l =  stpln_params;

		/* set input varaibles    */
	x = xin;
	y = yin;


		/* switch by projection code and convert data   */
	switch (projection)   {
	case 1:				/*   lambert                 */
		/*  equivalance the variable array l[] with  stpln_params[] so that
			the work below will visually match the formulae from the book  */
		l =  stpln_params;

		theta = atan( (x - l[1]) / (l[4] - y));          /* theta  (radians) */
		theta_sec = theta / M_PI * 180 * 60 *60;         /* theta  (seconds) */

		lambda_sec = l[2] - ( theta_sec / l[6] );        /* lambda (seconds) */
		longitude = - (lambda_sec / 60 / 60);            /* E lon (dec degrees*/	

		r = (l[4] - y) / (cos(theta));

		s1 = (l[4] - l[3] -y + 2*r*pow(sin(theta/2),2.0)) / (l[5]);

		s1t = s1/100000000;                          
		s2 = s1 / (1 + l[9]*pow(s1t,2.0) - l[10]*pow(s1t,3.0) + 
			l[11]*pow(s1t,4.0));

		s2t = s2/100000000;
		s3 = s1 / (1 + l[9]*pow(s2t,2.0) - l[10]*pow(s2t,3.0) + 
			l[11]*pow(s2t,4.0));

		s3t = s3/100000000;
		s = s1 / (1 + l[9]*pow(s3t,2.0) - l[10]*pow(s3t,3.0) + 
			l[11]*pow(s3t,4.0));


		w_min = l[7] - 600;

		w_sec = 36000 + l[8] - 0.009873675553*s;

		w = w_min + w_sec/60;                            /*  in minutes   */
		w = w / 60 / 180 * M_PI;                         /*  in radians   */

		fi_min = l[7] - 600;

		fi_sec = w_sec + (1047.546710 + (6.192760 + 0.050912*pow(cos(w),2.0))*
				pow(cos(w),2.0))*sin(w)*cos(w);

		fi = fi_min + fi_sec/60;                        /*  in minutes   */
		latitude = fi/60;                               /* N lat in degrees */

		break;

 			/*-------------------------------------------------------*/


	case 2:				/*   transverse mercator     */
		/*  equivalance the variable array t[] with  stpln_params[] so that
			the work below will visually match the formulae from the book  */
		t =  stpln_params;

		sg1 = x - t[1] - t[6]*pow(((x - t[1])/100000),3.0);

		sm = (0.3048006099 / t[5]) * (x - t[1] - t[6]*pow((sg1/100000),3.0));

		w_min = t[3];

		w_sec = t[4] + (0.009873675553 / t[5]) * y;

		w = w_min + w_sec/60;                            /* in minutes  */
		w = w / 60 / 180 * M_PI;                         /* in radians  */

		fi_min_min = t[3];

		fi_min_sec = w_sec + (1047.546710 + (6.192760 + 0.050912*
			pow(cos(w),2.0))* pow(cos(w),2.0)) * sin(w) * cos(w);

		fi_min = fi_min_min + fi_min_sec/60;

		fi_sec = fi_min_sec - 25.52381 * 
			pow( (1 - 0.0067686580 * pow(sin(fi_min/60/180*M_PI),2.0)),2.0) *
			pow((sm/100000),2.0) * tan(fi_min/60/180*M_PI);
		

		fi = fi_min_min + fi_sec/60;                /* in minutes   */
		latitude = fi / 60;             /* North latitude, decimal degrees */
		fi = latitude / 180 * M_PI;     /*  in radians   */


		sa = sm - 4.0831 * pow((sm/100000),3.0);

		s1 = sm - 4.0831 * pow((sa/100000),3.0);

		dlambda1 = s1 * pow(( 1 - 0.0067686580*pow(sin(fi),2.0)),0.5) /
			(30.92241724 * cos(fi));

		dlambdaa = dlambda1 + 3.9174 * pow(dlambda1/10000, 3.0);

		lambda_sec = t[2] - dlambda1 - 3.9174 * pow(dlambdaa/10000,3.0);
		longitude = - (lambda_sec / 60 / 60);   /* E. longitude, dec. degrees*/

		break;

 			/*-------------------------------------------------------*/


	case 3:				/*   Alaska zones 2-9        */
		/*  equivalance the variables c and cm with  stpln_params[1&2] so that
			the work below will visually match the formulae from the book  */
		c =  stpln_params[1];
		cm =  stpln_params[2];

		w_min_sec = 193900.054420 + 0.00987466302498*y;
		w_min = w_min_sec/60/60/180*M_PI;                /*  radians  */

		fi_min_sec = w_min_sec + ( 1047.546691 + 6.193011*
			pow(cos(w_min),2.0) + 5.0699/100 * pow(cos(w_min),4.0)) *
			pow((1 - pow(cos(w_min),2.0)), 0.5) * cos(w_min);
		fi_min = fi_min_sec/60/60/180*M_PI;               /*  radians  */

		pcf1 = cos(fi_min);
		pcf2 = pcf1 * pcf1;
		pcf4 = pcf2 * pcf2;
		fi_sec = fi_min_sec - 233.9736450 * pow(((x-c)/1000000),2.0) *
			pow((1 + 0.0068147849*pcf2),2.0) *
			pow(((1 / pcf2 ) - 1), 0.5) *
			( 1 - 1.89056040/10000 * pow(((x - c) / 1000000),2.0) *
				( 1.9591113 + 3/pcf2 + 8.1359/100 * pcf2 + 
					2.79/10000 * pcf4 ) +
				1.42969/100000000 * pow(((x-c)/1000000),4.0) *
				( 1 + 0.0068147849 * pcf2 ) *
				( 15.5 + 45/pcf4 - 0.307/pcf2 + 1.53*pcf2));
		latitude = fi_sec / 60 / 60;              /*  N lat, decimal degrees */

		lambda_sec = cm - 9824.513072 * 
			pow((1 + 0.0068147849 * pcf2), 0.5) / pcf1 * ((x-c) / 1000000) *
			( 1 - 3.78112080/10000 * (1 + 0.0068147849 * pcf2) *
				pow(((x - c) / 1000000 ), 2.0) *
				( -1 + 2/pcf2 + 0.0068147849 * pcf2 ) +
				4.2890624/100000000 *
				pow((1 + 0.0068147849 * pcf2), 2.0) *
				pow(((x - c)/1000000), 4.0) *
				( 1.054 + 24/pcf4 - 20/pcf2 - 1.36/100*pcf2));
		longitude = - (lambda_sec / 60 / 60);     /*  E lon, decimal degrees */

		break;

 			/*-------------------------------------------------------*/


	case 4:				/*   Alaska zone 1           */
		/*  equivalance the variables a-i and e-lambda0 with 
			 stpln_params[1-15] so that the work below will visually 
			match the formulae from the book  */
		a =  stpln_params[1];
		b =  stpln_params[2];
		c =  stpln_params[3];
		d =  stpln_params[4];
		f =  stpln_params[5];
		g =  stpln_params[6];
		h =  stpln_params[7];
		i =  stpln_params[8];
		e =  stpln_params[9];
		e2 =  stpln_params[10];
		pi =  stpln_params[11];
		epsilon =  stpln_params[12];
		kc =  stpln_params[13];
		alpha =  stpln_params[14];
		lambda0 =  stpln_params[15];

		u = (-0.182880365761)*x + 0.243840487681*y + 7000000.0000;

		v = 0.243840487681*x + 0.182880365761*y - 1000000.0000;

		r = (pow(epsilon,v/d) - pow(epsilon,-v/d)) / 2;

		s = (pow(epsilon,v/d) + pow(epsilon,-v/d)) / 2;

		mu = (1 / (2*b)) * 
			log((s + f*r + g*sin(u/d)) / (s- f*r - g*sin(u/d)) ) -
			c / b;

		xr = 2 * atan(pow(epsilon,mu)) - pi/2;

		fi_rad = xr + (0.006761032571 + 0.000053172205*pow(cos(xr),2.0) +
				0.000000573027*pow(cos(xr),4.0) +
				0.000000007128*pow(cos(xr),6.0) ) * sin(xr) * cos(xr);

		fi_sec = 206264.806247 * fi_rad;
		latitude = fi_sec / 60 / 60;		/* N latitude in decimal degrees */


		lambda = lambda0 + 1/b * 180/M_PI * 
			atan((f * sin(u/d) - g*r) / (cos(u/d)));
		longitude = -lambda;      /* E longitude in decimal degrees   */
			

		break;

 			/*-------------------------------------------------------*/


	case 5:				/*   azmthl equdist (Guam)   */
		/*  No variables are necessary in the Guam projection as they are  
			all imbeded in its unique projection formula.    */

		w1_sec = 48263.2837702 + 0.03239388390 *
			(y - 50000.0000 - pow((x - 50000.0000)/10000,2.0) * 1.87770 );
		w1 = w1_sec / 60 / 60 / 180 * M_PI;     /*   in radians  */

		fi1_sec = w1_sec + ( 1047.546691 + 6.193011*pow(cos(w1),2.0) +
			0.050699*pow(cos(w1),4.0)) * sin(w1) * cos(w1);
		fi1 = fi1_sec / 60 / 60 / 180 * M_PI;   /*   in radians  */

		w2_sec = 48263.2837702 + 0.03239388390 * 
			( y - 50000.0000 - pow((( x - 50000.0000 ) / 10000), 2.0) *
				((tan(fi1) * pow((1 - 0.0067686580 * pow(sin(fi1),2.0)), 0.5)) 
					/ 0.127564128));
		w2 = w2_sec / 60 / 60 / 180 * M_PI;     /*   in radians  */

		fi2_sec = w2_sec + ( 1047.546691 + 6.193011*pow(cos(w2),2.0) +
				0.050699 * pow(cos(w2),4.0) ) * sin(w2) * cos(w2);
		fi2 = fi2_sec / 60 / 60 / 180 * M_PI;   /*   in radians  */

		w3_sec = 48263.2837702 + 0.03239388390 *
			( y - 50000.0000 - pow((( x - 50000.0000) / 10000),2.0) *
				((tan(fi2) * pow((1 - 0.0067686580 * 
					pow( sin(fi2), 2.0)), 0.5)) / 0.127564128 ));
		w3 = w3_sec / 60 / 60 / 180 * M_PI;     /*   in radians  */

		fi_sec = w3_sec + ( 1047.546691 + 6.193011 * pow(cos(w3),2.0) +
				0.050699 * pow(cos(w3), 4.0) ) * sin(w3) * cos(w3);
		fi = fi_sec / 60 / 60 / 180 * M_PI;     /*  in radians   */
		latitude = fi_sec / 60 / 60;            /*  N lat in decimal degrees */

		lambda_sec = 521095.50254 + 
			( x - 50000.0000) * pow((1 - 0.0067686580 * 
				pow(sin(fi),2.0)), 0.5) / (30.92241724 * cos(fi));
		longitude = lambda_sec / 60 / 60;       /* E longitude in dec. degrees*/
				
		break;
	}

	*xout = longitude;
	*yout = latitude;

	return 0;
}
