/*******************************************************************************
*       transform       (see manual, Internal Functions)                <in>   *

     Purpose:	convert plane coordinates freely between coordinate systems

     Arguments:	long      coordsysin     coordinate system number of input
				long	coordsysout		coordinate system number of output
				double  xin,yin,xout,yout

     Returned Values: (long) 0=ok,    -1=failure.

     Global Variables Modified:  - none -

	Note:  this code was adapted from LIDES (SUB5, SUB Transform),
	and the comments are directly from that code.

*                      j.dabritz  3/9/89                                       *
*******************************************************************************/

#include <math.h>
#include "../../incl/globals.h"


long transform(xinout,yinout,in,out)
    double *xinout, *yinout;       /*   pointers to x & y coordinates   */
    long   in,out;       /*   numbers to input and output coordinate systems */
{
    /*     CORNER(A,B,C)                                                  */
    /*   A - SPECIFIES CORNER COORDINATE SYSTEM                     */
    /*   B - SPECIFIES CORNER 0=SW, 1=NW, 2=NE, 3=SE                    */
    /*   C - SPECIFIES COORDINATE, 0=X, 1=Y                              */
    /*                                                            */
    /*-------------------------------------------------------           */
    /*        (Proportional Rubbersheeting)                              */
    /*   This transformation is between two convex quadrilaterals.          */
    /*   Any line made by connecting points simmiliarly proportioned          */
    /*   from Left to Right on the Top and Bottom (or Bottom to Top on the*/
    /*   Left and Right Sides) will be mapped to a similiarly defined*/
    /*   line in the new mapping system.  Points are transformed by           */
    /*   finding the porporitonal lines through the point, constructing*/
    /*   similiar lines in the new system, and taking their intersection.*/
    /*                                                            */
    /*   The transformation assumes that no 3 corner points are in a          */
    /*   line, and is defined for all points except the intersections of */
    /*   of lines containing opposite sides, and points inside the parabola*/
    /*   tangent to all 4 lines containing the sides. Corners are mapped*/
    /*   to corners, sides are mapped to sides.  Points beyond the           */
    /*   intersection of opposite sides exhibit a projection of a          */
    /*   folded plane.                                                  */
    /*                                                            */
    /*   Note:  this transformation can be linear or nonlinear depending*/
    /*   on the corners input.  It will translate, rotate, change the */
    /*   scale in any direction, reflect, and rubbersheet in any combination*/
    /*   at the same time depending on the input and output corners.          */
    /*                                                            */
    /*   Any error (caused by non convex quadralaterals or undefined points)*/
    /*   causes the return of (0,0)).                                        */
	/*   This routine finds the point in question by direct methods.  The    */
	/*   dirivation follows.  John Dabritz  12/5/86                           */
    /*                                                             */
    /*                                                            */
    /* SOLUTION:   (USE PICTURE BELOW) OF THE TRAMSFORMATION          */
    /*                                                            */
    /*                              Map Border Shown with corners A,B,C,D*/
    /*                B**                                                 */
    /*                *  **F**                                        */
    /*                *       *****                                        */
    /*                *            *****                              */
    /*               *                  *****                              */
    /*               *                       *****                    */
    /*               G                            ****C                    */
    /*               *    Z                           *                    */
    /*               *                               *                    */
    /*              *                               *                    */
    /*              *                               H                    */
    /*              *                              *                    */
    /*              *                             *                    */
    /*              *                            *                    */
    /*             *                            *                    */
    /*             *                            *                    */
    /*             *                           *                    */
    /*             *             **************D                    */
    /*             A***E*********                                        */
    /*                                                            */
    /*  Given:  Corners A,B,C,D and any point Z                              */
    /*                      Define V1=D-A                              */
    /*                             V2=C-B                              */
    /*                             V3=Z-A                              */
    /*                             V4=Z-B                              */
    /*                                                            */
    /*  Define scaler n:    E=A+nV1                                        */
    /*                      F=B+nV2                                        */
    /*                 and  Z is an element of line EF                     */
    /*         scaler m:    G=A+m(B-A)                                        */
    /*                      H=D+m(C-D)                              */
    /*                 and  Z is an element of line GH                    */
    /*                                                  */
    /*  Note:  The fact is used, (but the proof not shown here) that the*/
    /*         ratio n=AE:AD=BF:BC is also equal to GZ:GH.                    */
    /*                                                            */
    /*  Find scaler n.  (Similiar solution for scaler m.)                    */
    /*                                                            */
    /*      Formula for computation of scaler n:                              */
    /*                                                            */
    /*      (Zx-Fx)(Zy-Ey)-(Zx-Ex)(Zy-Fy)=0        Det=0 since Z is on EF*/
    /*                                                            */
    /*      (Zx-Bx-nV2x)(Zy-Ay-nV1y)-(Zx-Ax-nV1x)(Zy-By-nV2y)=0   subst.*/
    /*                                                            */
    /*      (V4x-nV2x)(V3y-nV1y)-(V3x-nV1x)(V4y-nV2y)=0           subst.*/
    /*                                                            */
    /*      (n^2)(V2xV1y-V1xV2y)+                                 mult. &*/
    /*        (n)(-V4xV1y-V2xV3y+V3xV2y+V1xV4y)+                    assoc.*/
    /*           (V4xV3y-V3xV4y)=0                                        */
    /*                                                            */
    /*      n=[-K+-(K^2-4JL)^.5]/(2J)                             quadratic*/
    /*         where   J=V2xV1y-V1xV2y                              */
    /*                 K=-V4xV1y-V2xV3y+V3xV2y+V1xV4y                    */
    /*                 L=V4xV3y-V3xV4y                              */
    /*                                                            */
    /*      Note:  Both solutions may be legitimate, we want the one          */
    /*             closest to the interval 0 to 1                    */
    /*                                                             */

char string[256];

	double ex,ey,fx,fy,gx,gy,hx,hy,j,k,l,n1,n2;
	double v1x,v1y,v2x,v2y,v3x,v3y,v4x,v4y;
	double radical;
	double ma[2][3];
	double x,y;
	double p[2];
	double vx,vy;

			/* check to see if in & out are the same     */
	if(in == out)
		return(0);

			/* check to see if geodetics are input or output  */
			/*        start here     */
	if(in == 1)       /* input is geodetics    */
	{		/*   convert to utm    */
		if(geo_to_utm( *xinout, *yinout, xinout, yinout))
		{	*xinout = *yinout = 0.0;
			return(-1);
		}
			/*   convert from utm to output     */
		return(transform(xinout,yinout,2,out));
	}
	else if(out == 1)    /*   output is geodetics   */
	{		/*   convert input to utm               */
		if(transform(xinout,yinout,in,2))
			return(-1);
			/*   convert utm to geodetics           */
		return(utm_to_geo(2,xinout,yinout,xinout,yinout));
	}

			/* set x and y and default *xinout, *yinout  */
	x = *xinout;
	y = *yinout;
	*xinout = 0;
	*yinout = 0;


	/*  find proportion along 2 sides       */
		/*   calculate side vectors       */
	v1x=corner[in][3][0]-corner[in][0][0];
	v1y=corner[in][3][1]-corner[in][0][1];
	v2x=corner[in][2][0]-corner[in][1][0];
	v2y=corner[in][2][1]-corner[in][1][1];
                                             
		/*  compute j using formula above   */
	j=v2x*v1y-v1x*v2y;
                                             
		/*  check oposite sides for parallelality   */
	if(fabs(j) <= (.0000001*sqrt((v1x*v1x+v1y*v1y)*(v2x*v2x+v2y*v2y))))
	{		/*  lines are parallel, solve matrix for proportion on sides */
		ma[0][0] = v1x;
		ma[0][1] = corner[in][1][0] - corner[in][0][0];
		ma[0][2] = x - corner[in][0][0];
		ma[1][0] = v1y;
		ma[1][1] = corner[in][1][1] - corner[in][0][1];
		ma[1][2] = y - corner[in][0][1];
			/*  check for an insolvable matrix              */
		if(mat_solve(2,3,ma))
		{
sprintf(string,"transform abort 1.  cannot solve matrix. %f,  %f",x,y);
message(string);
			return(-1);
		}
			/*  set proportion p                            */
		p[1] = ma[1][2];

			/*  solve for proportion on top & bottom   */
		gx = corner[in][0][0] + p[1]*(corner[in][1][0] - corner[in][0][0]);
		gy = corner[in][0][1] + p[1]*(corner[in][1][1] - corner[in][0][1]);

		hx = corner[in][3][0] + p[1]*(corner[in][2][0] - corner[in][3][0]);
		hy = corner[in][3][1] + p[1]*(corner[in][2][1] - corner[in][3][1]);

		vx = hx - gx;
		vy = hy - gy;
		if(fabs(vx) > fabs(vy))
		{	if(vx == 0)
			{
sprintf(string,"transform abort 2.  vx == 0. %f,  %f",x,y);
message(string);
				return(-1);
			}
			p[0] = (x - gx) / vx;
		}
		else
		{	if(vy == 0)
			{
sprintf(string,"transform abort 3.  vy == 0. %f,  %f",x,y);
message(string);
				return(-1);
			}
			p[0] = (y - gy) / vy;
		}
	}
	else
	{		/*  not parallel, solve for p using formula     */
		v3x=x-corner[in][0][0];
		v3y=y-corner[in][0][1];
		v4x=x-corner[in][1][0];
		v4y=y-corner[in][1][1];
		k = -v4x*v1y - v2x*v3y + v3x*v2y + v1x*v4y;
		l =  v4x*v3y - v3x*v4y;

		radical = ( k*k - 4*j*l);
		if(radical < 0)
		{
sprintf(string,"transform abort 4.  vy == 0. %f,  %f",x,y);
message(string);
			return(-1);
		}
		radical = sqrt(radical);
		n1 = (-k + radical) / (2*j);
		n2 = (-k - radical) / (2*j);

			/*  choose closest interval to .5  */
		if( (fabs(n1-.5)) < (fabs(n2-.5)) )
			p[0] = n1;
		else
			p[0] = n2;

			/*  solve for proportion on sides          */
		ex = corner[in][0][0] + p[0]*(corner[in][3][0] - corner[in][0][0]);
		ey = corner[in][0][1] + p[0]*(corner[in][3][1] - corner[in][0][1]);

		fx = corner[in][1][0] + p[0]*(corner[in][2][0] - corner[in][1][0]);
		fy = corner[in][1][1] + p[0]*(corner[in][2][1] - corner[in][1][1]);

		vx = fx - ex;
		vy = fy - ey;
		if(fabs(vx) > fabs(vy))
		{	if(vx == 0)
			{
	sprintf(string,"transform abort 5.  vx == 0 (2) . %f,  %f",x,y);
	message(string);
				return(-1);
			}
			p[1] = (x - ex) / vx;
		}
		else
		{	if(vy == 0)
			{
	sprintf(string,"transform abort 6.  vy == 0 (2) . %f,  %f",x,y);
	message(string);
				return(-1);
			}
			p[1] = (y - ey) / vy;
		}
	}
	
		/*  comput intersect in out coordinate system   */
		/*  note:  h,g corrispont to h & g above in the output coord system */
	hx = corner[out][3][0] + ((corner[out][2][0] - corner[out][3][0]) * p[1]);
	hy = corner[out][3][1] + ((corner[out][2][1] - corner[out][3][1]) * p[1]);
	gx = corner[out][0][0] + ((corner[out][1][0] - corner[out][0][0]) * p[1]);
	gy = corner[out][0][1] + ((corner[out][1][1] - corner[out][0][1]) * p[1]);

	*xinout = gx + ((hx - gx) * p[0]);
	*yinout = gy + ((hy - gy) * p[0]);

	return(0);
}
